// ESP.js implements functions for the Engineering Sketch Pad (ESP)
// written by John Dannenhoffer and Bob Haimes

// functions expected by wv
//    wvInitUI()
//    wvUpdateUI()
//    wvServerMessage(text)

// functions associated with button presses
//    activateBuildButton()
//    cmdUndo()
//    cmdEdit()
//    cmdSave()
//    cmdTest()
//    cmdHelp()

// functions associated with mouse clicks and context-menu selections
//    addPmtr()
//    clrVels()
//    editPmtr(e)
//    editPmtrOk()
//    editPmtrCancel()
//    compSens()
//    setVel(e)
//    addBrch()
//    addBrchOk()
//    addBrchCancel()
//    delBrch(e)
//    editBrch(e)
//    editBrchOk()
//    editBrchCancel()
//    addAttr()
//    editAttr(e)
//    buildTo()
//    editCsmOk()
//    editCsmCancel()

// functions associated with the mouse in the canvas
//    getCursorXY(e)
//    getMouseDown(e)
//    getMouseUp(e)
//    mouseLeftCancas(e)
//    getKeyPress(e)
//    getKeyDown(e)
//    getKeyUp(e)

// functions associated with the mouse in the key window
//    setKeyLimits(e)

// functions associated with toggling display settings
//    toggleViz(e)
//    toggleGrd(e)
//    toggleTrn(e)
//    toggleOri(e)

// functions associated with a Tree in the tlframe
//    Tree(doc, treeId) - constructor
//    TreeAddNode(iparent, name, gprim, click, cmenu, prop1, valu1, cbck1, prop2, valu2, cbck2, prop3, valu3, cbck3)
//    TreeBuild()
//    TreeClear()
//    TreeContract(inode)
//    TreeExpand(inode)
//    TreeProp(inode, iprop, onoff)
//    TreeUpdate()

// helper functions
//    changeMode(newMode)
//    postMessage(mesg)
//    resizeFrames()
//    setupEditBrchForm()
//    setupEditPmtrForm()
//    numberOfPmtrChanges()
//    numberOfBrchChanges()

"use strict";


//
// callback when the user interface is to be initialized (called by wv-render.js)
//
function wvInitUI()
{
    // alert("wvInitUI()");

    // set up extra storage for matrix-matrix multiplies
    wv.uiMatrix   = new J3DIMatrix4();
    wv.saveMatrix = new J3DIMatrix4(wv.mvMatrix);

                                   // ui cursor variables
    wv.cursorX   = -1;             // current cursor position
    wv.cursorY   = -1;
    wv.keyPress  = -1;             // last key pressed
    wv.keyCode   = -1;
    wv.startX    = -1;             // start of dragging position
    wv.startY    = -1;
    wv.button    = -1;             // button pressed
    wv.modifier  =  0;             // modifier (shift,alt,cntl) bitflag
    wv.flying    =  1;             // flying multiplier (do not set to 0)
    wv.offTop    =  0;             // offset to upper-left corner of the canvas
    wv.offLeft   =  0;
    wv.dragging  =  false;         // true during drag operation
    wv.picking   =  0;             // keycode of command that turned picking on
    wv.locating  =  0;             // keycode of command that turned locating on
    wv.focus     = [0, 0, 0, 1];   // focus data needed in locating
    wv.pmtrStat  =  0;             // -2 latest Parameters are in Tree
                                   // -1 latest Parameters not in Tree (yet)
                                   //  0 need to request Parameters
                                   // >0 waiting for Parameters (request already made)
    wv.brchStat  =  0;             // -2 latest Branches are in Tree
                                   // -1 latest Branches not in Tree (yet)
                                   //  0 need to request Branches
                                   // >0 waiting for Branches (request already made)
    wv.builtTo   = 99999;          // last Branch in previous successful build
    wv.menuEvent = undefined;      // event associated with click in Tree
    wv.server    = undefined;      // string passed back from "identify;"
    wv.loLimit   = -2;             // lower limit in key
    wv.upLimit   = +2;             // upper limit in key
    wv.nchanges  = -1;             // number of changes since last getCsmFile (-1 to initialize)
    wv.filename  = "";             // name of the .csm file
    wv.curFile   = "";             // contents of current .csm file
    wv.curMode   =  0;             //-1 to disable buttons, etc.
                                   // 0 show WebViewer in canvas
                                   // 1 show addBrchForm
                                   // 2 show editBrchForm with addBrchHeader
                                   // 3 show editBrchForm with editBrchHeader
                                   // 4 show editPmtrForm with addPmtrHeader
                                   // 5 show editPmtrForm with editPmtrHeader
                                   // 6 show editCsmForm
    wv.curPmtr   = -1;             // Parameter beibng editted (or -1)
    wv.curBrch   = -1;             // Branch being editted (or -1)
    wv.afterBrch = -1;             // Branch to add after (or -1)
    wv.numArgs   =  0;             // number of arguments in wv.curBrch
//  wv.centerV                     // set to 1 to center view and rotation
//  wv.pick                        // set to 1 to turn picking on
//  wv.locate                      // set to 1 to turn locating on
//  wv.sceneGraph                  // pointer to sceneGraph
//  wv.picked                      // sceneGraph object that was picked
//  wv.located                     // sceneGraph object that was located
//  wv.sceneUpd                    // should be set to 1 to re-render scene
//  wv.sgUpdate                    // =1 if the sceneGraph has been updated
//  wv.canvasID                    // ID of main canvas
//  wv.canvasKY                    // ID of key  canvas
//  wv.drawKey                     // =1 if key window needs to be redrawn
//  wv.socketUt.send(text)         // function to send text to server
//  wv.plotAttrs                   // plot attributes

    var canvas = document.getElementById(wv.canvasID);
      canvas.addEventListener('mousemove',  getCursorXY,     false);
      canvas.addEventListener('mousedown',  getMouseDown,    false);
      canvas.addEventListener('mouseup',    getMouseUp,      false);
      canvas.addEventListener('mouseout',   mouseLeftCanvas, false);

    var keycan = document.getElementById(wv.canvasKY)
      keycan.addEventListener('mouseup',    setKeyLimits,    false);

    document.addEventListener('keypress',   getKeyPress,     false);
    document.addEventListener('keydown',    getKeyDown,      false);
    document.addEventListener('keyup',      getKeyUp,        false);
}

//
// callback when the user interface should be updated (called by wv-render.js)
//
function wvUpdateUI()
{
    // alert("wvUpdateUI()");

    // special code for delayed-picking mode
    if (wv.picking > 0) {

        // if something is picked, post a message
        if (wv.picked !== undefined) {

            // second part of '^' operation
            if (wv.picking == 94) {
                var mesg = "Picked: "+wv.picked.gprim;

                try {
                    var attrs = sgData[wv.picked.gprim];
                    for (var i = 0; i < attrs.length; i+=2) {
                        mesg = mesg + "\n        "+attrs[i]+"= "+attrs[i+1];
                    }
                } catch (x) {
                }
                postMessage(mesg);
            }

            wv.picked  = undefined;
            wv.picking = 0;
            wv.pick    = 0;

        // abort picking on mouse motion
        } else if (wv.dragging) {
            postMessage("Picking aborted");

            wv.picking = 0;
            wv.pick    = 0;
        }

        wv.keyPress = -1;
        wv.dragging = false;
    }

    // special code for delayed-locating mode
    if (wv.locating > 0) {

        // if something is located, post a message
        if (wv.located !== undefined) {

            // second part of '@' operation
            if (wv.locating == 64) {
                var xloc = wv.focus[3] * wv.located[0] + wv.focus[0];
                var yloc = wv.focus[3] * wv.located[1] + wv.focus[1];
                var zloc = wv.focus[3] * wv.located[2] + wv.focus[2];

                postMessage("Located: x="+xloc.toFixed(4)+", y="+yloc.toFixed(4)+
                                   ", z="+zloc.toFixed(4));
            }

            wv.located  = undefined;
            wv.locating = 0;
            wv.locate   = 0;

        // abort locating on mouse motion
        } else if (wv.dragging) {
            postMessage("Locating aborted");

            wv.locating = 0;
            wv.locate   = 0;
        }

        wv.keyPress = -1;
        wv.dragging = false;
    }

    // if the Parameters are scheduled to be updated, send a message to
    //    get the Parameters now
    if (wv.pmtrStat > 0) {
        wv.pmtrStat--;
    } else if (wv.pmtrStat == 0) {
        try {
            wv.socketUt.send("identify;");
            wv.socketUt.send("getPmtrs;");
            wv.pmtrStat = 6000;
        } catch (e) {
            // could not send command
        }
    }

    // if the Branches are scheduled to be updated, send a message to
    //    get the Branches now
    if (wv.brchStat > 0) {
        wv.brchStat--;
    } else if (wv.brchStat == 0) {
        try {
            wv.socketUt.send("getBrchs;");
            wv.brchStat = 6000;
        } catch (e) {
            // could not send command
        }
    }

    // if the CsmFile has changed, send a message to
    //    get the CsmFile now
    if (wv.nchanges < 0) {
        try {
            wv.socketUt.send("getCsmFile;");
            wv.nchanges = 0;
        } catch (e) {
            // could not send message
        }
    }

    // if the scene graph and Parameters have been updated, (re-)build the Tree
    if ((wv.sgUpdate == 1 && wv.pmtrStat <= -1 && wv.brchStat <= -1) ||
        (                    wv.pmtrStat == -1 && wv.brchStat == -2) ||
        (                    wv.pmtrStat == -2 && wv.brchStat == -1)   ) {

        if (wv.sceneGraph === undefined) {
            alert("wv.sceneGraph is undefined --- but we need it");
        }

        // if there was a previous Tree, keep track of whether or not
        //    the Parameters, Branches, and Display was open
        var pmtr1Open = 0;
        var pmtr2Open = 0
        var brchsOpen = 0;

        if (myTree.opened.length > 4) {
            pmtr1Open = myTree.opened[1];
            pmtr2Open = myTree.opened[2];
            brchsOpen = myTree.opened[3];
        }

        // clear previous Nodes from the Tree
        myTree.clear();

        // put the group headers into the Tree
        myTree.addNode(0, "Design Parameters", "", addPmtr, "*");
        myTree.addNode(0, "Local Variables",   "", null,    "" );
        myTree.addNode(0, "Branches",          "", addBrch, "*");
        myTree.addNode(0, "Display",           "", null,    "" );

        // put the Design Parameters into the Tree
        for (var ipmtr = 0; ipmtr < pmtr.length; ipmtr++) {
            var name  = "\u00a0\u00a0"+pmtr[ipmtr].name;
            var type  =                pmtr[ipmtr].type;
            var nrow  =                pmtr[ipmtr].nrow;
            var ncol  =                pmtr[ipmtr].ncol;
            var value =                pmtr[ipmtr].value[0];

            if (nrow > 1 || ncol > 1) {
                value = "["+nrow+"x"+ncol+"]";
            }

            if (type == 500 && pmtr[ipmtr].dot[0] != 0) {
                myTree.addNode(1, "^"+name, "", editPmtr, "*", ""+value, "", "");
            } else if (type == 500) {
                myTree.addNode(1,     name, "", editPmtr, "*", ""+value, "", "");
            }
        }

        wv.pmtrStat = -2;

        // open the Design Parameters (if they were open before the Tree was rebuilt)
        if (pmtr1Open == 1) {
            myTree.opened[1] = 1;
        }

        // put the Local Variables into the Tree
        for (var ipmtr = 0; ipmtr < pmtr.length; ipmtr++) {
            var name  = "\u00a0\u00a0"+pmtr[ipmtr].name;
            var type  =                pmtr[ipmtr].type;
            var nrow  =                pmtr[ipmtr].nrow;
            var ncol  =                pmtr[ipmtr].ncol;
            var value =                pmtr[ipmtr].value[0];

            if (nrow > 1 || ncol > 1) {
                value = "["+nrow+"x"+ncol+"]";
            }

            if (type !== 500) {
                myTree.addNode(2,     name, "", null,     "",  ""+value, "", "");
                var inode = myTree.name.length - 1;

                var indx = 0;
                if (nrow > 1 || ncol > 1) {
                    for (var irow = 0; irow < nrow; irow++) {
                        for (var icol = 0; icol < ncol; icol++) {
                            name  = "\u00a0\u00a0\u00a0\u00a0["+(irow+1)+","+(icol+1)+"]";
                            value = pmtr[ipmtr].value[indx++];

                            myTree.addNode(inode, name, "", null, "",
                                           "\u00a0\u00a0\u00a0\u00a0\u00a0\u00a0"+value, "", "");
                        }
                    }
                }


            }
        }

        wv.pmtrStat = -2;

        // open the Local Variables (if they were open before the Tree was rebuilt)
        if (pmtr2Open == 1) {
            myTree.opened[2] = 1;
        }

        // put the Branches into the Tree
        for (var ibrch = 0; ibrch < brch.length; ibrch++) {
            var name  = "\u00a0\u00a0"+brch[ibrch].name;
            var type  =      brch[ibrch].type;
            var actv;
            if (ibrch >= wv.builtTo) {
                actv = "skipped";
            } else if (brch[ibrch].actv == 301) {
                actv = "suppressed";
            } else if (brch[ibrch].actv == 302) {
                actv = "inactive";
            } else if (brch[ibrch].actv == 303) {
                actv = "deferred";
            } else {
                actv = "";
            }

            myTree.addNode(3, name, "", editBrch, "*", type, "", "", actv, "", "");

            // add the Branch's attributes to the Tree
            var inode = myTree.name.length - 1;
            for (var iattr = 0; iattr < brch[ibrch].attrs.length; iattr++) {
                var aname  = brch[ibrch].attrs[iattr][0];
                var avalue = brch[ibrch].attrs[iattr][1];
                myTree.addNode(inode, "\u00a0\u00a0\u00a0\u00a0"+aname, "", editAttr, "*", avalue, "", "");
            }
        }

        wv.brchStat = -2;

        // open the Branches (if they were open before the Tree was rebuilt)
        if (brchsOpen == 1) {
            myTree.opened[3] = 1;
        }

        // put the Display attributes into the Tree
        for (var gprim in wv.sceneGraph) {

            // parse the name
            var matches = gprim.split(" ");

            var ibody = Number(matches[1]);
            if        (matches[2] == "Face") {
                var iface = matches[3];
            } else if (matches[2] == "Edge") {
                var iedge = matches[3];
            } else {
                alert("unknown type: "+matches[2]);
                continue;
            }

            // determine if Body does not exists
            var kbody = -1;
            for (var jnode = 1; jnode < myTree.name.length; jnode++) {
                if (myTree.name[jnode] == "\u00a0\u00a0Body "+ibody) {
                    kbody = jnode;
                }
            }

            // if Body does not exist, create it and its Face and Edge
            //    subnodes now
            var kface, kedge;
            if (kbody < 0) {
                myTree.addNode(4, "\u00a0\u00a0Body "+ibody, "", null, "",
                               "Viz", "on", toggleViz, "Grd", "off", toggleGrd);
                kbody = myTree.name.length - 1;

                myTree.addNode(kbody, "\u00a0\u00a0\u00a0\u00a0Faces", "", null, "",
                               "Viz", "on", toggleViz, "Grd", "off", toggleGrd, "Trn", "off", toggleTrn);
                kface = myTree.name.length - 1;

                myTree.addNode(kbody, "\u00a0\u00a0\u00a0\u00a0Edges", "", null, "",
                               "Viz", "on", toggleViz, "Grd", "off", toggleGrd, "Ori", "off", toggleOri);
                kedge = myTree.name.length - 1;

            // otherwise, get pointers to the face-group and edge-group Nodes
            } else {
                kface = myTree.child[kbody];
                kedge = kface + 1;
            }

            // make the Tree Node
            if        (matches[2] == "Face") {
                myTree.addNode(kface, "\u00a0\u00a0\u00a0\u00a0\u00a0\u00a0face "+iface, gprim, null, "",
                               "Viz", "on", toggleViz, "Grd", "off", toggleGrd, "Trn", "off", toggleTrn);
            } else if (matches[2] == "Edge") {
                myTree.addNode(kedge, "\u00a0\u00a0\u00a0\u00a0\u00a0\u00a0edge "+iedge, gprim, null, "",
                               "Viz", "on", toggleViz, "Grd", "off", toggleGrd, "Ori", "off", toggleOri);
            }
        }

        // open the Display (by default)
        myTree.opened[4] = 1;

        // mark that we have (re-)built the Tree
        wv.sgUpdate = 0;

        // convert the abstract Tree Nodes into an HTML table
        myTree.build();
    }

    // deal with key presses
    if (wv.keyPress != -1) {

        var myKeyPress = String.fromCharCode(wv.keyPress);

        // '?' -- help
        if (myKeyPress == "?") {
            postMessage(".................. Viewer Cursor options ..................\n" +
                        "Home - initial view                                        \n" +
                        "Left - rotate or xlate left     Rite - rotate or xlate rite\n" +
                        "Up   - rotate or zlate up       Down - rotate or xlate down\n" +
                        "PgUp - zoom in                  PgDn - zoom out            \n" +
                        ">    - save view                <    - recall view         \n" +
                        "^    - query at cursor          @    - locate at cursor    \n" +
                        "!    - toggle flying mode       *    - center view         \n" +
                        "..................... ? - get help ........................");

        // '^' -- query at cursor
        } else if (myKeyPress == '^') {
            wv.picking  = 94;
            wv.pick     = 1;
            wv.sceneUpd = 1;

        // '@' -- locate at cursor
        } else if (myKeyPress == '@') {
            wv.locating = 64;
            wv.locate   = 1;

        // '*' -- center view
        } else if (myKeyPress == '*') {
            wv.centerV = 1;
            postMessage("View being centered");

        // '!' -- toggle flying mode
        } else if (myKeyPress == "!") {
            if (wv.flying <= 1) {
                postMessage("Turning flying mode ON");
                wv.flying = 10;
            } else {
                postMessage("Turning flying mode OFF");
                wv.flying = 1;
            }

        // '>' -- save view
        } else if (myKeyPress == ">") {
            postMessage("Saving current view");
            wv.saveMatrix.load(wv.mvMatrix);
            wv.sceneUpd = 1;

        // '<' -- recall view
        } else if (myKeyPress == "<") {
            postMessage("Restoring saved view");
            wv.mvMatrix.load(wv.saveMatrix);
            wv.sceneUpd = 1;

        // '<esc>' -- not used
        } else if (wv.keyPress == 0 && wv.keyCode == 27) {
            postMessage("<Esc> is not supported.  Use '?' for help.");

        // '<Home>' -- initial view
        } else if (wv.keyPress == 0 && wv.keyCode == 36) {
            wv.mvMatrix.makeIdentity();
            wv.sceneUpd = 1;

        // '<end>' -- not used
        } else if (wv.keyPress == 0 && wv.keyCode == 35) {
            postMessage("<End> is not supported.  Use '?' for help.");

        // '<PgUp>' -- zoom in
        } else if (wv.keyPress == 0 && wv.keyCode == 33) {
            if (wv.modifier == 0) {
                wv.mvMatrix.scale(2.0, 2.0, 2.0);
            } else {
                wv.mvMatrix.scale(1.25, 1.25, 1.25);
            }
            wv.sceneUpd = 1;

        // '<PgDn>' -- zoom out
        } else if (wv.keyPress == 0 && wv.keyCode == 34) {
            if (wv.modifier == 0) {
                wv.mvMatrix.scale(0.5, 0.5, 0.5);
            } else {
                wv.mvMatrix.scale(0.8, 0.8, 0.8);
            }
            wv.sceneUpd = 1;

        // '<Delete>' -- not used
        } else if (wv.keyPress == 0 && wv.keyCode == 46) {
            postMessage("<Delete> is not supported.  Use '?' for help.");

        // '<Left>' -- rotate or translate object left
        } else if (wv.keyPress == 0 && wv.keyCode == 37) {
            if (wv.flying == 1) {
                if (wv.modifier == 0) {
                    wv.mvMatrix.rotate(-30, 0,1,0);
                } else {
                    wv.mvMatrix.rotate( -5, 0,1,0);
                }
            } else {
                if (wv.modifier == 0) {
                    wv.mvMatrix.translate(-0.5, 0.0, 0.0);
                } else {
                    wv.mvMatrix.translate(-0.1, 0.0, 0.0);
                }
            }
            wv.sceneUpd = 1;

        // '<Right>' -- rotate or translate object right
        } else if (wv.keyPress == 0 && wv.keyCode == 39) {
            if (wv.flying == 1) {
                if (wv.modifier == 0) {
                    wv.mvMatrix.rotate(+30, 0,1,0);
                } else {
                    wv.mvMatrix.rotate( +5, 0,1,0);
                }
            } else {
                if (wv.modifier == 0) {
                    wv.mvMatrix.translate(+0.5, 0.0, 0.0);
                } else {
                    wv.mvMatrix.translate(+0.1, 0.0, 0.0);
                }
            }
            wv.sceneUpd = 1;

        // '<Up>' -- rotate or translate object up
        } else if (wv.keyPress == 0 && wv.keyCode == 38) {
            if (wv.flying == 1) {
                if (wv.modifier == 0) {
                    wv.mvMatrix.rotate(-30, 1,0,0);
                } else {
                    wv.mvMatrix.rotate( -5, 1,0,0);
                }
            } else {
                if (wv.modifier == 0) {
                    wv.mvMatrix.translate(0.0, +0.5, 0.0);
                } else {
                    wv.mvMatrix.translate(0.0, +0.1, 0.0);
                }
            }
            wv.sceneUpd = 1;

        // '<Down>' -- rotate or translate object down
        } else if (wv.keyPress == 0 && wv.keyCode == 40) {
            if (wv.flying == 1) {
                if (wv.modifier == 0) {
                    wv.mvMatrix.rotate(+30, 1,0,0);
                } else {
                    wv.mvMatrix.rotate( +5, 1,0,0);
                }
            } else {
                if (wv.modifier == 0) {
                    wv.mvMatrix.translate(0.0, -0.5, 0.0);
                } else {
                    wv.mvMatrix.translate(0.0, -0.1, 0.0);
                }
            }
            wv.sceneUpd = 1;

        // unknown command
        } else {
            postMessage("Unknown command (keyPress="+myKeyPress+", keyCode="+
                        wv.keyCode+").  Use '?' for help.");
        }
    }

    wv.keyPress = -1;

    // UI is in screen coordinates (not object)
    wv.uiMatrix.load(wv.mvMatrix);
    wv.mvMatrix.makeIdentity();

    // deal with mouse movement
    if (wv.dragging) {

        // cntrl is down (rotate)
        if (wv.modifier == 4) {
            var angleX =  (wv.startY - wv.cursorY) / 4.0 / wv.flying;
            var angleY = -(wv.startX - wv.cursorX) / 4.0 / wv.flying;
            if ((angleX != 0.0) || (angleY != 0.0)) {
                wv.mvMatrix.rotate(angleX, 1,0,0);
                wv.mvMatrix.rotate(angleY, 0,1,0);
                wv.sceneUpd = 1;
            }

        // alt-shift is down (rotate)
        } else if (wv.modifier == 3) {
            var angleX =  (wv.startY - wv.cursorY) / 4.0 / wv.flying;
            var angleY = -(wv.startX - wv.cursorX) / 4.0 / wv.flying;
            if ((angleX != 0.0) || (angleY != 0.0)) {
                wv.mvMatrix.rotate(angleX, 1,0,0);
                wv.mvMatrix.rotate(angleY, 0,1,0);
                wv.sceneUpd = 1;
            }

        // alt is down (spin)
        } else if (wv.modifier == 2) {
            var xf = wv.startX - wv.width  / 2;
            var yf = wv.startY - wv.height / 2;

            if ((xf != 0.0) || (yf != 0.0)) {
                var theta1 = Math.atan2(yf, xf);
                xf = wv.cursorX - wv.width  / 2;
                yf = wv.cursorY - wv.height / 2;

                if ((xf != 0.0) || (yf != 0.0)) {
                    var dtheta = Math.atan2(yf, xf) - theta1;
                    if (Math.abs(dtheta) < 1.5708) {
                        var angleZ = 128*(dtheta) / 3.1415926 / wv.flying;
                        wv.mvMatrix.rotate(angleZ, 0,0,1);
                        wv.sceneUpd = 1;
                    }
                }
            }

        // shift is down (zoom)
        } else if (wv.modifier == 1) {
            if (wv.cursorY != wv.startY) {
                var scale = Math.exp((wv.cursorY - wv.startY) / 512.0 / wv.flying);
                wv.mvMatrix.scale(scale, scale, scale);
                wv.scale   *= scale;
                wv.sceneUpd = 1;
            }

        // no modifier (translate)
        } else {
            var transX = (wv.cursorX - wv.startX) / 256.0 / wv.flying;
            var transY = (wv.cursorY - wv.startY) / 256.0 / wv.flying;
            if ((transX != 0.0) || (transY != 0.0)) {
                wv.mvMatrix.translate(transX, transY, 0.0);
                wv.sceneUpd = 1;
            }
        }

        // if not flying, then update the start coordinates
        if (wv.flying <= 1) {
            wv.startX = wv.cursorX;
            wv.startY = wv.cursorY;
        }
    }
}


//
// callback when a (text) message is received from the server (called by wv-socket.js)
//
function wvServerMessage(text)
{
    // alert("wvServerMessage(text="+text+")");

    // if it is a null message, do nothing
    if (text.length <= 1) {

    // if it starts with "identify;" post a message */
    } else if (text.substring(0,9) == "identify;") {
        if (wv.server === undefined) {
            wv.server = text.substring(9,text.length-2);
            postMessage("ESP has been initialized and is attached to '"+wv.server+"'");
        }

    // if it starts with "sgData;" store the auxiliary scene graph data
    } else if (text.substring(0,7) == "sgData;") {
        if (text.length > 9) {
            sgData = JSON.parse(text.substring(7,text.length-1));
        } else {
            sgData = {};
        }

    // if it starts with "sgFocus;" store the scene graph focus data
    } else if (text.substring(0,8) == "sgFocus;") {
        if (text.length > 10) {
            wv.focus = JSON.parse(text.substring(8,text.length-1));
        } else {
            wv.focus = [];
        }

    // if it starts with "getPmtrs;" build the (global) pmtr array
    } else if (text.substring(0,9) == "getPmtrs;") {
        if (text.length > 11) {
            pmtr = JSON.parse(text.substring(9,text.length-1));
        } else {
            pmtr = new Array;
        }
        wv.pmtrStat = -1;

    // if it starts with "newPmtr;" do nothing
    } else if (text.substring(0,8) == "newPmtr;") {

    // if it starts with "clrVels;" do nothing
    } else if (text.substring(0,8) == "clrVels;") {

    // if it starts with "setPmtr;" do nothing
    } else if (text.substring(0,8) == "setPmtr;") {

    // if it starts with "setVel;" do nothing
    } else if (text.substring(0,7) == "setVel;") {

    // if it starts with "getBrchs;" build the (global) brch array
    } else if (text.substring(0,9) == "getBrchs;") {
        if (text.length > 11) {
            brch = JSON.parse(text.substring(9,text.length-1));
        } else {
            brch = new Array;
        }
        wv.brchStat = -1;

    // if it starts with "newBrch;" do nothing
    } else if (text.substring(0,8) == "newBrch;") {

    // if it starts with "setBrch;" do nothing
    } else if (text.substring(0,8) == "setBrch;") {

    // if it starts with "delBrch;" do nothing
    } else if (text.substring(0,8) == "delBrch;") {

    // if it starts with "setAttr;" do nothing
    } else if (text.substring(0,8) == "setAttr;") {

    // if it starts with "undo;" get the latest Parameters and
    //                           Branches and update tree (scene)
    } else if (text.substring(0,5) == "undo;") {
        var cmd = text.substring(5,text.length-2);

        wv.sgUpdate = 1;
        wv.brchStat = 0;
        wv.pmtrStat = 0;

        postMessage("Undoing \""+cmd+"\" ====> Re-build is needed <====");

    // if it starts with "save;" do nothing
    } else if (text.substring(0,5) == "save;") {

    // if it starts with "build;" reset the build button
    } else if (text.substring(0,6) == "build;") {
        var button = document.getElementById("buildButton");
        button.onclick = null;
        button["innerHTML"] = "Up to date";
        button.style.backgroundColor = null;

        var textList = text.split(";");
        var ibrch    = textList[1];
        var nbody    = textList[2];

        wv.builtTo = ibrch;

        if (ibrch == brch.length) {
            postMessage("Entire build complete, which generated "+nbody+
                        " Body(s)");
        } else {
            postMessage("Partial build (through "+brch[ibrch-1].name+
                        ") complete, which generated "+nbody+" Body(s)");
        }

        changeMode(0);

    // if it starts with "setLims;" do nothing
    } else if (text.substring(0,8) == "setLims;") {

    // if it starts with "getCsmFile;" store the results in wv.filename
    //    and wv.curFile
    } else if (text.substring(0,11) == "getCsmFile;") {
        var ichar = 11;

        wv.filename = "";
        while (text[ichar] != ";") {
            wv.filename += text[ichar++];
        }
        ichar++;

        wv.curFile = text.substring(ichar);

    // if it strart with "setCsmFileBeg;" do nothing
    } else if (text.substring(0,14) == "setCsmFileBeg;") {

    // if it strart with "setCsmFileMid;" do nothing
    } else if (text.substring(0,14) == "setCsmFileMid;") {

    // if it strart with "setCsmFileEnd;" do nothing
    } else if (text.substring(0,14) == "setCsmFileEnd;") {

    // if it starts with "ERROR:: there is nothing to undo" post the error
    } else if (text.substring(0,32) == "ERROR:: there is nothing to undo") {
        postMessage(text);
        alert("There is nothing to undo");

    // if it starts with "ERROR::" post the error
    } else if (text.substring(0,7) == "ERROR::") {
        postMessage(text);
        alert("Error encountered (see messages)");

        var button = document.getElementById("buildButton");

        button["innerHTML"] = "Fix before re-build";
        button.style.backgroundColor = "#FF3F3F";

        changeMode(0);

        button.onclick = function () {
        };

    // default is to post the message
    } else {
        postMessage(text);
    }
}


//
// activate the buildButton
//
function activateBuildButton () {
    // alert("in activateBuildButton()");

    var button = document.getElementById("buildButton");

    button["innerHTML"] = "Press to Re-build";
    button.style.backgroundColor = "#3FFF3F";

    button.onclick = function () {
        if (wv.curMode != 0) {
            alert("Command disabled,  Press 'Cancel' or 'OK' first.");
            return;
        } else {
            postMessage("Re-building...");
        }

        // build first so that parameters are updated
        wv.socketUt.send("build;0;");
        wv.brchStat = 6000;

        wv.socketUt.send("getPmtrs;");
        wv.pmtrStat = 6000;

        wv.socketUt.send("getBrchs;");

        var button = document.getElementById("buildButton");
        button.onclick = null;
        button["innerHTML"] = "Re-building...";
        button.style.backgroundColor = "#FFFF3F";

        //inactivate buttons until build is done
        changeMode(-1);
    };
}


//
// callback when "undoButton" is pressed (called by ESP.html)
//
function cmdUndo () {
    // alert("in cmdUndo()");

    if (wv.curMode != 0) {
        alert("Command disabled.  Press 'Cancel' or 'OK' first.");
        return;
    } else if (wv.server != "serveCSM") {
        alert("cmdUndo is not implemented for "+wv.server);
        return;
    }

    wv.socketUt.send("undo;");

    // update the UI
    activateBuildButton();
}


//
// callback when "editButton" is pressed (called by ESP.html)
//
function cmdEdit () {
    // alert("in cmdEdit()");

    if (wv.curMode != 0) {
        alert("Command disabled.  Press 'Cancel' or 'OK' first.");
        return;
    } else if (wv.server != "serveCSM") {
        alert("cmdEdit is not implemented for "+wv.server);
        return;
    }

    // if there have been any changes, tell user to save first
    if (wv.nchanges > 0) {
        alert("Changes have been made to the \"Design Parameters\"\n" +
              "and/or the \"Branches\" via the user-interface.\n\n" +
              "Save these changes before you can edit the .csm file");
        return;
    }

    // fill in the name of the .csm file
    var csmFilename = document.getElementById("editCsmFilename");

    csmFilename["innerHTML"] = "Contents of: "+wv.filename;

    // fill the textarea with the current .csm file
    var csmTextArea = document.getElementById("editCsmTextArea");

    csmTextArea.cols  = 84;
    csmTextArea.rows  = 25;
    csmTextArea.value = wv.curFile;

    // post the editCsmForm
    changeMode(6);
}


//
// callback when "saveButton" is pressed (called by ESP.html)
//
function cmdSave () {
    // alert("in cmdSave()");

    if (wv.curMode != 0) {
        alert("Command disabled.  Press 'Cancel' or 'OK' first.");
        return;
    } else if (wv.server != "serveCSM") {
        alert("cmdSave is not implemented for "+wv.server);
        return;
    }

    var filename = prompt("Enter filename:", wv.filename);
    if (filename !== null) {
        if (filename.search(".csm") == -1) {
            filename = filename + ".csm";
        }

        postMessage("Saving model to '"+filename+"'");
        wv.socketUt.send("save;"+filename+";");

        wv.socketUt.send("getFile;");

//        wv.filename = filename;
        wv.nchanges = -1;
    } else {
        postMessage("NOT saving since no filename specified");
    }
}


//
// callback when "testButton" is pressed (called by ESP.html)
//
function cmdTest () {
    // alert("in cmdTest()");

}


//
// callback when "helpButton" is pressed (called by ESP.html)
//
function cmdHelp () {

    // open help in another tab
    window.open("ESP-help.html");
}


//
// callback when "Design Parameters" is pressed in Tree
//
function addPmtr() {
    // alert("addPmtr()");

    if (wv.curMode != 0) {
        alert("Command disabled.  Press 'Cancel' or 'OK' first.");
        return;
    } else if (wv.server != "serveCSM") {
        alert("addPmtr is not implemented for "+wv.server);
        return;
    }

    // get the new name
    var name = prompt("Enter new Parameter name");
    if (name === null) {
        return;
    } else if (name.length <= 0) {
        return;
    }

    // check that name is valid
    if (name.match(/^[a-zA-Z]\w*$/) === null) {
        alert("'"+name+"' is not a valid name");
        return;
    }

    // check that the name does not exist already
    for (var ipmtr = 0; ipmtr < pmtr.length; ipmtr++) {
        if (name == pmtr[ipmtr].name) {
            alert("'"+name+"' already exists");
            return;
        }
    }

    // get the number of rows
    var nrow = prompt("Enter number of rows", "1");
    if (nrow !== null) {
        if (isNaN(nrow)) {
            alert("Number of rows must be positive integer");
            return;
        } else if (Number(nrow) < 1) {
            alert("Number of rows must be positive integer");
            return;
        } else if (Math.round(Number(nrow)) != Number(nrow)) {
            alert("Number of rows must be positive integer");
            return;
        }
    } else {
        return;
    }

    // get the number of columns
    var ncol = prompt("Enter number of columns", "1");
    if (ncol !== null) {
        if (isNaN(ncol)) {
            alert("Number of columns must be positive integer");
            return;
        } else if (Number(ncol) < 1) {
            alert("Number of columns must be positive integer");
            return;
        } else if (Math.round(Number(ncol)) != Number(ncol)) {
            alert("Number of columns must be positive integer");
            return;
        }
    } else {
        return;
    }

    // store the values locally
    var newPmtr = pmtr.length;

    pmtr[newPmtr] = new Array();

    pmtr[newPmtr].name = name;
    pmtr[newPmtr].type = 500;
    pmtr[newPmtr].nrow = nrow;
    pmtr[newPmtr].ncol = ncol;
    pmtr[newPmtr].value = new Array(nrow*ncol);

    for (var i = 0; i < nrow*ncol; i++) {
        pmtr[newPmtr].value[i] = "";
    }

    // remember info for Parameter
    wv.curPmtr = newPmtr;

    // set up editPmtrForm
    if (setupEditPmtrForm() > 0) {
        return;
    }

    // post the editPmtr form (with the addPmtr header)
    changeMode(4)
}


//
// callback when "Clear Design Velocities" is pressed in editPmtrForm (called by ESP.html)
//
function clrVels() {
    // alert("clrVels()");

    // disable this command if there were any changes to the Parameter
    if (numberOfPmtrChanges() > 0) {
        alert("Changes were made.  Press 'Cancel' or 'OK' first.");
        return;
    } else if (wv.server != "serveCSM") {
        alert("clrVels is not implemented for "+wv.server);
        return;
    }

    // get an updated Parameter list (so that added Parameter is listed)
    wv.socketUt.send("clrVels;");

    // update the UI
    postMessage("Design Velocities have been cleared ====> Re-build is needed <====");
    activateBuildButton();

    changeMode(0);
}


//
// callback when Design Parameter name is pressed in Tree
//
function editPmtr(e) {
    // alert("in editPmtr(e="+e+")");

    if (wv.curMode != 0) {
        alert("Command disabled.  Press 'Cancel' or 'OK' first.");
        return;
    } else if (wv.server != "serveCSM") {
        alert("editPmtr is not implemented for "+wv.server);
        return;
    }

    wv.menuEvent = e;

    // get the Tree Node
    var id    = wv.menuEvent["target"].id;
    var inode = Number(id.substring(4,id.length-4));

    // get the Parameter name
    var name = myTree.name[inode].replace(/\u00a0/g, "");
    name = name.replace(/\^/g, "");

    // get the Parameter index
    var ipmtr = -1;      // 0-bias
    var jpmtr;           // 1-bias (and temp)
    for (jpmtr = 0; jpmtr < pmtr.length; jpmtr++) {
        if (pmtr[jpmtr].name.replace(/\^/g, "") == name) {
            ipmtr = jpmtr;
            break;
        }
    }
    if (ipmtr < 0) {
        alert(name+" not found");
        return;
    } else {
        jpmtr = ipmtr + 1;
    }

    // remember info for te current Parameter
    wv.curPmtr = ipmtr;

    // set up editPmtrForm
    setupEditPmtrForm();

    // post the editPmtr form (with the editPmtr header)
    changeMode(5);
}


//
// callback when "OK" is pressed in editPmtrForm (called by ESP.html)
//
function editPmtrOk() {
    // alert("in editPmtrOk()");

    var editPmtrForm = document.getElementById("editPmtrForm");

    var ipmtr = wv.curPmtr;
    var name  = pmtr[ipmtr].name;
    var nrow  = pmtr[ipmtr].nrow;
    var ncol  = pmtr[ipmtr].ncol;
    var irow;
    var icol;

    // make sure that all entries have valid values
    var nchange = 0;
    var index   = -1;
    for (irow = 1; irow <= pmtr[ipmtr].nrow; irow++) {
        for (icol = 1; icol <= pmtr[ipmtr].ncol; icol++) {
            index++;

            // get the new value
            var myInput = editPmtrForm["row"+irow+"col"+icol];
            var value   = myInput.value.replace(/\s/g, "");

            if (value.length <= 0) {
                alert("Entry at (row "+irow+", col "+icol+") is blank");
                return;
            } else if (isNaN(value)) {
                alert("Illegal number format at (row "+irow+", col "+icol+")");
                return;
            }
        }
    }

    // send the new Parameter to the server if in add Brch mode
    if (wv.curMode == 4) {
        var mesg = "newPmtr;"+name+";"+nrow+";"+ncol+";";

        var index = -1;
        for (irow = 1; irow <= nrow; irow++) {
            for (icol = 1; icol <= ncol; icol++) {
                index++;
                mesg = mesg+";";
            }
        }

        wv.socketUt.send(mesg);
    }

    // get each of the values
    index = -1;
    for (irow = 1; irow <= pmtr[ipmtr].nrow; irow++) {
        for (icol = 1; icol <= pmtr[ipmtr].ncol; icol++) {
            index++;

            // get the new value
            var myInput = editPmtrForm["row"+irow+"col"+icol];
            var value = myInput.value.replace(/\s/g, "");

            if (value != pmtr[ipmtr].value[index]) {
                postMessage("Parameter '"+pmtr[ipmtr].name+"["+irow+","+icol+
                            "]' has been changed to "+value+
                            " ====> Re-build is needed <====");
                nchange++;

                // store the value locally
                pmtr[ipmtr].value[index] = Number(value);

                // send the new value to the server
                wv.socketUt.send("setPmtr;"+pmtr[ipmtr].name+";"+irow+";"+icol+";"+value+";");
            }
        }
    }

    // update the UI
    if (nchange > 0) {
        wv.nchanges += nchange;

        if (wv.curMode != 4) {
            var id     = wv.menuEvent["target"].id;
            var myElem = document.getElementById(id);
            myElem.className = "fakelinkoff";

        // get an updated Parameter list (so that added Pmtr is listed)
        } else {
            wv.socketUt.send("getPmtrs;");
        }

        activateBuildButton();
    }

    // return to the WebViewer
    changeMode(0);
}


//
// callback when "Cancel" is pressed in editPmtrForm (called by ESP.html)
//
function editPmtrCancel() {
    // alert("in editPmtrCancel()");

    // if we are in process of adding a Parameter, remove it now
    if (wv.curMode == 4) {
        pmtr.splice(pmtr.length-1, 1);
    }

    // return to the WebViewer
    changeMode(0);
}


//
// callback when "Compute Sensitivity" is pressed in editPmtrForm (called by ESP.html)
//
function compSens() {
    // alert("in compSens()");

    // disable this command if there were any changes to the Parameter
    if (numberOfPmtrChanges() > 0) {
        alert("Changes were made.  Press 'Cancel' or 'OK' first.");
        return;
    }

    // get the Tree Node
    var id    = wv.menuEvent["target"].id;
    var inode = Number(id.substring(4,id.length-4));

    // get the Parameter name
    var name = myTree.name[inode].replace(/\u00a0/g, "");
    name = name.replace(/\^/g, "");

    // get the Parameter index
    var ipmtr = -1;      // 0-bias
    var jpmtr;           // 1-bias (and temp)
    for (jpmtr = 0; jpmtr < pmtr.length; jpmtr++) {
        if (pmtr[jpmtr].name.replace(/\^/g, "") == name) {
            ipmtr = jpmtr;
            break;
        }
    }
    if (ipmtr < 0) {
        alert(name+" not found");
        return;
    } else {
        jpmtr = ipmtr + 1;
    }

    // can only compute sensitivity for a scalar
    if (pmtr[ipmtr].nrow > 1 || pmtr[ipmtr].ncol > 1) {
        alert("Sensitivity can only be computed for a scalar");
        return;
    }

    // clear any previous velocities
    wv.socketUt.send("clrVels;");
    for (jpmtr = 0; jpmtr < pmtr.length; jpmtr++) {
        pmtr[jpmtr].dot[0] = 0;
    }

    // set velocity for ipmtr
    pmtr[ipmtr].dot[0] = 1.0;
    wv.socketUt.send("setVel;"+pmtr[ipmtr].name+";1;1;1;")
    postMessage("Computing sensitivity with respect to "+pmtr[ipmtr].name);

    // rebuild
    wv.socketUt.send("build;0;");
    wv.brchStat = 6000;

    wv.socketUt.send("getPmtrs;");
    wv.pmtrStat = 6000;

    wv.socketUt.send("getBrchs;");

    var button = document.getElementById("buildButton");
    button.onclick = null;
    button["innerHTML"] = "Re-building...";
    button.style.backgroundColor = "#FFFF3F";

    // inactivate buttons until build is done
    changeMode( 0);
    changeMode(-1);
}


//
// callback when "Set Design Velocity" is pressed in editPmtrForm (called by ESP.html)
//
function setVel() {
    // alert("in setVel()");

    // disable this command if there were any changes to the Parameter
    if (numberOfPmtrChanges() > 0) {
        alert("Changes were made.  Press 'Cancel' or 'OK' first.");
        return;
    }

    // get the Tree Node
    var id    = wv.menuEvent["target"].id;
    var inode = Number(id.substring(4,id.length-4));

    // get the Parameter name
    var name = myTree.name[inode].replace(/\u00a0/g, "");
    name = name.replace(/\^/g, "");

    // get the Parameter index
    var ipmtr = -1;      // 0-bias
    var jpmtr;           // 1-bias (and temp)
    for (jpmtr = 0; jpmtr < pmtr.length; jpmtr++) {
        if (pmtr[jpmtr].name.replace(/\^/g, "") == name) {
            ipmtr = jpmtr;
            break;
        }
    }
    if (ipmtr < 0) {
        alert(name+" not found");
        return;
    } else {
        jpmtr = ipmtr + 1;
    }

    // get each of the values
    var index  = -1;
    var nchange = 0;
    for (var irow = 1; irow <= pmtr[ipmtr].nrow; irow++) {
        for (var icol = 1; icol <= pmtr[ipmtr].ncol; icol++) {
            index++;

            // get the new value
            var newVel;
            if (pmtr[ipmtr].nrow == 1 && pmtr[ipmtr].ncol == 1) {
                newVel = prompt("Enter new Design Velocity for "+name,
                                pmtr[ipmtr].dot[index]);
            } else {
                newVel = prompt("Enter new Design Velocity for "+name+
                                "["+irow+","+icol+"]",
                                pmtr[ipmtr].dot[index]);
            }

            // make sure a valid number was entered
            if (newVel === null) {
                continue;
            } else if (isNaN(newVel)) {
                alert("Illegal number format, so Design Velocity not being changed");
                continue;
            } else if (newVel == pmtr[ipmtr].dot[index]) {
                continue;
            } else if (pmtr[ipmtr].nrow == 1 && pmtr[ipmtr].ncol == 1) {
                postMessage("Parameter '"+name+"' has new Design Velocity "+
                            newVel+" ====> Re-build is needed <====");
                nchange++;
            } else {
                postMessage("Parameter '"+name+"["+irow+","+icol+
                            "]' has new Design Velocity "+newVel+
                            " ====> Re-build is needed <====");
                nchange++;
            }

            // store the value locally
            pmtr[ipmtr].dot[index] = Number(newVel);

            // send the new Design Velocity to the server
            wv.socketUt.send("setVel;"+pmtr[ipmtr].name+";"+irow+";"+icol+";"+newVel+";");

        }
    }

    // update the UI
    if (nchange > 0) {
        wv.nchanges += nchange;

        var myElem = document.getElementById(id);
        myElem.className = "fakelinkoff";

        activateBuildButton();
    }

    // return to the WebViewer
    changeMode(0);
}


//
// callback when "Branch" is pressed in Tree
// callback when "Add new Branch after this Brnch" is pressed in editBrchForm (called by ESP.html)
//
function addBrch() {
    // alert("in addBrch()");

    // this check allows one to select AddBranchAfterThisBranch
    if (wv.menuEvent !== undefined && wv.curMode == 3) {

    } else if (wv.curMode != 0) {
        alert("Changes were made.  Press 'Cancel' or 'OK' first.");
        return;
    } else if (wv.server != "serveCSM") {
        alert("addBrch is not implemented for "+wv.server);
        return;
    }

    // remember
    if (wv.curBrch >= 0) {
        wv.afterBrch = wv.curBrch;
        wv.curBrch   = -1;
    } else {
        wv.afterBrch = brch.length - 1;
        wv.curBrch   = -1;
    }

    // post the addBrch form
    changeMode(1);
}


//
// callback when "OK" is pressed in addBrchForm (called by ESP.html)
//
function addBrchOk() {
    // alert("in addBrchOk()");

    // get the elements on the form
    var elements = document.getElementById("addBrchForm").elements;

    var jelem       = -1;
    for (var ielem = 0; ielem < elements.length; ielem++) {
        if (elements[ielem].checked) {
            jelem = ielem;
            break;
        }
    }

    // if nothing is selected, return to the WebViewer
    if (jelem < 0) {
        alert("Select a Branch type or press 'Cancel'.")
        return;
    }

    // if something was picked, initialize the new Branch
    var newBrch = brch.length;

    brch[newBrch] = new Array();

    brch[newBrch].name = "**name_automatically_assigned**";
    brch[newBrch].type = elements[jelem].value;
    brch[newBrch].actv = 300;
    brch[newBrch].args = new Array();

    // remember info for Branch to add after
    wv.curBrch = newBrch;

    // set up editBrchForm
    if (setupEditBrchForm() > 0) {
        return;
    }

    // increment the number of changes
    wv.nchanges++;

    // post the editBrch form (with the addBrch header)
    changeMode(2);
}


//
// callback when "Cancel" is pressed in addBrchForm (called by ESP.html)
//
function addBrchCancel() {
    // alert("in addBrchCancel()");

    // return to the WebViewer
    changeMode(0);
}


//
// callback when "Delete this Branch" is pressed in editBrchForm (called by ESP.html)
//
function delBrch() {
    // alert("in delBrch()");

    // disable this command if there were any chnges to the Branch
    if (numberOfBrchChanges() > 0) {
        alert("Changes were made.  Press 'Cancel' or 'OK' first.");
        return;
    }

    var ibrch = -1;
    var name  = "";

    // get the Tree node
    var id    = wv.menuEvent["target"].id;
    var inode = Number(id.substring(4,id.length-4));

    // if this was called by pressing "Delete last Branch", adjust inode
    //    to point to last visible node (as if "Delete this Branch"
    //    was pressed on the last Branch)
    if (inode == 2) {
        inode = myTree.child[inode];

        while (myTree.next[inode] >= 0) {
            inode = myTree.next[inode];
        }
    }

    // get the Branch name
    name = myTree.name[inode].replace(/\u00a0/g, "");

    // get the Branch index
    for (var jbrch = 0; jbrch < brch.length; jbrch++) {
        if (brch[jbrch].name == name) {
            ibrch = jbrch + 1;
        }
    }
    if (ibrch <= 0) {
        alert(name+" not found");
        return;
    }

    // send the message to the server
    wv.socketUt.send("delBrch;"+ibrch+";");

    // hide the Tree node that was just updated
    var element = myTree.document.getElementById("node"+inode);
    element.style.display = "none";

    // get an updated version of the Branches
    wv.brchStat = 0;

    // update the UI
    postMessage("Deleting Branch "+name+" ====> Re-build is needed <====");
    activateBuildButton();

    // return to t he WebViewer
    changeMode(0);
}


//
// callback when Branch name is pressed in Tree
//
function editBrch(e) {
    // alert("in editBrch(e="+e+")");

    if (wv.curMode != 0) {
        alert("Command disabled.  Press 'Cancel' or 'OK' first.");
        return;
    } else if (wv.server != "serveCSM") {
        alert("editBrch is not implemented for "+wv.server);
        return;
    }

    wv.menuEvent = e;

    // get the Tree node
    var id    = wv.menuEvent["target"].id;
    var inode = Number(id.substring(4,id.length-4));

    // get the Branch name
    var name = myTree.name[inode].replace(/\u00a0/g, "");

    // get the Branch index
    var ibrch = -1;           // 0-bias
    var jbrch;                // 1-bias (and temp)
    for (jbrch = 0; jbrch < brch.length; jbrch++) {
        if (brch[jbrch].name == name) {
            ibrch = jbrch;
            break;
        }
    }
    if (ibrch < 0) {
        alert(name+" not found");
        return;
    } else {
        jbrch = ibrch + 1;
    }

    // remember info for the current Branch
    wv.curBrch   = ibrch;

    // set up editBrchForm
    setupEditBrchForm();

    // post the editBrch form (with the editBrch header)
    changeMode(3);
}


//
// callback when "OK" button is pressed in editBrchForm (called by ESP.html)
//
function editBrchOk() {
    // alert("in editBrchOk()");

    var editBrchForm = document.getElementById("editBrchForm");

    var ibrch = wv.curBrch;
    var jbrch = ibrch + 1;

    var brchName = editBrchForm.brchName.value.replace(/\s/g, "");
    if (brchName.length <= 0) {
        alert("Name cannot be blank");
        return;
    }

    var mesg;
    var output;
    var nchange = 0;

    if (wv.curMode == 2) {
        mesg = "newBrch;"+(wv.afterBrch+1)+";"+brch[ibrch].type+";";
    } else {
        mesg = "setBrch;"+jbrch+";";
    }

    // make sure that name does not start with "Brch\u00a0"
    var newBrchName = editBrchForm.brchName.value;
    if (newBrchName.substring(0,5) == "Brch\u00a0") {
        alert("Changed name '"+newBrchName+
              "' cannot begin with 'Brch\u00a0'");
        return;
    }

    if (newBrchName != brch[ibrch].name) {
        // make sure that name does not already exist
        for (var kbrch = 0; kbrch < brch.length; kbrch++) {
            if (brch[kbrch].name == newBrchName) {
                alert("Name '"+newBrchName+"' already exists");
                return;
            }
        }

        // store the value locally
        postMessage("Changing name '"+brch[ibrch].name+"' to '"+
                    newBrchName+"' ====> Re-build is needed <====");
        brch[ibrch].name = newBrchName;

        nchange++;
    }

    if (wv.curMode == 3) {
        mesg = mesg + newBrchName + ";";
    }

    // update the activity
    var newActivity = editBrchForm.activity.value;
    if (brch[ibrch].actv != "none"     &&
        brch[ibrch].actv != newActivity  ) {

        // store the value locally
        brch[ibrch].actv = newActivity;
        if (brch[ibrch].actv == 300) {
            postMessage("Activating Branch "+newBrchName+
                        " ====> Re-build is needed <====");
        } else {
            postMessage("Suppressing Branch "+newBrchName+
                        " ====> Re-build is needed <====");
        }

        // get an updated version of the Branches
        wv.brchStat = 0;

        nchange++;
    }

    if (wv.curMode == 2) {

    } else if (newActivity == 300) {
        mesg = mesg + "active;";
    } else {
        mesg = mesg + "suppressed;"
    }

    // update any necessary arguments
    for (var iarg = 0; iarg < wv.numArgs; iarg++) {
        if        (iarg == 0) {
            var name  = document.getElementById("argName1").firstChild["data"];
            var value = editBrchForm.            argValu1.value.replace(/\s/g, "");
        } else if (iarg == 1) {
            var name  = document.getElementById("argName2").firstChild["data"];
            var value = editBrchForm.            argValu2.value.replace(/\s/g, "");
        } else if (iarg == 2) {
            var name  = document.getElementById("argName3").firstChild["data"];
            var value = editBrchForm.            argValu3.value.replace(/\s/g, "");
        } else if (iarg == 3) {
            var name  = document.getElementById("argName4").firstChild["data"];
            var value = editBrchForm.            argValu4.value.replace(/\s/g, "");
        } else if (iarg == 4) {
            var name  = document.getElementById("argName5").firstChild["data"];
            var value = editBrchForm.            argValu5.value.replace(/\s/g, "");
        } else if (iarg == 5) {
            var name  = document.getElementById("argName6").firstChild["data"];
            var value = editBrchForm.            argValu6.value.replace(/\s/g, "");
        } else if (iarg == 6) {
            var name  = document.getElementById("argName7").firstChild["data"];
            var value = editBrchForm.            argValu7.value.replace(/\s/g, "");
        } else if (iarg == 7) {
            var name  = document.getElementById("argName8").firstChild["data"];
            var value = editBrchForm.            argValu8.value.replace(/\s/g, "");
        } else if (iarg == 8) {
            var name  = document.getElementById("argName9").firstChild["data"];
            var value = editBrchForm.            argValu9.value.replace(/\s/g, "");
        }

        // check for blanks
        if (value.length <= 0) {
            if (brch[ibrch].type == "fillet" ||
                brch[ibrch].type == "chamfer"  ) {
                if (confirm(name+" is blank.  Continue?")) {
                } else {
                    return;
                }
            } else {
                alert(name+" should not be blank.");
                return;
            }
        }

        if (name.charAt(0) != "$" || value.length <= 0) {
            output =       value;
        } else {
            output = "$" + value;
        }

        // check for semi-colons
        if (value.indexOf(";") >= 0) {
            alert(name+" should not contain a semi-colon.");
            return;
        }

        // check if  there were any changes
        if (output != brch[ibrch].args[iarg]) {
            if (wv.curMode == 3) {
                postMessage("Changing "+name+" from '"+brch[ibrch].args[iarg]+
                            "' to '"+output+
                            "' ====> Re-build is needed <====");
            }

            brch[ibrch].args[iarg] = output;
            nchange++;
        }
        mesg = mesg + output + ";";
    }

    // put rest of semicolon at end of mesg
    for (var iarg = wv.numArgs; iarg < 9; iarg++) {
        mesg = mesg + ";";
    }

    if (wv.curMode == 3) {
        mesg = mesg + ";;";
    }

    if (nchange > 0 || wv.numArgs == 0) {
        wv.nchanges += nchange;

        // send the new values to the server
        wv.socketUt.send(mesg);

        // get an updated Branch list (so that added Branch is listed)
        if (wv.curMode == 2) {
            wv.socketUt.send("getBrchs;");
        }

        // update the UI
        if (wv.curMode == 3) {
            var id     = wv.menuEvent["target"].id;
            var myElem = document.getElementById(id);
            myElem.className = "fakelinkoff";
        }

        postMessage("Branch (type="+brch[ibrch].type+") has been added"+
                "  ====> Re-build is needed <====");

        activateBuildButton();
    } else {
        alert("no changes were made");
    }

    // return to the WebViewer
    changeMode(0);
}


//
// callback when "Cancel" is pressed in editBrchForm (called by ESP.html)
//
function editBrchCancel() {
    // alert("in editBrchCancel()");

    // if we are in process of adding a Branch, remove it now
    if (wv.curMode == 2) {
        brch.splice(brch.length-1, 1);
    }

    // return to the WebViewer
    changeMode(0);
}


//
// callback when "Add Attribute" is pressed in editBrchForm  (called by ESP.html)
//
function addAttr() {
    // alert("in addAttr()");

    // disable this command if there were any chnges to the Branch
    if (numberOfBrchChanges() > 0) {
        alert("Changes were made.  Press 'Cancel' or 'OK' first.");
        return;
    }

    // get the Tree node
    var id    = wv.menuEvent["target"].id;
    var inode = Number(id.substring(4,id.length-4));

    // get the Branch name
    var name = myTree.name[inode].replace(/\u00a0/g, "");

    // get the Branch index
    var ibrch = -1;           // 0-bias
    var jbrch;                // 1-bias (and temp)
    for (jbrch = 0; jbrch < brch.length; jbrch++) {
        if (brch[jbrch].name == name) {
            ibrch = jbrch;
            break;
        }
    }
    if (ibrch < 0) {
        alert(name+" not found");
        return;
    } else {
        jbrch = ibrch + 1;
    }

    // get the Attribute name and value
    var aname  = prompt("Enter Attribute name");
    if (aname === null) {
        return;
    } else if (aname.length <= 0) {
        return;
    }
    var avalue = prompt("Enter Attribute value");
    if (avalue === null) {
        return;
    } else if (avalue.length <= 0) {
        return;
    }

    // send the new value to the server
    var mesg = "setAttr;"+jbrch+";"+aname+";"+avalue+";";

    wv.socketUt.send(mesg);

    // update the UI
    postMessage("Adding attribute \""+aname+"\" (with value "+avalue+") to "+
                name+" ====> Re-build is needed <====");
    activateBuildButton();

    // get an updated version of the Branches
    wv.brchStat = 0;

    // return to  the WebViewer
    changeMode(0);
}


//
// callback when Attribute name is pressed in Tree
//
function editAttr(e) {
    // alert("in editAttr(e="+e+")");

    if (wv.curMode != 0) {
        alert("Command disabled.  Press 'Cancel' or 'OK' first.");
        return;
    } else if (wv.server != "serveCSM") {
        alert("editBrch is not implemented for "+wv.server);
        return;
    }

    wv.menuEvent = e;

    // get the Tree node
    var id    = wv.menuEvent["target"].id;
    var inode = Number(id.substring(4,id.length-4));

    // get the Attribute name and value
    var aname  = myTree.name[inode].replace(/\u00a0/g, "");
    var avalue = myTree.prop1[inode];

    // get the Branch name
    inode = myTree.parent[inode];

    var name = myTree.name[inode].replace(/\u00a0/g, "");

    // get the Branch index
    var ibrch = -1;           // 0-bias
    var jbrch;                // 1-bias (and temp)
    for (jbrch = 0; jbrch < brch.length; jbrch++) {
        if (brch[jbrch].name == name) {
            ibrch = jbrch;
            break;
        }
    }
    if (ibrch < 0) {
        alert(name+" not found");
        return;
    } else {
        jbrch = ibrch + 1;
    }

    // get the new Attribute value
    var anew = prompt("Enter new value for \""+aname+"\"", avalue);
    if (anew === null) {
        return;
    } else if (anew.length <= 0) {
        return;
    } else if (anew == avalue) {
        return;
    }

    // send the new value to the server
    var mesg = "setAttr;"+jbrch+";"+aname+";"+anew+";";

    wv.socketUt.send(mesg);

    // update the UI
    postMessage("Changing attribute \""+aname+"\" to "+
                anew+" ====> Re-build is needed <====");
    activateBuildButton();

    // get an updated version of the Branches
    wv.brchStat = 0;
}


//
// callback when "Build to the Branch" is pressed in editBrchForm (called by ESP.html)
//
function buildTo() {
    // alert("buildTo()");

    // disable this command if there were any chnges to the Branch
    if (numberOfBrchChanges() > 0) {
        alert("Changes were made.  Press 'Cancel' or 'OK' first.");
        return;
    }

    var ibrch = -1;
    var name  = "";

    // get the Tree node
    var id    = wv.menuEvent["target"].id;
    var inode = Number(id.substring(4,id.length-4));

    // get the Branch name
    name = myTree.name[inode].replace(/\u00a0/g, "");

    // get the Branch index
    for (var jbrch = 0; jbrch < brch.length; jbrch++) {
        if (brch[jbrch].name == name) {
            ibrch = jbrch + 1;
        }
    }
    if (ibrch <= 0) {
        alert(name+" not found");
        return;
    }

    // send the message to the server
    wv.socketUt.send("build;"+ibrch+";");

    // update the UI
    postMessage("Re-building only to "+name+"...");

    var button = document.getElementById("buildButton");
    button.onclick = null;
    button["innerHTML"] = "Re-building...";
    button.style.backgroundColor = "#FFFF3F";

    // inactivate buttons until build is done
    changeMode(-1);
}


//
// callback when "OK" button is pressed in editCsmForm (called by ESP.html)
//
function editCsmOk() {
    // alert("in editCsmOk()");

    var editCsmTextArea = document.getElementById("editCsmTextArea");

    // tell user if no changes were made
    if (wv.curFile == editCsmTextArea.value) {
        alert("No changes were made.");
        changeMode(0);
        return;
    } else {
        if (confirm("This will overwrite your input file.  Continue?")) {
        } else {
            return;
        }
    }

    // because of an apparent limit on the size of text
    //    messages that can be sent from the browser to the
    //    server, we need to send the new file back in
    //    pieces and then reassemble on the server
    var maxMessageSize = 1000;

    var ichar = 0;
    var part  = editCsmTextArea.value.substring(ichar, ichar+maxMessageSize);
    wv.socketUt.send("setCsmFileBeg;"+part);
    ichar += maxMessageSize;

    while (ichar < editCsmTextArea.value.length) {
        part = editCsmTextArea.value.substring(ichar, ichar+maxMessageSize);
        wv.socketUt.send("setCsmFileMid;"+part);
        ichar += maxMessageSize;
    }

    wv.socketUt.send("setCsmFileEnd;");

    // remember the edited .csm file
    wv.curFile = editCsmTextArea.value;

    postMessage("\""+wv.filename+"\" file has been changed.  Re-building...");

    // get an updated version of the Parameters and Branches
    wv.pmtrStat = 0;
    wv.brchStat = 0;

    // reset the number of changes
    wv.nchanges = 0;

    // inform the user that a rebuild is in process
    var button = document.getElementById("buildButton");
    button.onclick = null;
    button["innerHTML"] = "Re-building...";
    button.style.backgroundColor = "#FFFF3F";

    // inactivate buttons until build is done
    changeMode( 0);
    changeMode(-1);
}


//
// callback when "Cancel" is pressed in editCsmForm (called by ESP.html)
//
function editCsmCancel() {
    // alert("in editCsmCancel()");

    // return to the WebViewer
    changeMode(0);
}


//
// callback to get the cursor location
//
function getCursorXY(e)
{
    if (!e) var e = event;

    wv.cursorX  =  e.clientX - wv.offLeft             - 1;
    wv.cursorY  = -e.clientY + wv.offTop  + wv.height + 1;

                    wv.modifier  = 0;
    if (e.shiftKey) wv.modifier |= 1;
    if (e.altKey  ) wv.modifier |= 2;
    if (e.ctrlKey ) wv.modifier |= 4;
}


//
// callback when any mouse is pressed in canvas
//
function getMouseDown(e)
{
    if (!e) var e = event;

    wv.startX   =  e.clientX - wv.offLeft             - 1;
    wv.startY   = -e.clientY + wv.offTop  + wv.height + 1;

    wv.dragging = true;
    wv.button   = e.button;

                    wv.modifier  = 0;
    if (e.shiftKey) wv.modifier |= 1;
    if (e.altKey  ) wv.modifier |= 2;
    if (e.ctrlKey ) wv.modifier |= 4;
}


//
// callback when the mouse is released
//
function getMouseUp(e)
{
    wv.dragging = false;
}


//
// callback when the mouse leaves the canvas
//
function mouseLeftCanvas(e)
{
    if (wv.dragging) {
        wv.dragging = false;
    }
}


//
// callback when a key is pressed
//
function getKeyPress(e)
{
    if (wv.curMode != 0) {
        return;
    }

//    if (!e) var e = event;

    wv.keyPress = e.charCode;
    wv.keyCode  = e.keyCode;

                    wv.modifier  = 0;
    if (e.shiftKey) wv.modifier |= 1;
    if (e.altKey  ) wv.modifier |= 2;
    if (e.ctrlKey ) wv.modifier |= 4;

    // make sure that Firefox does not stop websockets if <esc> is pressed
    if (e.charCode == 0) {
        e.preventDefault();
    }
}


//
// callback when an arrow... or shift is pressed (needed for Chrome)
//
function getKeyDown(e) {
    if (e.charCode == 0) {
        if (e.keyCode == 33 ||          // PgUp
            e.keyCode == 34 ||          // PgDn
            e.keyCode == 36 ||          // Home
            e.keyCode == 37 ||          // Left
            e.keyCode == 38 ||          // Up
            e.keyCode == 39 ||          // Right
            e.keyCode == 40   ) {       // Down
            wv.keyCode  = e.keyCode;
            wv.keyPress = 0;
        } else if (e.keyCode == 16) {   // Shift
            wv.modifier = 1;
        }
    }
}


//
// callback when a shift is released (needed for Chrome)
//
function getKeyUp(e) {
    if (e.charCode == 0 && e.keyCode == 16) {
        wv.modifier = 0;
    }
}


//
// callback when the mouse is pressed in key window
//
function setKeyLimits(e)
{

    // get new limits
    var templo = prompt("Enter new lower limit", wv.loLimit);
    if (isNaN(templo)) {
        alert("Lower limit must be a number");
        return;
    }

    var tempup = prompt("Enter new upper limit", wv.upLimit);
    if (isNaN(tempup)) {
        alert("Upper limit must be a number");
        return;
    }

    if (templo != wv.loLimit || tempup != wv.uplimit) {
        wv.loLimit = templo;
        wv.upLimit = tempup;

        // send the limits back to the server
        wv.socketUt.send("setLims;"+wv.loLimit+";"+wv.upLimit+";");
    }
}


//
// callback to toggle Viz property
//
function toggleViz(e) {
    // alert("in toggleViz(e="+e+")");

    if (wv.curMode != 0) {
        alert("Command disabled.  Press 'Cancel' or 'OK' first.");
        return;
    }

    // get the Tree Node
    var inode = e["target"].id.substring(4);
    inode     = inode.substring(0,inode.length-4);
    inode     = Number(inode);

    // toggle the Viz property
    if        (myTree.valu1[inode] == "off") {
        myTree.prop(inode, 1, "on");
    } else if (myTree.valu1[inode] == "on") {
        myTree.prop(inode, 1, "off");
    } else {
        alert("illegal Viz property:"+myTree.valu1[inode]);
        return;
    }
}


//
// callback to toggle Grd property
//
function toggleGrd(e) {
    // alert("in toggleGrd(e="+e+")");

    if (wv.curMode != 0) {
        alert("Command disabled.  Press 'Cancel' or 'OK' first.");
        return;
    }

    // get the Tree Node
    var inode = e["target"].id.substring(4);
    inode     = inode.substring(0,inode.length-4);
    inode     = Number(inode);

    // toggle the Grd property
    if        (myTree.valu2[inode] == "off") {
        myTree.prop(inode, 2, "on");
    } else if (myTree.valu2[inode] == "on") {
        myTree.prop(inode, 2, "off");
    } else {
        alert("illegal Grd property:"+myTree.valu2[inode]);
        return;
    }
}


//
// callback to toggle Trn property
//
function toggleTrn(e) {
    // alert("in toggleTrn(e="+e+")");

    if (wv.curMode != 0) {
        alert("Command disabled.  Press 'Cancel' or 'OK' first.");
        return;
    }

    // get the Tree Node
    var inode = e["target"].id.substring(4);
    inode     = inode.substring(0,inode.length-4);
    inode     = Number(inode);

    // toggle the Trn property
    if        (myTree.valu3[inode] == "off") {
        myTree.prop(inode, 3, "on");
    } else if (myTree.valu3[inode] == "on") {
        myTree.prop(inode, 3, "off");
    } else {
        alert("illegal Trn property:"+myTree.valu3[inode]);
        return;
    }
}


//
// callback to toggle Ori property
//
function toggleOri(e) {
    // alert("in toggleOri(e="+e+")");

    if (wv.curMode != 0) {
        alert("Command disabled.  Press 'Cancel' or 'OK' first.");
        return;
    }

    // get the Tree Node
    var inode = e["target"].id.substring(4);
    inode     = inode.substring(0,inode.length-4);
    inode     = Number(inode);

    // toggle the Ori property
    if        (myTree.valu3[inode] == "off") {
        myTree.prop(inode, 3, "on");
    } else if (myTree.valu3[inode] == "on") {
        myTree.prop(inode, 3, "off");
    } else {
        alert("illegal Ori property:"+myTree.valu3[inode]);
        return;
    }
}


//
// constructor for a Tree
//
function Tree(doc, treeId) {
    // alert("in Tree(doc="+doc+", treeId="+treeId+")");

    // remember the document
    this.document = doc;
    this.treeId   = treeId;

    // arrays to hold the Nodes
    this.name   = new Array();
    this.gprim  = new Array();
    this.click  = new Array();
    this.cmenu  = new Array();
    this.parent = new Array();
    this.child  = new Array();
    this.next   = new Array();
    this.nprop  = new Array();
    this.opened = new Array();

    this.prop1  = new Array();
    this.valu1  = new Array();
    this.cbck1  = new Array();
    this.prop2  = new Array();
    this.valu2  = new Array();
    this.cbck2  = new Array();
    this.prop3  = new Array();
    this.valu3  = new Array();
    this.cbck3  = new Array();

    // initialize Node=0 (the root)
    this.name[  0] = "**root**";
    this.gprim[ 0] = "";
    this.click[ 0] = null;
    this.cmenu[ 0] = "";
    this.parent[0] = -1;
    this.child[ 0] = -1;
    this.next[  0] = -1;
    this.nprop[ 0] =  0;
    this.prop1[ 0] = "";
    this.valu1[ 0] = "";
    this.cbck1[ 0] = null;
    this.prop2[ 0] = "";
    this.valu2[ 0] = "";
    this.cbck2[ 0] = null;
    this.prop3[ 0] = "";
    this.valu3[ 0] = "";
    this.cbck3[ 0] = null;
    this.opened[0] = +1;

    // add methods
    this.addNode  = TreeAddNode;
    this.expand   = TreeExpand;
    this.contract = TreeContract;
    this.prop     = TreeProp;
    this.clear    = TreeClear;
    this.build    = TreeBuild;
    this.update   = TreeUpdate;
}


//
// add a Node to the Tree
//
function TreeAddNode(iparent, name, gprim, click, cmenu, prop1, valu1, cbck1,
                     prop2, valu2, cbck2, prop3, valu3, cbck3)
{
    // alert("in TreeAddNode(iparent="+iparent+", name="+name+", gprim="+gprim+", click="+click+
    //                       ", cmenu="+cmenu+", prop1="+prop1+", valu1="+valu1+",cbck1="+cbck1+
    //                       ", prop2="+prop2+", valu2="+valu2+", cbck2="+cbck2+", prop3="+prop3+
    //                       ", valu3="+valu3+", cbck3="+cbck3+")");

    // validate the input
    if (iparent < 0 || iparent >= this.name.length) {
        alert("iparent="+iparent+" is out of range");
        return;
    }

    // find the next Node index
    var inode = this.name.length;

    // store this Node's values
    this.name[  inode] = name;
    this.gprim[ inode] = gprim;
    this.click[ inode] = click;
    this.cmenu[ inode] = cmenu;
    this.parent[inode] = iparent;
    this.child[ inode] = -1;
    this.next[  inode] = -1;
    this.nprop[ inode] =  0;
    this.opened[inode] =  0;

    // store the properties
    if (prop1 !== undefined) {
        this.nprop[inode] = 1;
        this.prop1[inode] = prop1;
        this.valu1[inode] = valu1;
        this.cbck1[inode] = cbck1;
    }

    if (prop2 !== undefined) {
        this.nprop[inode] = 2;
        this.prop2[inode] = prop2;
        this.valu2[inode] = valu2;
        this.cbck2[inode] = cbck2;
    }

    if (prop3 !== undefined) {
        this.nprop[inode] = 3;
        this.prop3[inode] = prop3;
        this.valu3[inode] = valu3;
        this.cbck3[inode] = cbck3;
    }

    // if the parent does not have a child, link this
    //    new Node to the parent
    if (this.child[iparent] < 0) {
        this.child[iparent] = inode;

    // otherwise link this Node to the last parent's child
    } else {
        var jnode = this.child[iparent];
        while (this.next[jnode] >= 0) {
            jnode = this.next[jnode];
        }

        this.next[jnode] = inode;
    }
}


//
// build the Tree (ie, create the html table from the Nodes)
//
function TreeBuild() {
    // alert("in TreeBuild()");

    var doc = this.document;

    // if the table already exists, delete it and all its children (3 levels)
    var thisTable = doc.getElementById(this.treeId);
    if (thisTable) {
        var child1 = thisTable.lastChild;
        while (child1) {
            var child2 = child1.lastChild;
            while (child2) {
                var child3 = child2.lastChild;
                while (child3) {
                    child2.removeChild(child3);
                    child3 = child2.lastChild;
                }
                child1.removeChild(child2);
                child2 = child1.lastChild;
            }
            thisTable.removeChild(child1);
            child1 = thisTable.lastChild;
        }
        thisTable.parentNode.removeChild(thisTable);
    }

    // build the new table
    var newTable = doc.createElement("table");
    newTable.setAttribute("id", this.treeId);
    doc.getElementById("tlframe").appendChild(newTable);

    // traverse the Nodes using depth-first search
    var inode = 1;
    while (inode > 0) {

        // table row "node"+inode
        var newTR = doc.createElement("TR");
        newTR.setAttribute("id", "node"+inode);
        newTable.appendChild(newTR);

        // table data "node"+inode+"col1"
        var newTDcol1 = doc.createElement("TD");
        newTDcol1.setAttribute("id", "node"+inode+"col1");
        newTDcol1.className = "fakelinkon";
        newTR.appendChild(newTDcol1);

        var newTexta = doc.createTextNode("");
        newTDcol1.appendChild(newTexta);

        // table data "node"+inode+"col2"
        var newTDcol2 = doc.createElement("TD");
        newTDcol2.setAttribute("id", "node"+inode+"col2");
        if (this.cmenu[inode] != "") {
            newTDcol2.className = "fakelinkcmenu";
        }
        newTR.appendChild(newTDcol2);

        var newTextb = doc.createTextNode(this.name[inode]);
        newTDcol2.appendChild(newTextb);

        var name = this.name[inode].replace(/\u00a0/g, "");

        var ibrch = 0;
        for (var jbrch = 0; jbrch < brch.length; jbrch++) {
            if (brch[jbrch].name == name) {
                if (brch[jbrch].ileft == -2) {
                    newTDcol2.className = "errorTD";
                }
                break;
            }
        }

        // table data "node"+inode+"col3"
        if (this.nprop[inode] > 0) {
            var newTDcol3 = doc.createElement("TD");
            newTDcol3.setAttribute("id", "node"+inode+"col3");
            if (this.cbck1[inode] != "") {
                newTDcol3.className = "fakelinkon";
            }
            newTR.appendChild(newTDcol3);

            if (this.nprop[inode] == 1) {
                newTDcol3.setAttribute("colspan", "3");
            }

            var newTextc = doc.createTextNode(this.prop1[inode]);
            newTDcol3.appendChild(newTextc);
        }

        // table data "node:+inode+"col4"
        if (this.nprop[inode] > 1) {
            var newTDcol4 = doc.createElement("TD");
            newTDcol4.setAttribute("id", "node"+inode+"col4");
            if (this.cbck2[inode] != "") {
                newTDcol4.className = "fakelinkon";
            }
            newTR.appendChild(newTDcol4);

            if (this.nprop[inode] == 2) {
                newTDcol4.setAttribute("colspan", "2");
            }

            var newTextd = doc.createTextNode(this.prop2[inode]);
            newTDcol4.appendChild(newTextd);
        }

        // table data "node:+inode+"col5"
        if (this.nprop[inode] > 2) {
            var newTDcol5 = doc.createElement("TD");
            newTDcol5.setAttribute("id", "node"+inode+"col5");
            if (this.cbck3[inode] != "") {
                newTDcol5.className = "fakelinkon";
            }
            newTR.appendChild(newTDcol5);

            var newTextd = doc.createTextNode(this.prop3[inode]);
            newTDcol5.appendChild(newTextd);
        }

        // go to next row
        if        (this.child[inode] >= 0) {
            inode = this.child[inode];
        } else if (this.next[inode] >= 0) {
            inode = this.next[inode];
        } else {
            while (inode > 0) {
                inode = this.parent[inode];
                if (this.parent[inode] == 0) {
                    newTR = doc.createElement("TR");
                    newTR.setAttribute("height", "10px");
                    newTable.appendChild(newTR);
                }
                if (this.next[inode] >= 0) {
                    inode = this.next[inode];
                    break;
                }
            }
        }
    }

    this.update();
}


//
// clear the Tree
//
function TreeClear() {
    // alert("in TreeClear()");

    // remove all but the first Node
    this.name.splice(  1);
    this.gprim.splice( 1);
    this.click.splice( 1);
    this.cmenu.splice( 1);
    this.parent.splice(1);
    this.child.splice( 1);
    this.next.splice(  1);
    this.nprop.splice( 1);
    this.opened.splice(1);

    this.prop1.splice(1);
    this.valu1.splice(1);
    this.cbck1.splice(1);
    this.prop2.splice(1);
    this.valu2.splice(1);
    this.cbck2.splice(1);
    this.prop3.splice(1);
    this.valu3.splice(1);
    this.cbck3.splice(1);

    // reset the root Node
    this.parent[0] = -1;
    this.child[ 0] = -1;
    this.next[  0] = -1;
}


//
// expand a Node in the Tree
//
function TreeContract(inode) {
    // alert("in TreeContract(inode="+inode+")");

    // validate inputs
    if (inode < 0 || inode >= this.opened.length) {
        alert("inode="+inode+" is out of range");
        return;
    }

    // contract inode
    this.opened[inode] = 0;

    // contract all descendents of inode
    for (var jnode = 1; jnode < this.parent.length; jnode++) {
        var iparent = this.parent[jnode];
        while (iparent > 0) {
            if (iparent == inode) {
                this.opened[jnode] = 0;
                break;
            }

            iparent = this.parent[iparent];
        }
    }

    // update the display
    this.update();
}


//
// expand a Node in the Tree
//
function TreeExpand(inode) {
    // alert("in TreeExpand(inode="+inode+")");

    // validate inputs
    if (inode < 0 || inode >= this.opened.length) {
        alert("inode="+inode+" is out of range");
        return;
    }

    // expand inode
    this.opened[inode] = 1;

    // update the display
    this.update();
}


//
// change a property of a Node
//
function TreeProp(inode, iprop, onoff) {
    // alert("in TreeProp(inode="+inode+", iprop="+iprop+", onoff="+onoff+")");

    // validate inputs
    if (inode < 0 || inode >= this.opened.length) {
        alert("inode="+inode+" is out of range");
        return;
    } else if (onoff != "on" && onoff != "off") {
        alert("onoff="+onoff+" is not 'on' or 'off'");
        return;
    }

    // set the property for inode
    if (iprop == 1) {
        this.valu1[inode] = onoff;
    } else if (iprop == 2) {
        this.valu2[inode] = onoff;
    } else if (iprop == 3) {
        this.valu3[inode] = onoff;
    } else {
        alert("iprop="+iprop+" is not 1, 2, or 3");
        return;
    }

    // set property of all descendents of inode
    for (var jnode = 1; jnode < this.parent.length; jnode++) {
        var iparent = this.parent[jnode];
        while (iparent > 0) {
            if (iparent == inode) {
                if        (iprop == 1) {
                    this.valu1[jnode] = onoff;
                } else if (iprop == 2) {
                    this.valu2[jnode] = onoff;
                } else if (iprop == 3) {
                    this.valu3[jnode] = onoff;
                }
                break;
            }

            iparent = this.parent[iparent];
        }
    }

    this.update();
}


//
// update the Tree (after build/expension/contraction/property-set)
//
function TreeUpdate() {
    // alert("in TreeUpdate()");

    var doc = this.document;

    // traverse the Nodes using depth-first search
    for (var inode = 1; inode < this.opened.length; inode++) {
        var element = doc.getElementById("node"+inode);

        // unhide the row
        element.style.display = null;

        // hide the row if one of its parents has .opened=0
        var jnode = this.parent[inode];
        while (jnode != 0) {
            if (this.opened[jnode] == 0) {
                element.style.display = "none";
                break;
            }

            jnode = this.parent[jnode];
        }

        // if the current Node has children, set up appropriate event handler to expand/collapse
        if (this.child[inode] > 0) {
            if (this.opened[inode] == 0) {
                var myElem = doc.getElementById("node"+inode+"col1");
                var This   = this;

                myElem.firstChild.nodeValue = "+";
                myElem.onclick = function () {
                    var thisNode = this.id.substring(4);
                    thisNode     = thisNode.substring(0,thisNode.length-4);
                    This.expand(thisNode);
                };

            } else {
                var myElem = doc.getElementById("node"+inode+"col1");
                var This   = this;

                myElem.firstChild.nodeValue = "-";
                myElem.onclick = function () {
                    var thisNode = this.id.substring(4);
                    thisNode     = thisNode.substring(0,thisNode.length-4);
                    This.contract(thisNode);
                };
            }
        }

        if (this.click[inode] !== null) {
            var myElem = doc.getElementById("node"+inode+"col2");
            myElem.onclick = this.click[inode];
        }

        // set the class of the properties
        if (this.nprop[inode] >= 1) {
            var myElem = doc.getElementById("node"+inode+"col3");
            myElem.onclick = this.cbck1[inode];

            if (this.prop1[inode] == "Viz") {
                if (this.valu1[inode] == "off") {
                    myElem.setAttribute("class", "fakelinkoff");
                    if (this.gprim[inode] != "") {
                        wv.sceneGraph[this.gprim[inode]].attrs &= ~wv.plotAttrs.ON;
                        wv.sceneUpd = 1;
                    }
                } else {
                    myElem.setAttribute("class", "fakelinkon");
                    if (this.gprim[inode] != "") {
                        wv.sceneGraph[this.gprim[inode]].attrs |=  wv.plotAttrs.ON;
                        wv.sceneUpd = 1;
                    }
                }
            }
        }

        if (this.nprop[inode] >= 2) {
            var myElem = doc.getElementById("node"+inode+"col4");
            myElem.onclick = this.cbck2[inode];

            if (this.prop2[inode] == "Grd") {
                if (this.valu2[ inode] == "off") {
                    myElem.setAttribute("class", "fakelinkoff");

                    if (this.gprim[inode] != "") {
                        wv.sceneGraph[this.gprim[inode]].attrs &= ~wv.plotAttrs.LINES;
                        wv.sceneGraph[this.gprim[inode]].attrs &= ~wv.plotAttrs.POINTS;
                        wv.sceneUpd = 1;
                    }
                } else {
                    myElem.setAttribute("class", "fakelinkon");

                    if (this.gprim[inode] != "") {
                        wv.sceneGraph[this.gprim[inode]].attrs |=  wv.plotAttrs.LINES;
                        wv.sceneGraph[this.gprim[inode]].attrs |=  wv.plotAttrs.POINTS;
                        wv.sceneUpd = 1;
                    }
                }
            }
        }

        if (this.nprop[inode] >= 3) {
            var myElem = doc.getElementById("node"+inode+"col5");
            myElem.onclick = this.cbck3[inode];

            if (this.prop3[inode] == "Trn") {
                if (this.valu3[ inode] == "off") {
                    myElem.setAttribute("class", "fakelinkoff");

                    if (this.gprim[inode] != "") {
                        wv.sceneGraph[this.gprim[inode]].attrs &= ~wv.plotAttrs.TRANSPARENT;
                        wv.sceneUpd = 1;
                    }
                } else {
                    myElem.setAttribute("class", "fakelinkon");

                    if (this.gprim[inode] != "") {
                        wv.sceneGraph[this.gprim[inode]].attrs |=  wv.plotAttrs.TRANSPARENT;
                        wv.sceneUpd = 1;
                    }
                }
            } else if (this.prop3[inode] == "Ori") {
                if (this.valu3[ inode] == "off") {
                    myElem.setAttribute("class", "fakelinkoff");

                    if (this.gprim[inode] != "") {
                        wv.sceneGraph[this.gprim[inode]].attrs &= ~wv.plotAttrs.ORIENTATION;
                        wv.sceneUpd = 1;
                    }
                } else {
                    myElem.setAttribute("class", "fakelinkon");

                    if (this.gprim[inode] != "") {
                        wv.sceneGraph[this.gprim[inode]].attrs |=  wv.plotAttrs.ORIENTATION;
                        wv.sceneUpd = 1;
                    }
                }
            }
        }
    }
}


//
// change mode for trframe
//
function changeMode(newMode) {
    // alert("in changeMode(newMode="+newMode+")");

    var webViewer      = document.getElementById("WebViewer");
    var addBrchForm    = document.getElementById("addBrchForm");
    var editBrchForm   = document.getElementById("editBrchForm");
    var addBrchHeader  = document.getElementById("addBrchHeader");
    var editBrchHeader = document.getElementById("editBrchHeader");
    var editPmtrForm   = document.getElementById("editPmtrForm");
    var addPmtrHeader  = document.getElementById("addPmtrHeader");
    var editPmtrHeader = document.getElementById("editPmtrHeader");
    var editCsmForm    = document.getElementById("editCsmForm");

    if (newMode == wv.curMode) {
        return;
    } else if (newMode < 0) {
        // used to cause buttons such as "cmdSave" not to be active
        wv.curMode = newMode;
        return;
    } else if (newMode == 0) {
        webViewer.hidden    = false;
        addBrchForm.hidden  = true;
        editBrchForm.hidden = true;
        editPmtrForm.hidden = true;
        editCsmForm.hidden  = true;

        wv.curMode   = 0;
        wv.curPmtr   = -1;
        wv.curBrch   = -1;
        wv.afterBrch = -1;
        wv.menuEvent = undefined;
    } else if (newMode == 1) {
        webViewer.hidden    = true;
        addBrchForm.hidden  = false;
        editBrchForm.hidden = true;
        editPmtrForm.hidden = true;
        editCsmForm.hidden  = true;

        // unselect all items
        var elements = document.getElementById("addBrchForm").elements;
        for (var ielem = 0; ielem < elements.length; ielem++) {
            elements[ielem].checked = false;
        }

        wv.curMode = 1;
    } else if (newMode == 2) {
        webViewer.hidden    = true;
        addBrchForm.hidden  = true;
        editBrchForm.hidden = false;
        editPmtrForm.hidden = true
        editCsmForm.hidden  = true;

        addBrchHeader.hidden  = false;
        editBrchHeader.hidden = true;

        wv.curMode = 2;
    } else if (newMode == 3) {
        webViewer.hidden    = true;
        addBrchForm.hidden  = true;
        editBrchForm.hidden = false;
        editPmtrForm.hidden = true
        editCsmForm.hidden  = true;

        addBrchHeader.hidden  = true;
        editBrchHeader.hidden = false;

        wv.curMode = 3;
    } else if (newMode == 4) {
        webViewer.hidden    = true;
        addBrchForm.hidden  = true;
        editBrchForm.hidden = true;
        editPmtrForm.hidden = false;
        editCsmForm.hidden  = true;

        addPmtrHeader.hidden  = false;
        editPmtrHeader.hidden = true;

        wv.curMode = 4;
    } else if (newMode == 5) {
        webViewer.hidden    = true;
        addBrchForm.hidden  = true;
        editBrchForm.hidden = true;
        editPmtrForm.hidden = false;
        editCsmForm.hidden  = true;

        addPmtrHeader.hidden  = true;
        editPmtrHeader.hidden = false;

        wv.curMode = 5;
    } else if (newMode == 6) {
        webViewer.hidden    = true;
        addBrchForm.hidden  = true;
        editBrchForm.hidden = true;
        editPmtrForm.hidden = true;
        editCsmForm.hidden  = false;

        wv.curMode = 6;
    } else {
        alert("Bad new mode = "+newMode);
    }
}


//
// post a message into the brframe
//
function postMessage(mesg) {
    // alert("in postMessage(mesg="+mesg+")");

    var botm = document.getElementById("brframe");

    var pre  = document.createElement("pre");
    var text = document.createTextNode(mesg);
    pre.appendChild(text);
    botm.insertBefore(pre, botm.lastChild);

    pre.scrollIntoView();
}


//
// callback when "onresize" event occurs (called by ESP.html)
//
// resize the frames (with special handling to width of tlframe and height of brframe
//
function resizeFrames() {
    // alert("resizeFrames()");

    var scrollSize;
    if (BrowserDetect.browser == "Chrome") {
        scrollSize = 24;
    } else {
        scrollSize = 20;
    }

    // get the size of the client (minus amount to account for scrollbars)
    var body = document.getElementById("mainBody");
    var bodyWidth  = body.clientWidth  - scrollSize;
    var bodyHeight = body.clientHeight - scrollSize;

    // get the elements associated with the frames and the canvas
    var topleft = document.getElementById("tlframe");
    var toprite = document.getElementById("trframe");
    var botleft = document.getElementById("blframe")
    var botrite = document.getElementById("brframe");
    var canv    = document.getElementById(wv.canvasID);

    // compute and set the widths of the frames
    //    (do not make tlframe larger than 250px)
    var leftWidth = Math.round(0.25 * bodyWidth);
    if (leftWidth > 250)   leftWidth = 250;
    var riteWidth = bodyWidth - leftWidth;
    var canvWidth = riteWidth - scrollSize;

    topleft.style.width = leftWidth+"px";
    toprite.style.width = riteWidth+"px";
    botleft.style.width = leftWidth+"px";
    botrite.style.width = riteWidth+"px";
    canv.style.width    = canvWidth+"px";
    canv.width          = canvWidth;

    // compute and set the heights of the frames
    //    (do not make brframe larger than 200px)
    var botmHeight = Math.round(0.20 * bodyHeight);
    if (botmHeight > 200)   botmHeight = 200;
    var  topHeight = bodyHeight - botmHeight;
    var canvHeight =  topHeight - scrollSize - 5;
    var keyHeight  = botmHeight - 25;

    topleft.style.height =  topHeight+"px";
    toprite.style.height =  topHeight+"px";
    botleft.style.height = botmHeight+"px";
    botrite.style.height = botmHeight+"px";
    canv.style.height    = canvHeight+"px";
    canv.height          = canvHeight;

    // set up canvas associated with key
    if (wv.canvasKY != undefined) {
        var canf = document.getElementById(wv.canvasKY);

        var keyfWidth     = leftWidth - 20;
        canf.style.width  = keyfWidth + "px";
        canf.width        = keyfWidth;

        var keyfHeight    = botmHeight - 25;
        canf.style.height = keyfHeight + "px";
        canf.height       = keyfHeight;

        // force a key redraw
        wv.drawKey = 1;
    }
}


//
// load info into editBrchForm
//
function setupEditBrchForm() {
    // alert("setupEditBrchForm()");

    var ibrch = wv.curBrch;
    var name  = brch[ibrch].name;
    var type  = brch[ibrch].type;

    var editBrchForm = document.getElementById("editBrchForm");

    // turn all arguments off (by default)
    document.getElementById("argName1").parentNode.style.display = "none";
    document.getElementById("argName2").parentNode.style.display = "none";
    document.getElementById("argName3").parentNode.style.display = "none";
    document.getElementById("argName4").parentNode.style.display = "none";
    document.getElementById("argName5").parentNode.style.display = "none";
    document.getElementById("argName6").parentNode.style.display = "none";
    document.getElementById("argName7").parentNode.style.display = "none";
    document.getElementById("argName8").parentNode.style.display = "none";
    document.getElementById("argName9").parentNode.style.display = "none";

    // fill in the Branch type and name
    document.getElementById("brchType").firstChild["data"] = type;
    editBrchForm.            brchName.value                = name;

    // set up arguments based upon the type
    var argList;
    var defValue;
    var suppress = 0;   // set to 1 if type can be suppressed

    if (type == "box") {
        argList  = ["xmin", "ymin", "zmin", "xsize", "ysize", "zsize"];
        defValue = ["",     "",     "",     "",      "",      ""     ];
        suppress = 1;
    } else if (type == "sphere") {
        argList  = ["xcent", "ycent", "zcent", "radius"];
        defValue = ["",      "",      "",      ""      ];
        suppress = 1;
    } else if (type == "cone") {
        argList  = ["xvrtx", "yvrtx", "zvrtx", "xbase", "ybase", "zbase", "radius"];
        defValue = ["",      "",      "",      "",      "",      "",      ""      ];
        suppress = 1;
    } else if (type == "cylinder") {
        argList  = ["xbeg", "ybeg", "zbeg", "xend", "yend", "zend", "radius"];
        defValue = ["",     "",     "",     "",     "",     "",     ""      ];
        suppress = 1;
    } else if (type == "torus") {
        argList  = ["xcent", "ycent", "zcent", "dxaxis", "dyaxis", "dzaxis", "majorRad", "minorRad"];
        defValue = ["",      "",      "",      "",       "",       "",       "",         ""        ];
        suppress = 1;
    } else if (type == "import") {
        argList  = ["$filename", "bodynumber"];
        defValue = ["",          "1"         ];
        suppress = 1;
    } else if (type == "udprim") {
        argList  = ["$primtype", "$argName1", "argValue1", "$argName2", "argValue2", "$argName3", "argValue3", "$argName4", "argValue4"];
        defValue = ["",          "",          "",          "",          "",          "",          "",          "",          ""         ];
        suppress = 1;
    } else if (type == "extrude") {
        argList  = ["dx", "dy", "dz"];
        defValue = ["",   "",   ""  ];
        suppress = 1;
    } else if (type == "rule") {
        argList  = [];
        defValue = [];
        suppress = 1;
    } else if (type == "blend") {
        argList  = ["$noseList", "$tailList"];
        defValue = ["0",         "0"        ];
        suppress = 1;
    } else if (type == "revolve") {
        argList  = ["xorig", "yorig", "zorig", "dxaxis", "dyaxis", "dzaxis", "angDeg"];
        defValue = ["",      "",      "",      "",       "",       "",       ""      ];
        suppress = 1;
    } else if (type == "sweep") {
        argList  = [];
        defValue = [];
        suppress = 1;
    } else if (type == "loft") {
        alert("Consider using 'rule' or 'blend'");
        argList  = ["smooth"];
        defValue = [""      ];
        suppress = 1;
    } else if (type == "fillet") {
        argList  = ["radius", "$edgeList"];
        defValue = ["",       "0"        ];
        suppress = 1;
    } else if (type == "chamfer") {
        argList  = ["radius", "$edgeList"];
        defValue = ["",       "0"        ];
        suppress = 1;
    } else if (type == "hollow") {
        argList  = ["thick", "$faceList"];
        defValue = ["",      "0"        ];
        suppress = 1;
    } else if (type == "intersect") {
        argList  = ["$order", "index"];
        defValue = ["none",   "1"    ];
    } else if (type == "subtract") {
        argList  = ["$order", "index"];
        defValue = ["none",   "1"    ];
    } else if (type == "union") {
        argList  = ["tomark"];
        defValue = ["0"     ];
    } else if (type == "translate") {
        argList  = ["dx", "dy", "dz"];
        defValue = ["",   "",   ""  ];
        suppress = 1;
    } else if (type == "rotatex") {
        argList  = ["angDeg", "yaxis", "zaxis"];
        defValue = ["",       "",      ""     ];
        suppress = 1;
    } else if (type == "rotatey") {
        argList  = ["angDeg", "zaxis", "xaxis"];
        defValue = ["",       "",      ""     ];
        suppress = 1;
    } else if (type == "rotatez") {
        argList  = ["angDeg", "xaxis", "yaxis"];
        defValue = ["",       "",      ""     ];
        suppress = 1;
    } else if (type == "scale") {
        argList  = ["fact"];
        defValue = [""    ];
        suppress = 1;
    } else if (type == "mirror") {
        argList  = ["nx", "ny", "nz", "dist"];
        defValue = ["",   "",   "",   "0"   ];
        suppress = 1;
    } else if (type == "skbeg") {
        argList  = ["x", "y", "z"];
        defValue = ["",  "",  "" ];
    } else if (type == "linseg") {
        argList  = ["x", "y", "z"];
        defValue = ["",  "",  "" ];
    } else if (type == "cirarc") {
        argList  = ["xon", "yon", "zon", "xend", "yend", "zend"];
        defValue = ["",    "",    "",    "",     "",     ""    ];
    } else if (type == "spline") {
        argList  = ["x", "y", "z"];
        defValue = ["",  "",  "" ];
    } else if (type == "skend") {
        argList  = [];
        defValue = [];
    } else if (type == "solbeg") {
        argList  = ["$varList"];
        defValue = [""        ];
    } else if (type == "solcon") {
        argList  = ["$expr"];
        defValue = [""     ];
    } else if (type == "solend") {
        argList  = [];
        defValue = [];
    } else if (type == "set") {
        argList  = ["$pmtrName", "$exprs"];
        defValue = ["",          ""      ];
    } else if (type == "udparg") {
        argList  = ["$primtype", "$argName", "argvValue"];
        defValue = ["",          "",         ""         ];
    } else if (type == "mark") {
        argList  = [];
        defValue = [];
    } else if (type == "dump") {
        argList  = ["$filename", "remove"];
        defValue = ["",          "0"     ];
    } else if (type == "patbeg") {
        argList  = ["$pmtrName", "ncopy"];
        defValue = ["",          ""     ];
    } else if (type == "patend") {
        argList  = [];
        defValue = [];
    } else if (type == "store") {
        alert("'store' is not implemented yet.");
        changeMode(0);
        return 1;
    } else if (type == "restore") {
        alert("'restore' is not implemented yet.");
        changeMode(0);
        return 1;
    } else if (type == "include") {
        alert("'include' is not implemented yet.");
        changeMode(0);
        return 1;
    } else if (type == "return") {
        alert("'return' is not implemented yet.");
        changeMode(0);
        return 1;
    } else if (type == "select") {
        argList  = ["$type", "arg1", "arg2", "arg3", "arg4", "arg5", "arg6", "arg7", "arg8"];
        defValue = ["",      "",     "",     "",     "",     "",     "",     "",     ""    ];
    } else if (type == "macbeg") {
        alert("'macbeg' is deprecated");
        argList  = ["imacro"];
        defValue = [""      ];
    } else if (type == "macend") {
        alert("'macend' is deprecated");
        argList  = [];
        defValue = [];
    } else if (type == "recall") {
        aleft("'recall' is deprecated");
        argList  = ["imicro"];
        defValue = [""      ];
    }

    // set up the activity
    if (suppress) {
        if        (brch[ibrch].actv == 300) {     // Active
            editBrchForm.activity.value = 300;
        } else if (brch[ibrch].actv == 301) {     // Suppressed
            editBrchForm.activity.value = 301;
        } else {                                  // Inactive or Deferred
            editBrchForm.activity.value = 300;
        }
        editBrchForm.activity.style.display = "table-row";
    } else {
        editBrchForm.activity.style.display = "none";
    }

    // set up the arguments
    wv.numArgs = argList.length;

    if (wv.numArgs >= 1) {
        document.getElementById("argName1").parentNode.style.display = "table-row";
        document.getElementById("argName1").firstChild["data"]       = argList[0];
        if (wv.curMode == 1 || brch[ibrch].args[0] === undefined) {
            editBrchForm.        argValu1.value = defValue[0];
        } else if (argList[0].charAt(0) == "$") {
            editBrchForm.        argValu1.value = brch[ibrch].args[0].substr(1,brch[ibrch].args[0].length);
        } else {
            editBrchForm.        argValu1.value = brch[ibrch].args[0];
        }
    }
    if (wv.numArgs >= 2) {
        document.getElementById("argName2").parentNode.style.display = "table-row";
        document.getElementById("argName2").firstChild["data"]       = argList[1];
        if (wv.curMode == 1 || brch[ibrch].args[1] === undefined) {
            editBrchForm.        argValu2.value = defValue[1];
        } else if (argList[1].charAt(0) == "$") {
            editBrchForm.        argValu2.value = brch[ibrch].args[1].substr(1,brch[ibrch].args[1].length);
        } else {
            editBrchForm.        argValu2.value = brch[ibrch].args[1];
        }
    }
    if (wv.numArgs >= 3) {
        document.getElementById("argName3").parentNode.style.display = "table-row";
        document.getElementById("argName3").firstChild["data"]       = argList[2];
        if (wv.curMode == 1 || brch[ibrch].args[2] === undefined) {
            editBrchForm.        argValu3.value = defValue[2];
        } else if (argList[2].charAt(0) == "$") {
            editBrchForm.        argValu3.value = brch[ibrch].args[2].substr(1,brch[ibrch].args[2].length);
        } else {
            editBrchForm.        argValu3.value = brch[ibrch].args[2];
        }
    }
    if (wv.numArgs >= 4) {
        document.getElementById("argName4").parentNode.style.display = "table-row";
        document.getElementById("argName4").firstChild["data"]       = argList[3];
        if (wv.curMode == 1 || brch[ibrch].args[3] === undefined) {
            editBrchForm.        argValu4.value = defValue[3];
        } else if (argList[3].charAt(0) == "$") {
            editBrchForm.        argValu4.value = brch[ibrch].args[3].substr(1,brch[ibrch].args[3].length);
        } else {
            editBrchForm.        argValu4.value = brch[ibrch].args[3];
        }
    }
    if (wv.numArgs >= 5) {
        document.getElementById("argName5").parentNode.style.display = "table-row";
        document.getElementById("argName5").firstChild["data"]       = argList[4];
        if (wv.curMode == 1 || brch[ibrch].args[4] === undefined) {
            editBrchForm.        argValu5.value = defValue[4];
        } else if (argList[4].charAt(0) == "$") {
            editBrchForm.        argValu5.value = brch[ibrch].args[4].substr(1,brch[ibrch].args[4].length);
        } else {
            editBrchForm.        argValu5.value = brch[ibrch].args[4];
        }
    }
    if (wv.numArgs >= 6) {
        document.getElementById("argName6").parentNode.style.display = "table-row";
        document.getElementById("argName6").firstChild["data"]       = argList[5];
        if (wv.curMode == 1 || brch[ibrch].args[5] === undefined) {
            editBrchForm.        argValu6.value = defValue[5];
        } else if (argList[5].charAt(0) == "$") {
            editBrchForm.        argValu6.value = brch[ibrch].args[5].substr(1,brch[ibrch].args[5].length);
        } else {
            editBrchForm.        argValu6.value = brch[ibrch].args[5];
        }
    }
    if (wv.numArgs >= 7) {
        document.getElementById("argName7").parentNode.style.display = "table-row";
        document.getElementById("argName7").firstChild["data"]       = argList[6];
        if (wv.curMode == 1 || brch[ibrch].args[6] === undefined) {
            editBrchForm.        argValu7.value = defValue[6];
        } else if (argList[6].charAt(0) == "$") {
            editBrchForm.        argValu7.value = brch[ibrch].args[6].substr(1,brch[ibrch].args[6].length);
        } else {
            editBrchForm.        argValu7.value = brch[ibrch].args[6];
        }
    }
    if (wv.numArgs >= 8) {
        document.getElementById("argName8").parentNode.style.display = "table-row";
        document.getElementById("argName8").firstChild["data"]       = argList[7];
        if (wv.curMode == 1 || brch[ibrch].args[7] === undefined) {
            editBrchForm.        argValu8.value = defValue[7];
        } else if (argList[7].charAt(0) == "$") {
            editBrchForm.        argValu8.value = brch[ibrch].args[7].substr(1,brch[ibrch].args[7].length);
        } else {
            editBrchForm.        argValu8.value = brch[ibrch].args[7];
        }
    }
    if (wv.numArgs >= 9) {
        document.getElementById("argName9").parentNode.style.display = "table-row";
        document.getElementById("argName9").firstChild["data"]       = argList[8];
        if (wv.curMode == 1 || brch[ibrch].args[8] === undefined) {
            editBrchForm.        argValu9.value = defValue[8];
        } else if (argList[8].charAt(0) == "$") {
            editBrchForm.        argValu9.value = brch[ibrch].args[8].substr(1,brch[ibrch].args[8].length);
        } else {
            editBrchForm.        argValu9.value = brch[ibrch].args[8];
        }
    }

    return 0;
}


//
// load info into editPmtrForm
//
function setupEditPmtrForm() {
    // alert("setupEditPmtrForm()");

    var ipmtr = wv.curPmtr;
    var name  = pmtr[ipmtr].name;
    var nrow  = pmtr[ipmtr].nrow;
    var ncol  = pmtr[ipmtr].ncol;

    var editPmtrForm = document.getElementById("editPmtrForm");

    // fill in the Parameter name
    document.getElementById("pmtrName").firstChild["data"] = name;

    var pmtrTable = document.getElementById("editPmtrTable");

    // remove old table entries
    if (pmtrTable) {
        var child1 = pmtrTable.lastChild;
        while (child1) {
            var child2 = child1.lastChild;
            while (child2) {

                var child3 = child2.lastChild;
                while (child3) {
                    child2.removeChild(child3);
                    child3 = child2.lastChild;
                }
                child1.removeChild(child2);
                child2 = child1.lastChild;
            }
            pmtrTable.removeChild(child1);
            child1 = pmtrTable.lastChild;
        }
    }

    // build the table that will contain values
    for (var irow = 0; irow <= nrow; irow++) {
        var newTR = document.createElement("TR");
        pmtrTable.appendChild(newTR);

        // fill the row
        if (irow == 0) {
            var newTD = document.createElement("TD");
            newTR.appendChild(newTD);

            var newText = document.createTextNode("");
            newTD.appendChild(newText);

            for (var icol = 1; icol <= ncol; icol++) {
                var newTD = document.createElement("TD");
                newTR.appendChild(newTD);

                var newText = document.createTextNode("col\u00a0"+icol);
                newTD.appendChild(newText);
            }
        } else{
            var newTD = document.createElement("TD");
            newTR.appendChild(newTD);

            var newText = document.createTextNode("row\u00a0"+irow);
            newTD.appendChild(newText);

            for (var icol = 1; icol <= ncol; icol++) {
                var indx = icol-1 + (irow-1)*pmtr[ipmtr].ncol;

                var newTD = document.createElement("TD");
                newTR.appendChild(newTD);

                var newInput = document.createElement("input");
                newInput.type  = "text";
                newInput.name  = "row"+irow+"col"+icol;
                newInput.size  = 12;
                newInput.value = pmtr[ipmtr].value[indx];
                newTD.appendChild(newInput);
            }
        }
    }

    return 0;
}


//
// count number of Parameter value changes
//
function numberOfPmtrChanges() {
    // alert("in numberOfPmtrChanges()");

    var nchange = 0;

    var editPmtrForm = document.getElementById("editPmtrForm");

    var ipmtr = wv.curPmtr;

    // examine each of the values
    var index   = -1;
    for (var irow = 1; irow <= pmtr[ipmtr].nrow; irow++) {
        for (var icol = 1; icol <= pmtr[ipmtr].ncol; icol++) {
            index++;

            // get the new value
            var myInput = editPmtrForm["row"+irow+"col"+icol];
            var value   = myInput.value.replace(/\s/g, "");

            if (value != pmtr[ipmtr].value[index]) {
                nchange++;
            }
        }
    }

    // return the number of changes
    return nchange;
}


//
// count number of Branch changes
//
function numberOfBrchChanges() {
    // alert("in numberOfBrchChanges()");

    var nchange = 0;

    var editBrchForm = document.getElementById("editBrchForm");

    var ibrch = wv.curBrch;

    // check the name
    if (brch[ibrch].name != editBrchForm.brchName.value.replace(/\s/g, "")) {
        nchange++;
    }

    // check the activity
    if (brch[ibrch].actv != editBrchForm.activity.value) {
        nchange++;
    }

    // check the arguments
    for (var iarg = 0; iarg < wv.numArgs; iarg++) {
        if        (iarg == 0) {
            var name  = document.getElementById("argName1").firstChild["data"];
            var value = editBrchForm.            argValu1.value.replace(/\s/g, "");
        } else if (iarg == 1) {
            var name  = document.getElementById("argName2").firstChild["data"];
            var value = editBrchForm.            argValu2.value.replace(/\s/g, "");
        } else if (iarg == 2) {
            var name  = document.getElementById("argName3").firstChild["data"];
            var value = editBrchForm.            argValu3.value.replace(/\s/g, "");
        } else if (iarg == 3) {
            var name  = document.getElementById("argName4").firstChild["data"];
            var value = editBrchForm.            argValu4.value.replace(/\s/g, "");
        } else if (iarg == 4) {
            var name  = document.getElementById("argName5").firstChild["data"];
            var value = editBrchForm.            argValu5.value.replace(/\s/g, "");
        } else if (iarg == 5) {
            var name  = document.getElementById("argName6").firstChild["data"];
            var value = editBrchForm.            argValu6.value.replace(/\s/g, "");
        } else if (iarg == 6) {
            var name  = document.getElementById("argName7").firstChild["data"];
            var value = editBrchForm.            argValu7.value.replace(/\s/g, "");
        } else if (iarg == 7) {
            var name  = document.getElementById("argName8").firstChild["data"];
            var value = editBrchForm.            argValu8.value.replace(/\s/g, "");
        } else if (iarg == 8) {
            var name  = document.getElementById("argName9").firstChild["data"];
            var value = editBrchForm.            argValu9.value.replace(/\s/g, "");
        }

        var output;
        if (name.charAt(0) != "$" || value.length <= 0) {
            output =       value;
        } else {
            output = "$" + value;
        }

        if (output != brch[ibrch].args[iarg]) {
            nchange++;
        }
    }

    // return the number of changes
    return nchange;
}
