/*
 ************************************************************************
 *                                                                      *
 * serveCSM.c -- server for driving OpenCSM                             *
 *                                                                      *
 *              Written by John Dannenhoffer @ Syracuse University      *
 *                                                                      *
 ************************************************************************
 */

/*
 * Copyright (C) 2012/2014  John F. Dannenhoffer, III (Syracuse University)
 *
 * This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation; either
 *    version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 *    License along with this library; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *     MA  02110-1301  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <unistd.h>

#ifdef WIN32
    #define WIN32_LEAN_AND_MEAN
    #include <windows.h>
    #include <winsock2.h>
#endif

#if   defined(GEOM_CAPRI)
    #include "capri.h"

    #define CINT    int
    #define CDOUBLE double
    #define CCHAR   char
#elif defined(GEOM_EGADS)
    #include "egads.h"

    #define CINT    const int
    #define CDOUBLE const double
    #define CCHAR   const char
#else
    #error no_geometry_system_specified
#endif

#ifdef WIN32
    #define snprintf _snprintf
#endif

#define STRNCPY(A, B, LEN) strncpy(A, B, LEN); A[LEN-1] = '\0';

#include "common.h"
#include "OpenCSM.h"

#include "wsserver.h"

/***********************************************************************/
/*                                                                     */
/* macros (including those that go along with common.h)                */
/*                                                                     */
/***********************************************************************/

#ifdef DEBUG
   #define DOPEN {if (dbg_fp == NULL) dbg_fp = fopen("serveCSM.dbg", "w");}
   static  FILE *dbg_fp=NULL;
#endif

//static void *realloc_temp=NULL;            /* used by RALLOC macro */

#define  RED(COLOR)      (float)(COLOR / 0x10000        ) / (float)(255)
#define  GREEN(COLOR)    (float)(COLOR / 0x00100 % 0x100) / (float)(255)
#define  BLUE(COLOR)     (float)(COLOR           % 0x100) / (float)(255)

#define MAX_EXPR_LEN     128
#define MAX_STR_LEN    32767

/***********************************************************************/
/*                                                                     */
/* structures                                                          */
/*                                                                     */
/***********************************************************************/

static float color_map[256*3] =
{ 0.0000,    0.0000,    1.0000,     0.0000,    0.0157,    1.0000,
  0.0000,    0.0314,    1.0000,     0.0000,    0.0471,    1.0000,
  0.0000,    0.0627,    1.0000,     0.0000,    0.0784,    1.0000,
  0.0000,    0.0941,    1.0000,     0.0000,    0.1098,    1.0000,
  0.0000,    0.1255,    1.0000,     0.0000,    0.1412,    1.0000,
  0.0000,    0.1569,    1.0000,     0.0000,    0.1725,    1.0000,
  0.0000,    0.1882,    1.0000,     0.0000,    0.2039,    1.0000,
  0.0000,    0.2196,    1.0000,     0.0000,    0.2353,    1.0000,
  0.0000,    0.2510,    1.0000,     0.0000,    0.2667,    1.0000,
  0.0000,    0.2824,    1.0000,     0.0000,    0.2980,    1.0000,
  0.0000,    0.3137,    1.0000,     0.0000,    0.3294,    1.0000,
  0.0000,    0.3451,    1.0000,     0.0000,    0.3608,    1.0000,
  0.0000,    0.3765,    1.0000,     0.0000,    0.3922,    1.0000,
  0.0000,    0.4078,    1.0000,     0.0000,    0.4235,    1.0000,
  0.0000,    0.4392,    1.0000,     0.0000,    0.4549,    1.0000,
  0.0000,    0.4706,    1.0000,     0.0000,    0.4863,    1.0000,
  0.0000,    0.5020,    1.0000,     0.0000,    0.5176,    1.0000,
  0.0000,    0.5333,    1.0000,     0.0000,    0.5490,    1.0000,
  0.0000,    0.5647,    1.0000,     0.0000,    0.5804,    1.0000,
  0.0000,    0.5961,    1.0000,     0.0000,    0.6118,    1.0000,
  0.0000,    0.6275,    1.0000,     0.0000,    0.6431,    1.0000,
  0.0000,    0.6588,    1.0000,     0.0000,    0.6745,    1.0000,
  0.0000,    0.6902,    1.0000,     0.0000,    0.7059,    1.0000,
  0.0000,    0.7216,    1.0000,     0.0000,    0.7373,    1.0000,
  0.0000,    0.7529,    1.0000,     0.0000,    0.7686,    1.0000,
  0.0000,    0.7843,    1.0000,     0.0000,    0.8000,    1.0000,
  0.0000,    0.8157,    1.0000,     0.0000,    0.8314,    1.0000,
  0.0000,    0.8471,    1.0000,     0.0000,    0.8627,    1.0000,
  0.0000,    0.8784,    1.0000,     0.0000,    0.8941,    1.0000,
  0.0000,    0.9098,    1.0000,     0.0000,    0.9255,    1.0000,
  0.0000,    0.9412,    1.0000,     0.0000,    0.9569,    1.0000,
  0.0000,    0.9725,    1.0000,     0.0000,    0.9882,    1.0000,
  0.0000,    1.0000,    0.9961,     0.0000,    1.0000,    0.9804,
  0.0000,    1.0000,    0.9647,     0.0000,    1.0000,    0.9490,
  0.0000,    1.0000,    0.9333,     0.0000,    1.0000,    0.9176,
  0.0000,    1.0000,    0.9020,     0.0000,    1.0000,    0.8863,
  0.0000,    1.0000,    0.8706,     0.0000,    1.0000,    0.8549,
  0.0000,    1.0000,    0.8392,     0.0000,    1.0000,    0.8235,
  0.0000,    1.0000,    0.8078,     0.0000,    1.0000,    0.7922,
  0.0000,    1.0000,    0.7765,     0.0000,    1.0000,    0.7608,
  0.0000,    1.0000,    0.7451,     0.0000,    1.0000,    0.7294,
  0.0000,    1.0000,    0.7137,     0.0000,    1.0000,    0.6980,
  0.0000,    1.0000,    0.6824,     0.0000,    1.0000,    0.6667,
  0.0000,    1.0000,    0.6510,     0.0000,    1.0000,    0.6353,
  0.0000,    1.0000,    0.6196,     0.0000,    1.0000,    0.6039,
  0.0000,    1.0000,    0.5882,     0.0000,    1.0000,    0.5725,
  0.0000,    1.0000,    0.5569,     0.0000,    1.0000,    0.5412,
  0.0000,    1.0000,    0.5255,     0.0000,    1.0000,    0.5098,
  0.0000,    1.0000,    0.4941,     0.0000,    1.0000,    0.4784,
  0.0000,    1.0000,    0.4627,     0.0000,    1.0000,    0.4471,
  0.0000,    1.0000,    0.4314,     0.0000,    1.0000,    0.4157,
  0.0000,    1.0000,    0.4000,     0.0000,    1.0000,    0.3843,
  0.0000,    1.0000,    0.3686,     0.0000,    1.0000,    0.3529,
  0.0000,    1.0000,    0.3373,     0.0000,    1.0000,    0.3216,
  0.0000,    1.0000,    0.3059,     0.0000,    1.0000,    0.2902,
  0.0000,    1.0000,    0.2745,     0.0000,    1.0000,    0.2588,
  0.0000,    1.0000,    0.2431,     0.0000,    1.0000,    0.2275,
  0.0000,    1.0000,    0.2118,     0.0000,    1.0000,    0.1961,
  0.0000,    1.0000,    0.1804,     0.0000,    1.0000,    0.1647,
  0.0000,    1.0000,    0.1490,     0.0000,    1.0000,    0.1333,
  0.0000,    1.0000,    0.1176,     0.0000,    1.0000,    0.1020,
  0.0000,    1.0000,    0.0863,     0.0000,    1.0000,    0.0706,
  0.0000,    1.0000,    0.0549,     0.0000,    1.0000,    0.0392,
  0.0000,    1.0000,    0.0235,     0.0000,    1.0000,    0.0078,
  0.0078,    1.0000,    0.0000,     0.0235,    1.0000,    0.0000,
  0.0392,    1.0000,    0.0000,     0.0549,    1.0000,    0.0000,
  0.0706,    1.0000,    0.0000,     0.0863,    1.0000,    0.0000,
  0.1020,    1.0000,    0.0000,     0.1176,    1.0000,    0.0000,
  0.1333,    1.0000,    0.0000,     0.1490,    1.0000,    0.0000,
  0.1647,    1.0000,    0.0000,     0.1804,    1.0000,    0.0000,
  0.1961,    1.0000,    0.0000,     0.2118,    1.0000,    0.0000,
  0.2275,    1.0000,    0.0000,     0.2431,    1.0000,    0.0000,
  0.2588,    1.0000,    0.0000,     0.2745,    1.0000,    0.0000,
  0.2902,    1.0000,    0.0000,     0.3059,    1.0000,    0.0000,
  0.3216,    1.0000,    0.0000,     0.3373,    1.0000,    0.0000,
  0.3529,    1.0000,    0.0000,     0.3686,    1.0000,    0.0000,
  0.3843,    1.0000,    0.0000,     0.4000,    1.0000,    0.0000,
  0.4157,    1.0000,    0.0000,     0.4314,    1.0000,    0.0000,
  0.4471,    1.0000,    0.0000,     0.4627,    1.0000,    0.0000,
  0.4784,    1.0000,    0.0000,     0.4941,    1.0000,    0.0000,
  0.5098,    1.0000,    0.0000,     0.5255,    1.0000,    0.0000,
  0.5412,    1.0000,    0.0000,     0.5569,    1.0000,    0.0000,
  0.5725,    1.0000,    0.0000,     0.5882,    1.0000,    0.0000,
  0.6039,    1.0000,    0.0000,     0.6196,    1.0000,    0.0000,
  0.6353,    1.0000,    0.0000,     0.6510,    1.0000,    0.0000,
  0.6667,    1.0000,    0.0000,     0.6824,    1.0000,    0.0000,
  0.6980,    1.0000,    0.0000,     0.7137,    1.0000,    0.0000,
  0.7294,    1.0000,    0.0000,     0.7451,    1.0000,    0.0000,
  0.7608,    1.0000,    0.0000,     0.7765,    1.0000,    0.0000,
  0.7922,    1.0000,    0.0000,     0.8078,    1.0000,    0.0000,
  0.8235,    1.0000,    0.0000,     0.8392,    1.0000,    0.0000,
  0.8549,    1.0000,    0.0000,     0.8706,    1.0000,    0.0000,
  0.8863,    1.0000,    0.0000,     0.9020,    1.0000,    0.0000,
  0.9176,    1.0000,    0.0000,     0.9333,    1.0000,    0.0000,
  0.9490,    1.0000,    0.0000,     0.9647,    1.0000,    0.0000,
  0.9804,    1.0000,    0.0000,     0.9961,    1.0000,    0.0000,
  1.0000,    0.9882,    0.0000,     1.0000,    0.9725,    0.0000,
  1.0000,    0.9569,    0.0000,     1.0000,    0.9412,    0.0000,
  1.0000,    0.9255,    0.0000,     1.0000,    0.9098,    0.0000,
  1.0000,    0.8941,    0.0000,     1.0000,    0.8784,    0.0000,
  1.0000,    0.8627,    0.0000,     1.0000,    0.8471,    0.0000,
  1.0000,    0.8314,    0.0000,     1.0000,    0.8157,    0.0000,
  1.0000,    0.8000,    0.0000,     1.0000,    0.7843,    0.0000,
  1.0000,    0.7686,    0.0000,     1.0000,    0.7529,    0.0000,
  1.0000,    0.7373,    0.0000,     1.0000,    0.7216,    0.0000,
  1.0000,    0.7059,    0.0000,     1.0000,    0.6902,    0.0000,
  1.0000,    0.6745,    0.0000,     1.0000,    0.6588,    0.0000,
  1.0000,    0.6431,    0.0000,     1.0000,    0.6275,    0.0000,
  1.0000,    0.6118,    0.0000,     1.0000,    0.5961,    0.0000,
  1.0000,    0.5804,    0.0000,     1.0000,    0.5647,    0.0000,
  1.0000,    0.5490,    0.0000,     1.0000,    0.5333,    0.0000,
  1.0000,    0.5176,    0.0000,     1.0000,    0.5020,    0.0000,
  1.0000,    0.4863,    0.0000,     1.0000,    0.4706,    0.0000,
  1.0000,    0.4549,    0.0000,     1.0000,    0.4392,    0.0000,
  1.0000,    0.4235,    0.0000,     1.0000,    0.4078,    0.0000,
  1.0000,    0.3922,    0.0000,     1.0000,    0.3765,    0.0000,
  1.0000,    0.3608,    0.0000,     1.0000,    0.3451,    0.0000,
  1.0000,    0.3294,    0.0000,     1.0000,    0.3137,    0.0000,
  1.0000,    0.2980,    0.0000,     1.0000,    0.2824,    0.0000,
  1.0000,    0.2667,    0.0000,     1.0000,    0.2510,    0.0000,
  1.0000,    0.2353,    0.0000,     1.0000,    0.2196,    0.0000,
  1.0000,    0.2039,    0.0000,     1.0000,    0.1882,    0.0000,
  1.0000,    0.1725,    0.0000,     1.0000,    0.1569,    0.0000,
  1.0000,    0.1412,    0.0000,     1.0000,    0.1255,    0.0000,
  1.0000,    0.1098,    0.0000,     1.0000,    0.0941,    0.0000,
  1.0000,    0.0784,    0.0000,     1.0000,    0.0627,    0.0000,
  1.0000,    0.0471,    0.0000,     1.0000,    0.0314,    0.0000,
  1.0000,    0.0157,    0.0000,     1.0000,    0.0000,    0.0000 };

static float lims[2]  = {-2.0, +2.0};
static int   haveDots = 0;
static char  dotName[129];

/***********************************************************************/
/*                                                                     */
/* global variables                                                    */
/*                                                                     */
/***********************************************************************/

/* global variable holding a MODL */
static void       *modl;               /* pointer to MODL */
static int        outLevel = 1;        /* default output level */
static int        batch = 0;           /* =0 to enable visualization */

/* global variables associated with graphical user interface (gui) */
static wvContext  *cntxt;              /* context for the WebViewer */
static int         port   = 7681;      /* port number */

/* global variables associated with undo */
#define MAX_UNDOS  100
static int         nundo  = 0;         /* number of undos */
static void       *undo_modl[MAX_UNDOS+1];
static char        undo_text[MAX_UNDOS+1][32];

/* global variables associated with scene graph meta-data */
#define MAX_METADATA_LEN 32000
static char        sgMetaData[MAX_METADATA_LEN];
static char        sgTempData[MAX_METADATA_LEN];
static char        sgFocusData[256];

/* global variables associated with bodyList */
#define MAX_BODYS        999
static int        nbody    = 0;        /* number of built Bodys */
static int        bodyList[MAX_BODYS]; /* array  of built Bodys */

/* global variables associated with pending errors */
static int        pendingError = 0;

/* global variables associated with the response buffer */
static int        max_resp_len = 0;
static int        response_len = 0;
static char       *response = NULL;

/* global variables associated with journals */
static FILE       *jrnl_out=NULL;    /* output journal file */
static char       filename[257];

/***********************************************************************/
/*                                                                     */
/* declarations                                                        */
/*                                                                     */
/***********************************************************************/

/* declarations for high-level routines defined below */
#if   defined(GEOM_CAPRI)
       int        CAPRI_MAIN(int argc, char *argv[]);
#elif defined(GEOM_EGADS)
#endif

/* declarations for support routines defined below */
static int        buildBodys(int buildTo, int *builtTo, int *buildStatus);
static int        storeUndo(char *cmd, char *arg);

/* declarations for graphic routines defined below */
static int        buildSceneGraph();
       void       browserMessage(void *wsi, char *text, /*@unused@*/ int lena);
static void       processMessage(char *text);
static void       addToResponse(char *text);
static int        getToken(char *text, int nskip, char *token);
static void       spec_col(float scalar, float lims[], float out[]);

/* external functions not declared in an .h file */
#if   defined(GEOM_CAPRI)
    extern void   *gi_alloc(int nbtyes);
    extern void   *gi_reall(void *old, int nbytes);
    extern void   gi_free(void *ptr);
#elif defined(GEOM_EGADS)
#endif


/***********************************************************************/
/*                                                                     */
/*   CAPRI_MAIN - main program                                         */
/*                                                                     */
/***********************************************************************/

#if   defined(GEOM_CAPRI)
    int
    CAPRI_MAIN(int       argc,          /* (in)  number of arguments */
               char      *argv[])       /* (in)  array of arguments */
#elif defined(GEOM_EGADS)
    int
    main(int       argc,                /* (in)  number of arguments */
         char      *argv[])             /* (in)  array of arguments */
#endif
{

    int       imajor, iminor, status, i, bias, showUsage=0;
    int       builtTo, buildStatus;
    float     fov, zNear, zFar;
    float     eye[3]    = {0.0, 0.0, 7.0};
    float     center[3] = {0.0, 0.0, 0.0};
    float     up[3]     = {0.0, 1.0, 0.0};
    char      casename[257], jrnlname[257], tempname[257];
    char      text[MAX_STR_LEN], *wv_start;
    void      *temp;
    FILE      *jrnl_in=NULL;
    modl_T    *MODL;

    #if   defined(GEOM_CAPRI)
        double    box[6];
    #elif defined(GEOM_EGADS)
        int       jbody, ibody;
    #endif

    clock_t   old_time, new_time;

    ROUTINE(MAIN);

    /* --------------------------------------------------------------- */

    DPRINT0("starting serveCSM");

    /* get the flags and casename(s) from the command line */
    casename[0] = '\0';
    jrnlname[0] = '\0';

    for (i = 1; i < argc; i++) {
        if        (strcmp(argv[i], "-port") == 0) {
            if (i < argc-1) {
                sscanf(argv[++i], "%d", &port);
            } else {
                showUsage = 1;
                break;
            }
        } else if (strcmp(argv[i], "-jrnl") == 0) {
            if (i < argc-1) {
                STRNCPY(jrnlname, argv[++i], 257);
            } else {
                showUsage = 1;
                break;
            }
        } else if (strcmp(argv[i], "-outLevel") == 0) {
            if (i < argc-1) {
                sscanf(argv[++i], "%d", &outLevel);
                if (outLevel < 0) outLevel = 0;
                if (outLevel > 3) outLevel = 3;
            } else {
                showUsage = 1;
                break;
            }
        } else if (strcmp(argv[i], "-batch") == 0) {
            batch = 1;
        } else if (strlen(casename) == 0) {
            STRNCPY(casename, argv[i], 257);
        } else {
            SPRINT0(0, "two casenames given");
            showUsage = 1;
            break;
        }
    }

    if (showUsage) {
        SPRINT0(0, "proper usage: 'serveCSM [-port X] [-jrnl jrnlname] [-outLevel X] [-batch] [casename[.csm]]'");
        SPRINT0(0, "STOPPING...\a");
        exit(0);
    }

    /* welcome banner */
    (void) ocsmVersion(&imajor, &iminor);

    SPRINT0(1, "**********************************************************");
    SPRINT0(1, "*                                                        *");
    SPRINT0(1, "*                    Program serveCSM                    *");
    SPRINT2(1, "*                     version %2d.%02d                      *", imajor, iminor);
    SPRINT0(1, "*                                                        *");
    SPRINT0(1, "*        written by John Dannenhoffer, 2010/2014         *");
    SPRINT0(1, "*                                                        *");
    SPRINT0(1, "**********************************************************");

    /* set OCSMs output level */
    (void) ocsmSetOutLevel(outLevel);

    /* allocate nominal response buffer */
    max_resp_len = 4096;
    response = (char*) malloc(max_resp_len*sizeof(char));
    if (response == NULL) {
        SPRINT0(0, "ERROR:: error mallocing response");
        exit(0);
    }

    /* strip off .csm (which is assumed to be at the end) if present */
    if (strlen(casename) > 0) {
        STRNCPY(filename, casename, 257);
        if (strstr(casename, ".csm") == NULL) {
            strcat(filename, ".csm");
        }
    } else {
        casename[0] = '\0';
        filename[0] = '\0';
    }

    #if   defined(GEOM_CAPRI)
        /* start CAPRI */
        gi_uRegister();

        status = gi_uStart();
        SPRINT1(1, "--> gi_uStart() -> status=%d", status);

        if (status < SUCCESS) {
            SPRINT0(0, "ERROR:: problem detected while starting CAPRI");
            exit(0);
        }
    #elif defined(GEOM_EGADS)
    #endif

    #if   defined(GEOM_CAPRI)
        /* make a "throw-away" volume so that CAPRI's startup
           message does not get produced during code below */
        box[0] = 0.0; box[1] = 0.0; box[2] = 0.0;
        box[3] = 1.0; box[4] = 1.0; box[5] = 1.0;
        status = gi_gCreateVolume(NULL, "Parasolid", 1, box);
        SPRINT1(1, "--> gi_gCreateVolume(dummy) -> status=%d", status);
    #elif defined(GEOM_EGADS)
    #endif

    /* read the .csm file and create the MODL */
    old_time = clock();
    status   = ocsmLoad(filename, &modl);
    new_time = clock();
    SPRINT3(1, "--> ocsmLoad(%s) -> status=%d (%s)", filename, status, ocsmGetText(status));
    SPRINT1(1, "==> ocsmLoad CPUtime=%9.3f sec",
            (double)(new_time-old_time) / (double)(CLOCKS_PER_SEC));

    if (status < SUCCESS) {
        SPRINT0(0, "ERROR:: problem in ocsmLoad");
        exit(0);
    } else {
        MODL = (modl_T*)modl;
    }


    /* check that Branches are properly ordered */
    old_time = clock();
    status   = ocsmCheck(modl);
    new_time = clock();
    SPRINT2(0, "--> ocsmCheck() -> status=%d (%s)",
            status, ocsmGetText(status));
    SPRINT1(0, "==> ocsmCheck CPUtime=%10.3f sec",
            (double)(new_time-old_time) / (double)(CLOCKS_PER_SEC));

    if (status < SUCCESS) {
        SPRINT0(0, "ERROR:: problem in ocsmCheck");
        exit(0);
    }

    /* print out the global Attributes, Parameters, and Branches */
    if (batch == 1) {
        SPRINT0(1, "External Parameter(s):");
        if (outLevel > 0) {
            status = ocsmPrintPmtrs(modl, stdout);
        }

        SPRINT0(1, "Branch(es):");
        if (outLevel > 0) {
            status = ocsmPrintBrchs(modl, stdout);
        }

        SPRINT0(1, "Global Attribute(s):");
        if (outLevel > 0) {
            status = ocsmPrintAttrs(modl, stdout);
        }
    }

    /* open the output journal file */
    snprintf(tempname, 257, "port%d.jrnl", port);
    jrnl_out = fopen(tempname, "w");

    /* initialize the scene graph meta data */
    if (batch == 0) {
        sgMetaData[ 0] = '\0';
        sgFocusData[0] = '\0';
    }

    /* create the WebViewer context */
    if (batch == 0) {
        bias  =  1;
        fov   = 30.0;
        zNear =  1.0;
        zFar  = 10.0;
        cntxt = wv_createContext(bias, fov, zNear, zFar, eye, center, up);
        if (cntxt == NULL) {
            SPRINT0(0, "ERROR:: failed to create wvContext");
            exit(0);
        }
    }

    /* build the Bodys from the MODL */
    status = buildBodys(0, &builtTo, &buildStatus);

    if (builtTo < 0) {
        SPRINT2(0, "build() detected \"%s\" at %s",
                ocsmGetText(buildStatus), MODL->brch[1-builtTo].name);
        SPRINT0(0, "Configuration only built up to detected error\a");
        pendingError = -builtTo;
    } else if (status != SUCCESS) {
        SPRINT1(0, "ERROR:: build() detected \"%s\"",
                ocsmGetText(buildStatus));
        exit(0);
    }

    /* print out the Bodys */
    if (batch == 1) {
        SPRINT0(1, "Body(s):");
        if (outLevel > 0) {
            status = ocsmPrintBodys(modl, stdout);
        }
    }

    /* process the input journal file if jrnlname exists */
    if (strlen(jrnlname) > 0) {
        SPRINT1(0, "\n==> Opening input journal file \"%s\"\n", jrnlname);

        jrnl_in = fopen(jrnlname, "r");
        if (jrnl_in == NULL) {
            SPRINT0(0, "ERROR:: Journal file cannot be opened");
            exit(0);
        } else {
            while (1) {
                temp = fgets(text, MAX_STR_LEN, jrnl_in);
                if (temp == NULL) break;

                if (feof(jrnl_in)) break;

                text[strlen(text)-1] = '\0';

                processMessage(text);
            }

            fclose(jrnl_in);

            SPRINT0(0, "\n==> Closing input journal file\n");
        }
    }

    /* get the command to start the client (if any) */
    if (batch == 0) {
        wv_start = getenv("WV_START");
    }

    /* start the server */
    if (batch == 0) {
        status = SUCCESS;
        if (wv_startServer(port, NULL, NULL, NULL, 0, cntxt) == 0) {

            /* stay alive a long as we have a client */
            while (wv_statusServer(0)) {
                usleep(500000);

                /* start the browser if the first time through this loop */
                if (status == SUCCESS) {
                    if (wv_start != NULL) {
                        status += system(wv_start);
                    }

                    status++;
                }
            }
        }
    }

    /* cleanup and exit */
    if (jrnl_out != NULL) fclose(jrnl_out);
    jrnl_out = NULL;

    MODL = (modl_T*)modl;

    #if   defined(GEOM_CAPRI)
    #elif defined(GEOM_EGADS)
        for (jbody = 0; jbody < nbody; jbody++) {
            ibody = bodyList[jbody];

            if (MODL->body[ibody].etess != NULL) {
                status = EG_deleteObject(MODL->body[ibody].etess);
                SPRINT2(2, "--> EG_deleteObject(etess[%d]) -> status=%d", ibody, status);
            }

            if (MODL->body[ibody].ebody != NULL) {
                status = EG_deleteObject(MODL->body[ibody].ebody);
                SPRINT2(2, "--> EG_deleteObject(ebody[%d]) => status=%d", ibody, status);
            }
        }

        if (MODL->context != NULL) {
            status = EG_setOutLevel(MODL->context, 0);
            status = EG_close(MODL->context);
            SPRINT1(1, "--> EG_close() -> status=%d", status);

            /* closing the context automatically removed all ebody and etess */
            for (ibody = 1; ibody <= MODL->nbody; ibody++) {
                MODL->body[ibody].ebody = NULL;
                MODL->body[ibody].etess = NULL;
            }
        }
    #endif

    status = ocsmFree(modl);
    SPRINT2(1, "--> ocsmFree() -> status=%d (%s)", status, ocsmGetText(status));

    status = ocsmFree(NULL);
    SPRINT2(1, "--> ocsmFree(NULL) -> status=%d (%s)", status, ocsmGetText(status));

    free(response);
    response = NULL;

    wv_cleanupServers();
    status = SUCCESS;

    SPRINT0(0, "==> serveCSM completed successfully");

//cleanup:
    return status;
}


/***********************************************************************/
/*                                                                     */
/*   buildBodys - build Bodys and update scene graph                   */
/*                                                                     */
/***********************************************************************/

static int
buildBodys(int     buildTo,             /* (in)  last Branch to execute */
           int     *builtTo,            /* (out) last Branch successfully executed */
           int     *buildStatus)        /* (out) status returned from build */
{
    int       status = SUCCESS;         /* return status */

    modl_T    *MODL = (modl_T*)modl;

    int       ibody, jbody, status_min;
    double    box[6], size, params[3];
    clock_t   old_time, new_time;

    #if   defined(GEOM_CAPRI)
        int       ivol;
    #elif defined(GEOM_EGADS)
        ego       ebody;
    #endif

    ROUTINE(buildBodys);

    /* --------------------------------------------------------------- */

    /* if there are no Branches, simply return */
    if (MODL->nbrch <= 0) {
        SPRINT0(1, "--> No Branches, so skipping build");
        *builtTo     = 0;
        *buildStatus = SUCCESS;

    } else {
        /* check that Branches are properly ordered */
        old_time = clock();
        status   = ocsmCheck(modl);
        new_time = clock();
        SPRINT2(0, "--> ocsmCheck() -> status=%d (%s)",
                status, ocsmGetText(status));
        SPRINT1(0, "==> ocsmCheck CPUtime=%10.3f sec",
                (double)(new_time-old_time) / (double)(CLOCKS_PER_SEC));

        if (status < SUCCESS) {
            goto cleanup;
        }

        /* build the Bodys */
        nbody    = MAX_BODYS;
        old_time = clock();
        status   = ocsmBuild(modl, buildTo, builtTo, &nbody, bodyList);
        new_time = clock();
        SPRINT5(0, "--> ocsmBuild(buildTo=%d) -> status=%d (%s), builtTo=%d, nbody=%d",
                buildTo, status, ocsmGetText(status), *builtTo, nbody);
        SPRINT1(0, "==> ocsmBuild CPUtime=%10.3f sec",
                (double)(new_time-old_time) / (double)(CLOCKS_PER_SEC));

        if (batch == 1 && status < SUCCESS) {
            SPRINT1(0, "ERROR:: build not completed because error %d was detected", status);
            exit(0);
        }

        *buildStatus = status;

        /* tessellate the Bodys */
        old_time = clock();
        status_min = SUCCESS;
        for (jbody = 0; jbody < nbody; jbody++) {
            ibody = bodyList[jbody];

            /* if we already have an etess, no need to regenerate it here */
            #if   defined(GEOM_CAPRI)
                ivol   = MODL->body[ibody].ivol;
                if (ivol < 0) continue;
                status = gi_dBox(ivol, box);
            #elif defined(GEOM_EGADS)
                if (MODL->body[ibody].etess != NULL) {
                    SPRINT1(0, "--> reusing TessBody(ibody=%d)", ibody);
                    continue;
                }

                ebody  = MODL->body[ibody].ebody;
                status = EG_getBoundingBox(ebody, box);
            #endif

            size = sqrt(SQR(box[3]-box[0]) + SQR(box[4]-box[1]) + SQR(box[5]-box[2]));

            /* vTess parameters */
            params[0] = 0.0250 * size;
            params[1] = 0.0010 * size;
            params[2] = 15.0;

            #if   defined(GEOM_CAPRI)
                SPRINT0(0, "--> default tessellation used");
            #elif defined(GEOM_EGADS)
                ebody = MODL->body[ibody].ebody;

                if (MODL->context != NULL) {
                    (void) EG_setOutLevel(MODL->context, 1);
                    status = EG_makeTessBody(ebody, params, &(MODL->body[ibody].etess));
                    (void) EG_setOutLevel(MODL->context, outLevel);
                    SPRINT6(0, "--> EG_makeTessBody(ibody=%4d, params=%10.5f, %10.5f, %10.5f) -> status=%d (%s)",
                            ibody, params[0], params[1], params[2], status, ocsmGetText(status));

                    status_min = MIN(status_min, status);
                }
            #endif
        }

        new_time = clock();
        SPRINT1(0, "==> EG_makeTessBody CPUtime=%10.3f sec",
                (double)(new_time-old_time) / (double)(CLOCKS_PER_SEC));

        status = status_min;
        if (status < SUCCESS) {
            goto cleanup;
        }
    }

    /* build the scene graph */
    if (batch == 0) {
        buildSceneGraph();
    }

cleanup:
    return status;
}


/***********************************************************************/
/*                                                                     */
/*   storeUndo - store an undo for the current command                 */
/*                                                                     */
/***********************************************************************/

static int
storeUndo(char   *cmd,                  /* (in)  current command */
          char   *arg)                  /* (in)  current argument */
{
    int       status = SUCCESS;         /* return status */

    int       iundo;
    char      text[MAX_EXPR_LEN];

    modl_T    *MODL = (modl_T*)modl;

    ROUTINE(storeUndo);

    /* --------------------------------------------------------------- */

    /* if the undos are full, discard the most ancient one */
    if (nundo >= MAX_UNDOS) {
        status = ocsmFree(undo_modl[0]);
        CHECK_STATUS(ocsmFree);

        for (iundo = 0; iundo < nundo; iundo++) {
            undo_modl[iundo] = undo_modl[iundo+1];
            (void) STRNCPY(undo_text[iundo], undo_text[iundo+1], 32);
        }

        nundo--;
    }

    /* store an undo snapshot */
    snprintf(text, MAX_EXPR_LEN, "%s %s", cmd, arg);
    STRNCPY(undo_text[nundo], text, 31);

    status = ocsmCopy(MODL, &undo_modl[nundo]);
    CHECK_STATUS(ocsmCopy);

    nundo++;

    SPRINT2(1, "~~> ocsmCopy() -> status=%d  (nundo=%d)",
            status, nundo);

cleanup:
    return status;
}


/***********************************************************************/
/*                                                                     */
/*   buildSceneGraph - make a scene graph for wv                       */
/*                                                                     */
/***********************************************************************/

static int
buildSceneGraph()
{
    int       status = SUCCESS;

    int       ibody, jbody, iface, iedge, iattr, nattr, ipmtr, irc;
    int       npnt, ipnt, ntri, itri, igprim, nseg, k, attrs, head;
    int       *segs=NULL, *ivrts=NULL;
    CINT      *ptype, *pindx, *tris, *tric;
    float     focus[4], color[3], *pcolors;
    double    bigbox[6], box[6], size, xyz_dum[3], *vel;
    CDOUBLE   *xyz, *uv, *t;
    char      gpname[33];

    #if   defined(GEOM_CAPRI)
        int       ivol;
    #elif defined(GEOM_EGADS)
        ego       ebody, etess, eface, eedge;
        int       i, nlist, itype;
        CINT      *tempIlist;
        CDOUBLE   *tempRlist;
        CCHAR     *attrName, *tempClist;
    #endif

    wvData    items[5];

    modl_T    *MODL = (modl_T*)modl;

    ROUTINE(buildSceneGraph);

    /* --------------------------------------------------------------- */

    /* remove any graphic primitives that already exist */
    wv_removeAll(cntxt);

    /* find the values needed to adjust the vertices */
    bigbox[0] = bigbox[1] = bigbox[2] = +HUGEQ;
    bigbox[3] = bigbox[4] = bigbox[5] = -HUGEQ;

    for (jbody = 0; jbody < nbody; jbody++) {
        ibody = bodyList[jbody];

        #if   defined(GEOM_CAPRI)
            ivol  = MODL->body[ibody].ivol;
            status = gi_dBox(ivol, box);
        #elif defined(GEOM_EGADS)
            ebody = MODL->body[ibody].ebody;
            etess = MODL->body[ibody].etess;

            status = EG_getBoundingBox(ebody, box);
        #endif

        if (box[0] < bigbox[0]) bigbox[0] = box[0];
        if (box[1] < bigbox[1]) bigbox[1] = box[1];
        if (box[2] < bigbox[2]) bigbox[2] = box[2];
        if (box[3] > bigbox[3]) bigbox[3] = box[3];
        if (box[4] > bigbox[4]) bigbox[4] = box[4];
        if (box[5] > bigbox[5]) bigbox[5] = box[5];
    }

                                    size = bigbox[3] - bigbox[0];
    if (size < bigbox[4]-bigbox[1]) size = bigbox[4] - bigbox[1];
    if (size < bigbox[5]-bigbox[2]) size = bigbox[5] - bigbox[2];

    focus[0] = (bigbox[0] + bigbox[3]) / 2;
    focus[1] = (bigbox[1] + bigbox[4]) / 2;
    focus[2] = (bigbox[2] + bigbox[5]) / 2;
    focus[3] = size;

    /* generate the scene graph focus data */
    snprintf(sgFocusData, 256, "sgFocus;[%20.12e,%20.12e,%20.12e,%20.12e]", focus[0], focus[1], focus[2], focus[3]);

    /* initialize the scene graph meta data */
    STRNCPY(sgMetaData, "sgData;{", MAX_METADATA_LEN);

    /* special scene graph if nbody<=0 */
    if (nbody <= 0) {
        snprintf(gpname, 33, "Body 0 Face 0");
        attrs = WV_ON;

        xyz_dum[0] = 0;
        xyz_dum[1] = 0;
        xyz_dum[2] = 0;

        status = wv_setData(WV_REAL64, 1, (void*)xyz_dum, WV_VERTICES, &(items[0]));
        wv_adjustVerts(&(items[0]), focus);

        color[0] = 0;
        color[1] = 0;
        color[2] = 0;
        status = wv_setData(WV_REAL32, 1, (void*)color, WV_PCOLOR, &(items[1]));

        igprim = wv_addGPrim(cntxt, gpname, WV_LINE, attrs, 2, items);
    }

    /* loop through the Bodys */
    for (jbody = 0; jbody < nbody; jbody++) {
        ibody = bodyList[jbody];

        #if   defined(GEOM_CAPRI)
            ivol  = MODL->body[ibody].ivol;
        #elif defined(GEOM_EGADS)
            ebody = MODL->body[ibody].ebody;
            etess = MODL->body[ibody].etess;
        #endif

        /* determine if any of the external Parameters have a velocity */
        haveDots   = 0;
        dotName[0] = '\0';

        for (ipmtr = 1; ipmtr <= MODL->npmtr; ipmtr++) {
            if (MODL->pmtr[ipmtr].type == OCSM_EXTERNAL) {
                for (irc = 0; irc < (MODL->pmtr[ipmtr].nrow)*(MODL->pmtr[ipmtr].ncol); irc++) {
                    if (MODL->pmtr[ipmtr].dot[irc] != 0) {
                        if (fabs(MODL->pmtr[ipmtr].dot[irc]-1) < EPS06) {
                            haveDots++;
                            sprintf(dotName, "d(norm)/d(%s)", MODL->pmtr[ipmtr].name);
                        } else if (MODL->pmtr[ipmtr].dot[irc] != 0) {
                            haveDots++;
                            sprintf(dotName, "d(norm)/d(***)");
                        }
                    }
                }
            }
        }

        /* loop through the Faces within each Body */
        for (iface = 1; iface <= MODL->body[ibody].nface; iface++) {
            #if   defined(GEOM_CAPRI)
                status = gi_dTesselFace(ivol, iface, &ntri, &tris, &tric,
                                        &npnt, &xyz, &ptype, &pindx, &uv);
            #elif defined(GEOM_EGADS)
                status = EG_getTessFace(etess, iface,
                                        &npnt, &xyz, &uv, &ptype, &pindx,
                                        &ntri, &tris, &tric);
            #endif

            /* name and attributes */
            snprintf(gpname, 33, "Body %d Face %d", ibody, iface);
            if (haveDots >= 1) {
                attrs = WV_ON | WV_SHADING | WV_ORIENTATION;
            } else {
                attrs = WV_ON              | WV_ORIENTATION;
            }

            /* vertices */
            status = wv_setData(WV_REAL64, npnt, (void*)xyz, WV_VERTICES, &(items[0]));
            wv_adjustVerts(&(items[0]), focus);

            /* triangles */
            status = wv_setData(WV_INT32, 3*ntri, (void*)tris, WV_INDICES, &(items[1]));

            /* smooth triangle colors */
            if (haveDots >= 1) {
                vel = (double *) malloc(npnt * sizeof(double));
                if (vel == NULL) goto cleanup;

                pcolors = (float *) malloc(3*npnt*sizeof(float));
                if (pcolors == NULL) goto cleanup;

                status = ocsmGetVel(MODL, ibody, OCSM_FACE, iface, npnt, NULL, vel);
                if (status != SUCCESS) {
                    pendingError = -1;
                }

                for (ipnt = 0; ipnt < npnt; ipnt++) {
                    spec_col((float)(vel[ipnt]), lims, &(pcolors[3*ipnt]));
                }
                status = wv_setData(WV_REAL32, npnt, (void*)pcolors, WV_COLORS, &(items[2]));

                free(pcolors);
                free(vel);

            /* constant triangle colors */
            } else {
                color[0] = RED(  MODL->body[ibody].face[iface].gratt.color);
                color[1] = GREEN(MODL->body[ibody].face[iface].gratt.color);
                color[2] = BLUE( MODL->body[ibody].face[iface].gratt.color);
                status = wv_setData(WV_REAL32, 1, (void*)color, WV_COLORS, &(items[2]));
            }

            /* triangle sides (segments) */
            nseg = 0;
            for (itri = 0; itri < ntri; itri++) {
                for (k = 0; k < 3; k++) {
                    if (tric[3*itri+k] < itri+1) {
                        nseg++;
                    }
                }
            }

            segs = (int*) malloc(2*nseg*sizeof(int));
            if (segs == NULL) goto cleanup;

            nseg = 0;
            for (itri = 0; itri < ntri; itri++) {
                for (k = 0; k < 3; k++) {
                    if (tric[3*itri+k] < itri+1) {
                        segs[2*nseg  ] = tris[3*itri+(k+1)%3];
                        segs[2*nseg+1] = tris[3*itri+(k+2)%3];
                        nseg++;
                    }
                }
            }

            status = wv_setData(WV_INT32, 2*nseg, (void*)segs, WV_LINDICES, &(items[3]));

            free(segs);

            /* segment colors */
            color[0] = 0.0;
            color[1] = 0.0;
            color[2] = 0.0;
            status = wv_setData(WV_REAL32, 1, (void*)color, WV_LCOLOR, &(items[4]));

            /* make graphic primitive */
            igprim = wv_addGPrim(cntxt, gpname, WV_TRIANGLE, attrs, 5, items);
            if (igprim >= 0) {

                /* make line width 1 */
                cntxt->gPrims[igprim].lWidth = 1.0;
            }

            #if   defined(GEOM_CAPRI)
                nattr = 0;
            #elif defined(GEOM_EGADS)
                eface  = MODL->body[ibody].face[iface].eface;
                status = EG_attributeNum(eface, &nattr);
            #endif

            /* add Face to meta data (if there is room) */
            if (strlen(sgMetaData) < MAX_METADATA_LEN-1000) {
                if (nattr > 0) {
                    STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                    snprintf(sgMetaData, MAX_METADATA_LEN, "%s\"%s\":[",  sgTempData, gpname);
                } else {
                    STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                    snprintf(sgMetaData, MAX_METADATA_LEN, "%s\"%s\":[]", sgTempData, gpname);
                }

                for (iattr = 1; iattr <= nattr; iattr++) {
                    #if   defined(GEOM_CAPRI)
                    #elif defined(GEOM_EGADS)
                        status = EG_attributeGet(eface, iattr, &attrName, &itype, &nlist,
                                                 &tempIlist, &tempRlist, &tempClist);

                        STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                        snprintf(sgMetaData, MAX_METADATA_LEN, "%s\"%s\",\"", sgTempData, attrName);

                        if        (itype == ATTRINT) {
                            for (i = 0; i < nlist ; i++) {
                                STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                                snprintf(sgMetaData, MAX_METADATA_LEN, "%s %d", sgTempData, tempIlist[i]);
                            }
                        } else if (itype == ATTRREAL) {
                            for (i = 0; i < nlist ; i++) {
                                STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                                snprintf(sgMetaData, MAX_METADATA_LEN, "%s %f", sgTempData, tempRlist[i]);
                            }
                        } else if (itype == ATTRSTRING) {
                            STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                            snprintf(sgMetaData, MAX_METADATA_LEN, "%s %s ", sgTempData, tempClist);
                        }

                        STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                        snprintf(sgMetaData, MAX_METADATA_LEN, "%s\",", sgTempData);
                    #endif
                }
                sgMetaData[strlen(sgMetaData)-1] = '\0';
                STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                snprintf(sgMetaData, MAX_METADATA_LEN, "%s],", sgTempData);
            }
        }

        /* loop through the Edges within each Body */
        for (iedge = 1; iedge <= MODL->body[ibody].nedge; iedge++) {
            #if   defined(GEOM_CAPRI)
                status = gi_dTesselEdge(ivol, iedge, &npnt, &xyz, &t);
            #elif defined(GEOM_EGADS)
                status = EG_getTessEdge(etess, iedge, &npnt, &xyz, &t);
            #endif

            /* name and attributes */
            snprintf(gpname, 33, "Body %d Edge %d", ibody, iedge);
            attrs = WV_ON;

            /* vertices */
            status = wv_setData(WV_REAL64, npnt, (void*)xyz, WV_VERTICES, &(items[0]));
            wv_adjustVerts(&(items[0]), focus);

            /* segments */
            ivrts = (int*) malloc(2*(npnt-1)*sizeof(int));
            if (ivrts == NULL) goto cleanup;

            for (ipnt = 0; ipnt < npnt-1; ipnt++) {
                ivrts[2*ipnt  ] = ipnt + 1;
                ivrts[2*ipnt+1] = ipnt + 2;
            }

            status = wv_setData(WV_INT32, 2*(npnt-1), (void*)ivrts, WV_INDICES, &(items[1]));

            free(ivrts);

            /* line colors */
            color[0] = RED(  MODL->body[ibody].edge[iedge].gratt.color);
            color[1] = GREEN(MODL->body[ibody].edge[iedge].gratt.color);
            color[2] = BLUE( MODL->body[ibody].edge[iedge].gratt.color);
            status = wv_setData(WV_REAL32, 1, (void*)color, WV_COLORS, &(items[2]));

            /* points */
            ivrts = (int*) malloc(npnt*sizeof(int));
            if (ivrts == NULL) goto cleanup;

            for (ipnt = 0; ipnt < npnt; ipnt++) {
                ivrts[ipnt] = ipnt + 1;
            }

            status = wv_setData(WV_INT32, npnt, (void*)ivrts, WV_PINDICES, &(items[3]));

            free(ivrts);

            /* point colors */
            color[0] = 0;
            color[1] = 0;
            color[2] = 0;
            status = wv_setData(WV_REAL32, 1, (void*)color, WV_PCOLOR, &(items[4]));

            /* make graphic primitive */
            igprim = wv_addGPrim(cntxt, gpname, WV_LINE, attrs, 5, items);
            if (igprim >= 0) {

                /* make line width 2 */
                cntxt->gPrims[igprim].lWidth = 2.0;

                /* make point size 5 */
                cntxt->gPrims[igprim].pSize  = 5.0;

                /* add arrow heads */
                head = npnt - 1;
                status = wv_addArrowHeads(cntxt, igprim, 0.05, 1, &head);
            }

            #if   defined(GEOM_CAPRI)
                nattr = 0;
            #elif defined(GEOM_EGADS)
                eedge  = MODL->body[ibody].edge[iedge].eedge;
                status = EG_attributeNum(eedge, &nattr);
            #endif

            /* add Edge to meta data (if there is room) */
            if (strlen(sgMetaData) < MAX_METADATA_LEN-1000) {
                if (nattr > 0) {
                    STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                    snprintf(sgMetaData, MAX_METADATA_LEN, "%s\"%s\":[",  sgTempData, gpname);
                } else {
                    STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                    snprintf(sgMetaData, MAX_METADATA_LEN, "%s\"%s\":[]", sgTempData, gpname);
                }

                for (iattr = 1; iattr <= nattr; iattr++) {
                    #if   defined(GEOM_CAPRI)
                    #elif defined(GEOM_EGADS)
                        status = EG_attributeGet(eedge, iattr, &attrName, &itype, &nlist,
                                                 &tempIlist, &tempRlist, &tempClist);

                        STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                        snprintf(sgMetaData, MAX_METADATA_LEN, "%s\"%s\",\"", sgTempData, attrName);

                        if        (itype == ATTRINT) {
                            for (i = 0; i < nlist ; i++) {
                                STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                                snprintf(sgMetaData, MAX_METADATA_LEN, "%s %d", sgTempData, tempIlist[i]);
                            }
                        } else if (itype == ATTRREAL) {
                            for (i = 0; i < nlist ; i++) {
                                STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                                snprintf(sgMetaData, MAX_METADATA_LEN, "%s %f", sgTempData, tempRlist[i]);
                            }
                        } else if (itype == ATTRSTRING) {
                            STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                            snprintf(sgMetaData, MAX_METADATA_LEN, "%s %s ", sgTempData, tempClist);
                        }

                        STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                        snprintf(sgMetaData, MAX_METADATA_LEN, "%s\",", sgTempData);
                    #endif
                }
                sgMetaData[strlen(sgMetaData)-1] = '\0';
                STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                snprintf(sgMetaData, MAX_METADATA_LEN, "%s],", sgTempData);
            }
        }
    }

    /* finish the scene graph meta data */
    sgMetaData[strlen(sgMetaData)-1] = '\0';
    STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
    snprintf(sgMetaData, MAX_METADATA_LEN, "%s}", sgTempData);

cleanup:
    return status;
}


/***********************************************************************/
/*                                                                     */
/*   browserMessage - called when client sends a message to the server */
/*                                                                     */
/***********************************************************************/

void
browserMessage(void    *wsi,
               char    *text,
  /*@unused@*/ int     lena)
{
    int       status;

    modl_T    *MODL;

    ROUTINE(browserMessage);

    /* --------------------------------------------------------------- */

    /* process the Message */
    processMessage(text);

    /* send the response */
    SPRINT1(2, "response-> %s", response);
    wv_sendText(wsi, response);

    /* send the scene graph meta data if it has not already been sent */
    if (strlen(sgMetaData) > 0) {
        SPRINT1(2, "sgData-> %s", sgMetaData);
        wv_sendText(wsi, sgMetaData);

        /* nullify meta data so that it does not get sent again */
        sgMetaData[0] = '\0';
    }

    if (strlen(sgFocusData) > 0) {
        SPRINT1(2, "sgFocus-> %s", sgFocusData);
        wv_sendText(wsi, sgFocusData);

        /* nullify meta data so that it does not get sent again */
        sgFocusData[0] = '\0';
    }

    /* either open or close the key */
    if        (haveDots >  1) {
        status = wv_setKey(cntxt, 256, color_map, lims[0], lims[1], "d(norm)/d(***)");
    } else if (haveDots == 1) {
        status = wv_setKey(cntxt, 256, color_map, lims[0], lims[1], dotName         );
    } else {
        status = wv_setKey(cntxt,   0, NULL,      lims[0], lims[1], NULL            );
    }
    if (status != SUCCESS) {
        SPRINT1(0, "wv_setKet -> status=%d", status);
    }

    /* send an error message if one is pending */
    if (pendingError > 0) {

        MODL = (modl_T*)modl;

        snprintf(response, max_resp_len, "ERROR:: initial build() detected error after branch %s",
                 MODL->brch[pendingError].name);

        pendingError = 0;

        SPRINT1(2, "response-> %s", response);
        wv_sendText(wsi, response);

    } else if (pendingError == -1) {

        MODL = (modl_T*)modl;

        snprintf(response, max_resp_len, "ERROR:: could not find Design Velocities; shown as zeros");

        pendingError = 0;

        SPRINT1(2, "response-> %s", response);
        wv_sendText(wsi, response);

    }

}


/***********************************************************************/
/*                                                                     */
/*   processMessage - process the message and create the response      */
/*                                                                     */
/***********************************************************************/

static void
processMessage(char    *text)
{
    int       i, status, ibrch, itype, builtTo, buildStatus, ichar;
    int       ipmtr, jpmtr, nrow, ncol, irow, icol, index, iattr, actv, icount;
    char      *pEnd;

    char      *name=NULL, *type=NULL, *valu=NULL;
    char      *arg1=NULL, *arg2=NULL, *arg3=NULL;
    char      *arg4=NULL, *arg5=NULL, *arg6=NULL;
    char      *arg7=NULL, *arg8=NULL, *arg9=NULL;
    char      *entry=NULL;

    static FILE  *fp=NULL;

    modl_T    *MODL = (modl_T*)modl;

    ROUTINE(processMessage);

    /* --------------------------------------------------------------- */

    name  = (char *) malloc(MAX_EXPR_LEN*sizeof(char));
    type  = (char *) malloc(MAX_EXPR_LEN*sizeof(char));
    valu  = (char *) malloc(MAX_EXPR_LEN*sizeof(char));
    arg1  = (char *) malloc(MAX_EXPR_LEN*sizeof(char));
    arg2  = (char *) malloc(MAX_EXPR_LEN*sizeof(char));
    arg3  = (char *) malloc(MAX_EXPR_LEN*sizeof(char));
    arg4  = (char *) malloc(MAX_EXPR_LEN*sizeof(char));
    arg5  = (char *) malloc(MAX_EXPR_LEN*sizeof(char));
    arg6  = (char *) malloc(MAX_EXPR_LEN*sizeof(char));
    arg7  = (char *) malloc(MAX_EXPR_LEN*sizeof(char));
    arg8  = (char *) malloc(MAX_EXPR_LEN*sizeof(char));
    arg9  = (char *) malloc(MAX_EXPR_LEN*sizeof(char));
    entry = (char *) malloc(MAX_STR_LEN*sizeof(char));

    if (name == NULL || type == NULL || valu == NULL ||
        arg1 == NULL || arg2 == NULL || arg3 == NULL ||
        arg4 == NULL || arg5 == NULL || arg6 == NULL ||
        arg7 == NULL || arg8 == NULL || arg9 == NULL || entry == NULL) {
        SPRINT0(0, "ERROR:: error mallocing arrays in processMessage");
        exit(0);
    }

    SPRINT1(1, "==> processMessage(text=%s)", text);

    /* initialize the response */
    response_len = 0;
    response[0] = '\0';

    /* NO-OP */
    if (strlen(text) == 0) {

    /* "identify;" */
    } else if (strncmp(text, "identify;", 9) == 0) {

        /* build the response */
        snprintf(response, max_resp_len, "identify;serveCSM;");
        response_len = strlen(response);

    /* "getPmtrs;" */
    } else if (strncmp(text, "getPmtrs;", 9) == 0) {

        icount = 0;

        /* build the response in JSON format */
        snprintf(response, max_resp_len, "getPmtrs;[");
        response_len = strlen(response);

        /* external Parameters first */
        for (ipmtr = 1; ipmtr <= MODL->npmtr; ipmtr++) {
            if (MODL->pmtr[ipmtr].type != OCSM_EXTERNAL) continue;

            snprintf(entry, MAX_STR_LEN, "{\"name\":\"%s\",\"type\":%d,\"nrow\":%d,\"ncol\":%d,\"value\":[",
                     MODL->pmtr[ipmtr].name,
                     MODL->pmtr[ipmtr].type,
                     MODL->pmtr[ipmtr].nrow,
                     MODL->pmtr[ipmtr].ncol);
            addToResponse(entry);

            index = 0;
            for (irow = 1; irow <= MODL->pmtr[ipmtr].nrow; irow++) {
                for (icol = 1; icol <= MODL->pmtr[ipmtr].ncol; icol++) {
                    if (irow < MODL->pmtr[ipmtr].nrow || icol < MODL->pmtr[ipmtr].ncol) {
                        snprintf(entry, MAX_STR_LEN, "%lg,", MODL->pmtr[ipmtr].value[index++]);
                    } else {
                        snprintf(entry, MAX_STR_LEN, "%lg],\"dot\":[", MODL->pmtr[ipmtr].value[index++]);
                    }
                    addToResponse(entry);
                }
            }

            index = 0;
            for (irow = 1; irow <= MODL->pmtr[ipmtr].nrow; irow++) {
                for (icol = 1; icol <= MODL->pmtr[ipmtr].ncol; icol++) {
                    if (irow < MODL->pmtr[ipmtr].nrow || icol < MODL->pmtr[ipmtr].ncol) {
                        snprintf(entry, MAX_STR_LEN, "%lg,", MODL->pmtr[ipmtr].dot[index++]);
                    } else {
                        snprintf(entry, MAX_STR_LEN, "%lg]", MODL->pmtr[ipmtr].dot[index++]);
                    }
                    addToResponse(entry);
                }
            }

            icount++;
            if (icount < MODL->npmtr) {
                addToResponse("},");
            } else {
                addToResponse("}]");
            }
        }

        /* internal Parameters last */
        for (ipmtr = 1; ipmtr <= MODL->npmtr; ipmtr++) {
            if (MODL->pmtr[ipmtr].type != OCSM_INTERNAL) continue;

            snprintf(entry, MAX_STR_LEN, "{\"name\":\"%s\",\"type\":%d,\"nrow\":%d,\"ncol\":%d,\"value\":[",
                     MODL->pmtr[ipmtr].name,
                     MODL->pmtr[ipmtr].type,
                     MODL->pmtr[ipmtr].nrow,
                     MODL->pmtr[ipmtr].ncol);
            addToResponse(entry);

            index = 0;
            for (irow = 1; irow <= MODL->pmtr[ipmtr].nrow; irow++) {
                for (icol = 1; icol <= MODL->pmtr[ipmtr].ncol; icol++) {
                    if (irow < MODL->pmtr[ipmtr].nrow || icol < MODL->pmtr[ipmtr].ncol) {
                        snprintf(entry, MAX_STR_LEN, "%lg,", MODL->pmtr[ipmtr].value[index++]);
                    } else {
                        snprintf(entry, MAX_STR_LEN, "%lg],\"dot\":[", MODL->pmtr[ipmtr].value[index++]);
                    }
                    addToResponse(entry);
                }
            }

            for (irow = 1; irow <= MODL->pmtr[ipmtr].nrow; irow++) {
                for (icol = 1; icol <= MODL->pmtr[ipmtr].ncol; icol++) {
                    if (irow < MODL->pmtr[ipmtr].nrow || icol < MODL->pmtr[ipmtr].ncol) {
                        snprintf(entry, MAX_STR_LEN, "0,");
                    } else {
                        snprintf(entry, MAX_STR_LEN, "0]");
                    }
                    addToResponse(entry);
                }
            }

            icount++;
            if (icount < MODL->npmtr) {
                addToResponse("},");
            } else {
                addToResponse("}]");
            }
        }

    /* "newPmtr;name;nrow;ncol; value1; ..." */
    } else if (strncmp(text, "newPmtr;", 8) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
        }

        /* extract arguments */
        nrow = 0;
        ncol = 0;

        if (getToken(text,  1, name) == 0) name[0] = '\0';
        if (getToken(text,  2, arg1)     ) nrow = strtol(arg1, &pEnd, 10);
        if (getToken(text,  3, arg2)     ) ncol = strtol(arg2, &pEnd, 10);

        /* store an undo snapshot */
        status = storeUndo("newPmtr", name);

        /* build the response */
        status = ocsmNewPmtr(MODL, name, OCSM_EXTERNAL, nrow, ncol);

        if (status == SUCCESS) {
            ipmtr = MODL->npmtr;

            i = 4;
            for (irow = 1; irow <= nrow; irow++) {
                for (icol = 1; icol <= ncol; icol++) {
                    if (getToken(text, i, arg3)) {
                        (void) ocsmSetValu(MODL, ipmtr, irow, icol, arg3);
                    }

                    i++;
                }
            }

            snprintf(response, max_resp_len, "newPmtr;");
        } else {
            snprintf(response, max_resp_len, "ERROR:: newPmtr() detected: %s",
                     ocsmGetText(status));
        }
        response_len = strlen(response);

    /* "setPmtr;pmtrname;irow;icol;value1; " */
    } else if (strncmp(text, "setPmtr;", 8) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
        }

        /* extract arguments */
        ipmtr = 0;
        irow  = 0;
        icol  = 0;

        getToken(text, 1, arg1);

        for (jpmtr = 1; jpmtr <= MODL->npmtr; jpmtr++) {
            if (strcmp(arg1, MODL->pmtr[jpmtr].name) == 0) {
                ipmtr = jpmtr;
                break;
            }
        }

        if (ipmtr > 0) {
            if (getToken(text, 2, arg2)) irow  = strtol(arg2, &pEnd, 10);
            if (getToken(text, 3, arg3)) icol  = strtol(arg3, &pEnd, 10);

            /* store an undo snapshot */
            status = storeUndo("setPmtr", MODL->pmtr[ipmtr].name);

            if (getToken(text, 4, arg4)) {
                status = ocsmSetValu(MODL, ipmtr, irow, icol, arg4);
            } else {
                status = -999;
            }

            /* build the response */
            if (status == SUCCESS) {
                snprintf(response, max_resp_len, "setPmtr;");
            } else {
                snprintf(response, max_resp_len, "ERROR:: setPmtr(%d, %d, %d) detected: %s",
                         ipmtr, irow, icol, ocsmGetText(status));
            }
        } else {
            snprintf(response, max_resp_len, "ERROR:: setPmtr could not find name \"%s\"", arg1);
        }
        response_len = strlen(response);

    /* "clrVels;" */
    } else if (strncmp(text, "clrVels;", 8) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
        }

        status = ocsmSetVelD(MODL, 0, 0, 0, 0.0);

        /* store an undo snapshot */
        status = storeUndo("clrVels", "");

        /* build the response */
        if (status == SUCCESS) {
            snprintf(response, max_resp_len, "clrVels;");
        } else {
            snprintf(response, max_resp_len, "ERROR:: clrVels() detected: %s",
                     ocsmGetText(status));
        }
        response_len = strlen(response);

    /* "setVel;pmtrname;irow;icol;vel; " */
    } else if (strncmp(text, "setVel;", 7) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
        }

        /* extract arguments */
        ipmtr = 0;
        irow  = 0;
        icol  = 0;

        getToken(text, 1, arg1);

        for (jpmtr = 1; jpmtr <= MODL->npmtr; jpmtr++) {
            if (strcmp(arg1, MODL->pmtr[jpmtr].name) == 0) {
                ipmtr = jpmtr;
                break;
            }
        }

        if (ipmtr > 0) {
            if (getToken(text, 2, arg2)) irow  = strtol(arg2, &pEnd, 10);
            if (getToken(text, 3, arg3)) icol  = strtol(arg3, &pEnd, 10);

            if (getToken(text, 4, arg4)) {
                status = ocsmSetVel(MODL, ipmtr, irow, icol, arg4);
            } else {
                status = -999;
            }

            /* store an undo snapshot */
            status = storeUndo("setVel", MODL->pmtr[ipmtr].name);

            /* build the response */
            if (status == SUCCESS) {
                snprintf(response, max_resp_len, "setVel;");
            } else {
                snprintf(response, max_resp_len, "ERROR:: setVel(%d, %d, %d) detected: %s",
                         ipmtr, irow, icol, ocsmGetText(status));
            }
        } else {
            snprintf(response, max_resp_len, "ERROR:: setVel could not find name \"%s\"", arg1);
        }
        response_len = strlen(response);

    /* "getBrchs;" */
    } else if (strncmp(text, "getBrchs;", 9) == 0) {

        /* build the response in JSON format */
        snprintf(response, max_resp_len, "getBrchs;[");
        response_len = strlen(response);

        for (ibrch = 1; ibrch <= MODL->nbrch; ibrch++) {
            snprintf(entry, MAX_STR_LEN, "{\"name\":\"%s\",\"type\":\"%s\",\"actv\":%d,\"attrs\":[",
                     MODL->brch[ibrch].name,
         ocsmGetText(MODL->brch[ibrch].type),
                     MODL->brch[ibrch].actv);
            addToResponse(entry);

            for (iattr = 0; iattr < MODL->brch[ibrch].nattr; iattr++) {
                if (iattr < MODL->brch[ibrch].nattr-1) {
                    snprintf(entry, MAX_STR_LEN, "[\"%s\",\"%s\"],",
                             MODL->brch[ibrch].attr[iattr].name,
                             MODL->brch[ibrch].attr[iattr].value);
                } else {
                    snprintf(entry, MAX_STR_LEN, "[\"%s\",\"%s\"]",
                             MODL->brch[ibrch].attr[iattr].name,
                             MODL->brch[ibrch].attr[iattr].value);
                }
                addToResponse(entry);
            }

            snprintf(entry, MAX_STR_LEN, "],\"ileft\":%d,\"irite\":%d,\"ichld\":%d,\"args\":[",
                     MODL->brch[ibrch].ileft,
                     MODL->brch[ibrch].irite,
                     MODL->brch[ibrch].ichld);
            addToResponse(entry);

            if (MODL->brch[ibrch].narg >= 1) {
                snprintf(entry, MAX_STR_LEN, "\"%s\"", MODL->brch[ibrch].arg1);
                addToResponse(entry);
            }
            if (MODL->brch[ibrch].narg >= 2) {
                snprintf(entry, MAX_STR_LEN, ",\"%s\"", MODL->brch[ibrch].arg2);
                addToResponse(entry);
            }
            if (MODL->brch[ibrch].narg >= 3) {
                snprintf(entry, MAX_STR_LEN, ",\"%s\"", MODL->brch[ibrch].arg3);
                addToResponse(entry);
            }
            if (MODL->brch[ibrch].narg >= 4) {
                snprintf(entry, MAX_STR_LEN, ",\"%s\"", MODL->brch[ibrch].arg4);
                addToResponse(entry);
            }
            if (MODL->brch[ibrch].narg >= 5) {
                snprintf(entry, MAX_STR_LEN, ",\"%s\"", MODL->brch[ibrch].arg5);
                addToResponse(entry);
            }
            if (MODL->brch[ibrch].narg >= 6) {
                snprintf(entry, MAX_STR_LEN, ",\"%s\"", MODL->brch[ibrch].arg6);
                addToResponse(entry);
            }
            if (MODL->brch[ibrch].narg >= 7) {
                snprintf(entry, MAX_STR_LEN, ",\"%s\"", MODL->brch[ibrch].arg7);
                addToResponse(entry);
            }
            if (MODL->brch[ibrch].narg >= 8) {
                snprintf(entry, MAX_STR_LEN, ",\"%s\"", MODL->brch[ibrch].arg8);
                addToResponse(entry);
            }
            if (MODL->brch[ibrch].narg >= 9) {
                snprintf(entry, MAX_STR_LEN, ",\"%s\"", MODL->brch[ibrch].arg9);
                addToResponse(entry);
            }

            if (ibrch < MODL->nbrch) {
                snprintf(entry, MAX_STR_LEN, "]},");
            } else {
                snprintf(entry, MAX_STR_LEN, "]}]");
            }
            addToResponse(entry);
        }

    /* "newBrch;ibrch;type;arg1;arg2;arg3;arg4;arg5;arg6;arg7;arg8;arg9; aname1;avalu1;" */
    } else if (strncmp(text, "newBrch;", 8) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
        }

        /* extract arguments */
        ibrch = 0;
        itype = 0;

        if (getToken(text,  1, arg1)) ibrch = strtol(arg1, &pEnd, 10);

        if (getToken(text,  2, type) >  0) itype = ocsmGetCode(type);
        if (getToken(text,  3, arg1) == 0) arg1[0] = '\0';
        if (getToken(text,  4, arg2) == 0) arg2[0] = '\0';
        if (getToken(text,  5, arg3) == 0) arg3[0] = '\0';
        if (getToken(text,  6, arg4) == 0) arg4[0] = '\0';
        if (getToken(text,  7, arg5) == 0) arg5[0] = '\0';
        if (getToken(text,  8, arg6) == 0) arg6[0] = '\0';
        if (getToken(text,  9, arg7) == 0) arg7[0] = '\0';
        if (getToken(text, 10, arg8) == 0) arg8[0] = '\0';
        if (getToken(text, 11, arg9) == 0) arg9[0] = '\0';

        /* store an undo snapshot */
        status = storeUndo("newBrch", type);

        /* build the response */
        status = ocsmNewBrch(MODL, ibrch, itype, arg1, arg2, arg3,
                                                 arg4, arg5, arg6,
                                                 arg7, arg8, arg9);
        if (status == SUCCESS) {
            status = ocsmCheck(MODL);

            if (status == SUCCESS) {
                snprintf(response, max_resp_len, "newBrch;");
            } else {
                snprintf(response, max_resp_len, "ERROR:: newBrch(ibrch=%d) detected: %s",
                         ibrch, ocsmGetText(status));
            }
        } else {
            snprintf(response, max_resp_len, "ERROR:: newBrch(ibrch=%d) detected: %s",
                     ibrch, ocsmGetText(status));
        }
        response_len = strlen(response);


    /* "setBrch;ibrch;name;actv;arg1;arg2;arg3;arg4;arg5;arg6;arg7;arg8;arg9; aname1;avalu1; ..." */
    } else if (strncmp(text, "setBrch;", 8) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
        }

        /* extract arguments */
        ibrch = 0;
        actv  = 0;

        if (getToken(text, 1, arg1)) ibrch = strtol(arg1, &pEnd, 10);

        /* store an undo snapshot */
        status = storeUndo("setBrch", MODL->brch[ibrch].name);

        /* build the response */
        if (ibrch >= 1 && ibrch <= MODL->nbrch) {
            if (getToken(text,  2, name)) (void) ocsmSetName(MODL, ibrch, name);

            if (getToken(text,  3, arg1)) {
                if (strcmp(arg1, "suppressed") == 0) {
                    (void) ocsmSetBrch(MODL, ibrch, OCSM_SUPPRESSED);
                    actv = 1;
                } else {
                    (void) ocsmSetBrch(MODL, ibrch, OCSM_ACTIVE);
                    actv = 1;
                }
            }

            if (getToken(text,  4, arg1)) (void) ocsmSetArg(MODL, ibrch, 1, arg1);
            if (getToken(text,  5, arg2)) (void) ocsmSetArg(MODL, ibrch, 2, arg2);
            if (getToken(text,  6, arg3)) (void) ocsmSetArg(MODL, ibrch, 3, arg3);
            if (getToken(text,  7, arg4)) (void) ocsmSetArg(MODL, ibrch, 4, arg4);
            if (getToken(text,  8, arg5)) (void) ocsmSetArg(MODL, ibrch, 5, arg5);
            if (getToken(text,  9, arg6)) (void) ocsmSetArg(MODL, ibrch, 6, arg6);
            if (getToken(text, 10, arg7)) (void) ocsmSetArg(MODL, ibrch, 7, arg7);
            if (getToken(text, 11, arg8)) (void) ocsmSetArg(MODL, ibrch, 8, arg8);
            if (getToken(text, 12, arg9)) (void) ocsmSetArg(MODL, ibrch, 9, arg9);

            i = 13;
            while (1) {
                if (getToken(text, i++, name) == 0) break;
                if (getToken(text, i++, valu) == 0) break;

                (void) ocsmSetAttr(MODL, ibrch, name, valu);
            }

            if (actv > 0) {
                status = ocsmCheck(MODL);

                if (status >= SUCCESS) {
                    snprintf(response, max_resp_len, "setBrch;");
                } else {
                    snprintf(response, max_resp_len, "ERROR:: setBrch(ibrch=%d) detected: %s",
                             ibrch, ocsmGetText(status));
                }
            } else {
                snprintf(response, max_resp_len, "setBrch;");
            }
        } else {
            status = OCSM_ILLEGAL_BRCH_INDEX;
            snprintf(response, max_resp_len, "ERROR: setBrch(%d) detected: %s",
                     ibrch, ocsmGetText(status));
        }
        response_len = strlen(response);


    /* "delBrch;ibrch;" */
    } else if (strncmp(text, "delBrch;", 8) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
        }

        /* extract argument */
        ibrch = 0;

        if (getToken(text, 1, arg1)) ibrch = strtol(arg1, &pEnd, 10);

        /* store an undo snapshot */
        status = storeUndo("delBrch", MODL->brch[ibrch].name);

        /* delete the Branch */
        status = ocsmDelBrch(MODL, ibrch);

        /* check that the Branches are properly ordered */
        if (status == SUCCESS) {
            status = ocsmCheck(MODL);
        }

        /* build the response */
        if (status == SUCCESS) {
            snprintf(response, max_resp_len, "delBrch;");
        } else {
            snprintf(response, max_resp_len, "ERROR:: delBrch(%d) detected: %s",
                     ibrch, ocsmGetText(status));
        }
        response_len = strlen(response);

    /* "setAttr;ibrch;aname;avalue;" */
    } else if (strncmp(text, "setAttr;", 8) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
        }

        /* extract arguments */
        ibrch = 0;

        if (getToken(text, 1, arg1)) ibrch = strtol(arg1, &pEnd, 10);
        getToken(text, 2, arg2);
        getToken(text, 3, arg3);

        /* store an undo snapshot */
        status = storeUndo("setAttr", MODL->brch[ibrch].name);

        /* set the Attribute */
        status = ocsmSetAttr(MODL, ibrch, arg2, arg3);

        /* build the response */
        if (status == SUCCESS) {
            snprintf(response, max_resp_len, "setAttr;");
        } else {
            snprintf(response, max_resp_len, "ERROR: setAttr(%d, %s, %s) detected: %s",
                     ibrch, arg2, arg3, ocsmGetText(status));
        }
        response_len = strlen(response);

    /* "undo;" */
    } else if (strncmp(text, "undo;", 5) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
        }

        /* build the response */
        if (nundo <= 0) {
            snprintf(response, max_resp_len, "ERROR:: there is nothing to undo");

        } else {
            /* remove the current MODL */
            status = ocsmFree(modl);
            if (status < SUCCESS) {
                snprintf(response, max_resp_len, "ERROR:: undo(%s) detected: %s",
                         arg1, ocsmGetText(status));
            } else {
                /* repoint MODL to the saved modl */
                modl = undo_modl[--nundo];
                MODL = (modl_T*)modl;

                snprintf(response, max_resp_len, "undo;%s;", undo_text[nundo]);
            }
        }
        response_len = strlen(response);

    /* "save;filename" */
    } else if (strncmp(text, "save;", 5) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
        }

        /* extract argument */
        getToken(text, 1, filename);

        /* save the file */
        status = ocsmSave(MODL, filename);

        /* build the response */
        if (status == SUCCESS) {
            snprintf(response, max_resp_len, "save;");
        } else {
            snprintf(response, max_resp_len, "ERROR:: save(%s) detected: %s",
                     arg1, ocsmGetText(status));
        }
        response_len = strlen(response);

    /* "getCsmFile;" */
    } else if (strncmp(text, "getCsmFile;", 11) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
        }

        /* build the response */
        snprintf(response, max_resp_len, "getCsmFile;%s;", filename);
        response_len = strlen(response);

        /* add the input file, line by line */
        if (strlen(filename) > 0) {
            fp = fopen(filename, "r");

            while (1) {
                entry = fgets(entry, MAX_STR_LEN-1, fp);
                if (feof(fp) != 0) break;

                addToResponse(entry);
            }

            fclose(fp);
            fp = NULL;
        }

    /* "setCsmFileBeg;" */
    } else if (strncmp(text, "setCsmFileBeg;", 14) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
        }

        /* open the casefile and overwrite */
        fp = fopen(filename, "w");

        ichar = 14;
        while (text[ichar] != '\0') {
            fprintf(fp, "%c", text[ichar++]);
        }

    /* "setCsmFileMid;" */
    } else if (strncmp(text, "setCsmFileMid;", 14) == 0) {

        ichar = 14;
        while (text[ichar] != '\0') {
            fprintf(fp, "%c", text[ichar++]);
        }

    /* "setCsmFileEnd;" */
    } else if (strncmp(text, "setCsmFileEnd;", 14) == 0) {

        fclose(fp);
        fp = NULL;

        /* load the new case */
        status = ocsmFree(modl);
        status = ocsmFree(NULL);

        status = ocsmLoad(filename, &modl);

        if (status < 0) {
            snprintf(response, max_resp_len, "ERROR:: build() detected: %s",
                     ocsmGetText(status));
        } else {
            status = buildBodys(0, &builtTo, &buildStatus);

            if (builtTo < 0) {
                snprintf(response, max_resp_len, "ERROR:: build() detected \"%s\" in %s",
                         ocsmGetText(buildStatus), MODL->brch[1-builtTo].name);
            } else if (status == SUCCESS) {
                snprintf(response, max_resp_len, "build;%d;%d;", builtTo, nbody);
            } else {
                snprintf(response, max_resp_len, "ERROR:: build() detected: %s",
                         ocsmGetText(status));
            }
        }

        response_len = strlen(response);

    /* "build;"  */
    } else if (strncmp(text, "build;", 6) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
        }

        /* extract argument */
        ibrch = 0;

        if (getToken(text, 1, arg1)) ibrch = strtol(arg1, &pEnd, 10);

        /* build the response */
        status = buildBodys(ibrch, &builtTo, &buildStatus);

        if (builtTo < 0) {
            snprintf(response, max_resp_len, "ERROR:: build() detected \"%s\" in %s",
                     ocsmGetText(buildStatus), MODL->brch[1-builtTo].name);
        } else if (status == SUCCESS) {
            snprintf(response, max_resp_len, "build;%d;%d;", builtTo, nbody);
        } else {
            snprintf(response, max_resp_len, "ERROR:: build() detected: %s",
                     ocsmGetText(status));
        }
        response_len = strlen(response);

    /* "setLims;lo;hi;" */
    } else if (strncmp(text, "setLims;", 8) == 0) {

        /* extract arguments */
        if (getToken(text, 1, arg1)) lims[0] = strtod(arg1, &pEnd);
        if (getToken(text, 2, arg2)) lims[1] = strtod(arg2, &pEnd);

        /* build the response */
        snprintf(response, max_resp_len, "setLims;");
        response_len = strlen(response);

        /* update the scene graph */
        if (batch == 0) {
            buildSceneGraph();
        }
    }

    free(entry);
    free(arg9);
    free(arg8);
    free(arg7);
    free(arg6);
    free(arg5);
    free(arg4);
    free(arg3);
    free(arg2);
    free(arg1);
    free(valu);
    free(type);
    free(name);

}


/***********************************************************************/
/*                                                                     */
/*   addToResponse - add text to response (with buffer length protection) */
/*                                                                     */
/***********************************************************************/

static void
addToResponse(char   *text)             /* (in)  text to add */
{
    int    text_len;

    text_len = strlen(text);

    while (response_len+text_len > max_resp_len-2) {
        max_resp_len +=4096;
        SPRINT1(2, "increasing max_resp_len=%d", max_resp_len);

        response = (char*) realloc(response, max_resp_len*sizeof(char));
        if (response == NULL) {
            SPRINT0(0, "ERROR:: error mallocing response");
            exit(0);
        }
    }

    strcpy(&(response[response_len]), text);

    response_len += text_len;
}


/***********************************************************************/
/*                                                                     */
/*   getToken - get a token from a string                              */
/*                                                                     */
/***********************************************************************/

static int
getToken(char   *text,                  /* (in)  full text */
         int    nskip,                  /* (in)  tokens to skip */
         char   *token)                 /* (out) token */
{
    int    lentok, i, count, iskip;

    token[0] = '\0';
    lentok   = 0;

    /* count the number of semi-colons */
    count = 0;
    for (i = 0; i < strlen(text); i++) {
        if (text[i] == ';') {
            count++;
        }
    }

    if (count < nskip+1) return 0;

    /* skip over nskip tokens */
    i = 0;
    for (iskip = 0; iskip < nskip; iskip++) {
        while (text[i] != ';') {
            i++;
        }
        i++;
    }

    /* extract the token we are looking for */
    while (text[i] != ';') {
        token[lentok++] = text[i++];
        token[lentok  ] = '\0';
    }

    return strlen(token);
}



/***********************************************************************/
/*                                                                     */
/*   spec_col - return color for a given scalar value                  */
/*                                                                     */
/***********************************************************************/

static void
spec_col(float  scalar,
         float  lims[],
         float  color[])
{
    int   indx;
    float frac;

    if (lims[0] == lims[1]) {
        color[0] = 0.0;
        color[1] = 1.0;
        color[2] = 0.0;
    } else if (scalar <= lims[0]) {
        color[0] = color_map[0];
        color[1] = color_map[1];
        color[2] = color_map[2];
    } else if (scalar >= lims[1]) {
        color[0] = color_map[3*255  ];
        color[1] = color_map[3*255+1];
        color[2] = color_map[3*255+2];
    } else {
        frac  = 255.0 * (scalar - lims[0]) / (lims[1] - lims[0]);
        if (frac < 0  ) frac = 0;
        if (frac > 255) frac = 255;
        indx  = frac;
        frac -= indx;
        if (indx == 255) {
            indx--;
            frac += 1.0;
        }

        color[0] = frac * color_map[3*(indx+1)  ] + (1.0-frac) * color_map[3*indx  ];
        color[1] = frac * color_map[3*(indx+1)+1] + (1.0-frac) * color_map[3*indx+1];
        color[2] = frac * color_map[3*(indx-1)+2] + (1.0-frac) * color_map[3*indx+2];
    }
}
