/*
 ************************************************************************
 *                                                                      *
 * udpImport -- udp file to import from a STEP file                     *
 *                                                                      *
 *            Written by John Dannenhoffer @ Syracuse University        *
 *            Patterned after code written by Bob Haimes  @ MIT         *
 *                                                                      *
 ************************************************************************
 */

#ifdef GEOM_EGADS

/*
 * Copyright (C) 2011/2014  John F. Dannenhoffer, III (Syracuse University)
 *
 * This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation; either
 *    version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 *    License along with this library; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *     MA  02110-1301  USA
 */

#define NUMUDPARGS 3
#include "udpUtilities.h"

/* shorthands for accessing argument values and velocities */
#define FILENAME(  IUDP)  ((char   *) (udps[IUDP].arg[0].val))
#define BODYNUMBER(IUDP)  ((int    *) (udps[IUDP].arg[1].val))[0]
#define NUMBODIES( IUDP)  ((int    *) (udps[IUDP].arg[2].val))[0]

/* data about possible arguments */
static char  *argNames[NUMUDPARGS] = {"filename",  "bodynumber", "numbodies", };
static int    argTypes[NUMUDPARGS] = {ATTRSTRING,  ATTRINT,      -ATTRINT,    };
static int    argIdefs[NUMUDPARGS] = {0,           1,            0,           };
static double argDdefs[NUMUDPARGS] = {0.,          0.,           0.,          };

/* get utility routines: udpErrorStr, udpInitialize, udpReset, udpSet,
                         udpGet, udpVel, udpClean, udpMesh */
#include "udpUtilities.c"


/*
 ************************************************************************
 *                                                                      *
 *   udpExecute - execute the primitive                                 *
 *                                                                      *
 ************************************************************************
 */

int
udpExecute(ego  context,                /* (in)  EGADS context */
           ego  *ebody,                 /* (out) Body pointer */
           int  *nMesh,                 /* (out) number of associated meshes */
           char *string[])              /* (out) error message */
{
    int     status = EGADS_SUCCESS;

    int     oclass, mtype, nbody, *senses;
    ego     geom, *bodies, emodel;

#ifdef DEBUG
    printf("udpExecute(context=%lx)\n", (long)context);
    printf("filename(0)   = %s\n", FILENAME(  0));
    printf("bodynumber(0) = %d\n", BODYNUMBER(0));
#endif

    /* default return values */
    *ebody  = NULL;
    *nMesh  = 0;
    *string = NULL;

    /* check arguments */
    if (udps[0].arg[1].size > 1) {
        printf(" udpExecute: bodynumber should be a scalar\n");
        status  = EGADS_RANGERR;
        *string = udpErrorStr(status);
        goto cleanup;

    } else if (BODYNUMBER(0) <= 0) {
        printf(" udpExecute: BodyNumber = %d < 0\n", BODYNUMBER(0));
        status  = EGADS_RANGERR;
        *string = udpErrorStr(status);
        goto cleanup;
    }

    /* cache copy of arguments for future use */
    status = cacheUdp();
    if (status < 0) {
        printf(" udpExecute: problem caching arguments\n");
        *string = udpErrorStr(status);
        goto cleanup;
    }

#ifdef DEBUG
    printf("filaname(%d)   = %s\n", numUdp, FILENAME(numUdp  ));
    printf("bodynumber(%d) = %d\n", numUdp, BODYNUMBER(numUdp));
#endif

    /* load the model */
    status = EG_loadModel(context, 0, FILENAME(0), &emodel);
    if ((status != EGADS_SUCCESS) || (emodel == NULL)) {
        *string = udpErrorStr(status);
        goto cleanup;
    }

    /* extract the ebody */
    status = EG_getTopology(emodel, &geom, &oclass, &mtype, NULL, &nbody,
                          &bodies, &senses);
    if (status != EGADS_SUCCESS) {
        *string = udpErrorStr(status);
        EG_deleteObject(emodel);
        emodel = NULL;
        goto cleanup;
    }

    if (BODYNUMBER(0) > nbody) {
        printf(" udpExecute: BodyNumber %d > %d\n", BODYNUMBER(0), nbody);
        status = EGADS_RANGERR;
    } else {
        status = EG_copyObject(bodies[BODYNUMBER(0)-1], NULL, ebody);
    }
    if (status != EGADS_SUCCESS) {
        *string = udpErrorStr(status);
        ebody = NULL;
        goto cleanup;
    }

    /* set the output value(s) */
    NUMBODIES(0) = nbody;

    /* remember this model (body) */
    udps[numUdp].ebody = *ebody;

#ifdef DEBUG
    printf("udpExecute -> *ebody=%lx\n", (long)(*ebody));
#endif

cleanup:
    if (status != EGADS_SUCCESS) {
        *string = udpErrorStr(status);
    }

    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   udpSensitivity - return sensitivity derivatives for the "real" argument *
 *                                                                      *
 ************************************************************************
 */

int
udpSensitivity(ego    ebody,            /* (in)  Body pointer */
   /*@unused@*/int    npnt,             /* (in)  number of points */
   /*@unused@*/int    entType,          /* (in)  OCSM entity type */
   /*@unused@*/int    entIndex,         /* (in)  OCSM entity index (bias-1) */
   /*@unused@*/double uvs[],            /* (in)  parametric coordinates for evaluation */
   /*@unused@*/double vels[])           /* (out) velocities */
{
    int iudp, judp;

    /* check that ebody matches one of the ebodys */
    iudp = 0;
    for (judp = 1; judp <= numUdp; judp++) {
        if (ebody == udps[judp].ebody) {
            iudp = judp;
            break;
        }
    }
    if (iudp <= 0) {
        return EGADS_NOTMODEL;
    }

    /* this routine is not written yet */
    return EGADS_NOLOAD;
}

#endif
