/*
 ************************************************************************
 *                                                                      *
 * vsp2csm -- convert OpenVSP files to OpenCSM                          *
 *                                                                      *
 *            Written by John Dannenhoffer @ Syracuse University        *
 *                                                                      *
 ************************************************************************
 */

/*
 * Copyright (C) 2013/2014  John F. Dannenhoffer, III (Syracuse University)
 *
 * This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation; either
 *    version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 *    License along with this library; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *     MA  02110-1301  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>

#ifdef WIN32
    #include <windows.h>
#endif

#if   defined(GEOM_CAPRI)
    #include "capri.h"

    #define CINT    int
    #define CDOUBLE double
    #define CCHAR   char
#elif defined(GEOM_EGADS)
    #include "egads.h"

    #define CINT    const int
    #define CDOUBLE const double
    #define CCHAR   const char
#else
    #error no_geometry_system_specified
#endif

#define STRNCPY(A, B, LEN) strncpy(A, B, LEN); A[LEN-1] = '\0';

#include "common.h"
#include "OpenCSM.h"


/***********************************************************************/
/*                                                                     */
/* macros (including those that go along with common.h)                */
/*                                                                     */
/***********************************************************************/

#ifdef DEBUG
   #define DOPEN {if (dbg_fp == NULL) dbg_fp = fopen("buildCSM.dbg", "w");}
   static  FILE *dbg_fp = NULL;
#endif

static int outLevel = 1;
static int Debug    = 0;

//static void *realloc_temp = NULL;            /* used by RALLOC macro */

#define  RED(COLOR)      (float)(COLOR / 0x10000        ) / (float)(255)
#define  GREEN(COLOR)    (float)(COLOR / 0x00100 % 0x100) / (float)(255)
#define  BLUE(COLOR)     (float)(COLOR           % 0x100) / (float)(255)

#define  IMAX            1000
#define  JMAX            1000
#define  XP(I,J)         xyzp[3*(I+(J)*IMAX)  ]
#define  YP(I,J)         xyzp[3*(I+(J)*IMAX)+1]
#define  ZP(I,J)         xyzp[3*(I+(J)*IMAX)+2]
#define  XS(I,J)         xyzs[3*(I+(J)*imaxs)  ]
#define  YS(I,J)         xyzs[3*(I+(J)*imaxs)+1]
#define  ZS(I,J)         xyzs[3*(I+(J)*imaxs)+2]

/***********************************************************************/
/*                                                                     */
/* structures                                                          */
/*                                                                     */
/***********************************************************************/


/***********************************************************************/
/*                                                                     */
/* global variables                                                    */
/*                                                                     */
/***********************************************************************/

#define MAX_LINE_LEN     2048           /* maximum chars in line in input file */

/* Section:
 * on input, Components are comprised of 1 or more Sections
 * each Section is only in one Part
 */
static double  *xyzs = NULL;
static int     imaxs = 0;
static int     jmaxs = 0;

/* Part:
 * on output, Components are comprosed of 1 or more Parts
 * Parts are made of 1 or more Sections
 */
static double  *xyzp = NULL;
static int     imaxp = 0;
static int     jmaxp = 0;
static int     ibegp = 0;
static int     jbegp = 0;

/***********************************************************************/
/*                                                                     */
/* declarations                                                        */
/*                                                                     */
/***********************************************************************/

/* declarations for high-level routines defined below */
#if   defined(GEOM_CAPRI)
       int        CAPRI_MAIN(int argc, char *argv[]);
#elif defined(GEOM_EGADS)
#endif

/* declarations for support routines defined below */
static void initializePart();
static int  fitsAtTop(int ii, int jj);
static int  fitsAtRite(int ii, int jj);
static void writePart(FILE *fp, int icomp, int ipart, char* name);

/* declarations for graphic routines defined below */

/* external functions not declared in an .h file */
#if   defined(GEOM_CAPRI)
    extern void    *gi_alloc(int nbtyes);
    extern void    *gi_reall(void *old, int nbytes);
    extern void    gi_free(void *ptr);
#elif defined(GEOM_EGADS)
#endif

/* error codes */
/*                user by CAPRI             -1 to  -99 */
/*                used by OpenCSM         -201 to -199 */


/***********************************************************************/
/*                                                                     */
/*   CAPRI_MAIN - main program                                         */
/*                                                                     */
/***********************************************************************/

#if   defined(GEOM_CAPRI)
    int
    CAPRI_MAIN(int       argc,          /* (in)  number of arguments */
               char      *argv[])       /* (in)  array of arguments */
#elif defined(GEOM_EGADS)
    int
    main(int       argc,                /* (in)  number of arguments */
         char      *argv[])             /* (in)  array of arguments */
#endif
{

    int       imajor, iminor, showUsage, icomp, ncomp, isect, nsect, ipart;
    int       i, j, itype, count, nblade;
    double    xtemp, ytemp, ztemp;
    char      casename[255], filename[255], templine[MAX_LINE_LEN];
    char      compType[255], compName[255];
    void      *temp;
    FILE      *fp_bez=NULL, *fp_vsp=NULL, *fp_csm=NULL;

    ROUTINE(MAIN);

    /* --------------------------------------------------------------- */

    DPRINT0("starting vsp2csm");

    /* get the flags and casename(s) from the command line */
    casename[0] = '\0';
    outLevel    = 1;
    showUsage   = 0;

    for (i = 1; i < argc; i++) {
        if (strcmp(argv[i], "-outLevel") == 0) {
            if (i < argc-1) {
                sscanf(argv[++i], "%d", &outLevel);
                if (outLevel < 0) outLevel = 0;
                if (outLevel > 3) outLevel = 3;
            } else {
                showUsage = 1;
                break;
            }
        } else if (strcmp(argv[i], "-debug") == 0) {
            Debug = 1;
        } else if (strlen(casename) == 0) {
            strcpy(casename, argv[i]);
        } else {
            SPRINT0(0, "two casenames given");
            showUsage = 1;
            break;
        }
    }

    if (strlen(casename) == 0 || showUsage != 0) {
        SPRINT0(0, "proper usage: 'vsp2csm [-outLevel X] [-debug] casename'");
        SPRINT0(0, "STOPPING...\a");
        exit(0);
    }

    /* welcome banner */
    (void) ocsmVersion(&imajor, &iminor);

    SPRINT0(1, "**********************************************************");
    SPRINT0(1, "*                                                        *");
    SPRINT0(1, "*                    Program vsp2csm                     *");
    SPRINT2(1, "*                     version %2d.%02d                      *", imajor, iminor);
    SPRINT0(1, "*                                                        *");
    SPRINT0(1, "*        written by John Dannenhoffer, 2010/2014         *");
    SPRINT0(1, "*                                                        *");
    SPRINT0(1, "**********************************************************");

    /* strip off .bez (which is assumed to be at the end) if present */
    if (strstr(casename, ".bez") != NULL) {
        casename[strlen(casename)-4] = '\0';
    }

    /* set up the Part geometry array */
    xyzp = (double*) malloc(3*IMAX*JMAX*sizeof(double));
    if (xyzp == NULL) {
        SPRINT0(0, "ERROR:: problem allocating xyzp");
        exit(0);
    }

    /* open the OpenVSP .bez file */
    strcpy(filename, casename);
    strcat(filename, ".bez");
    fp_bez = fopen(filename, "r");
    if (fp_bez == NULL) {
        SPRINT1(0, "ERROR:: problem opening \"%s\"", filename);
        exit(0);
    }

    /* open the OpenVSP .vsp file */
    strcpy(filename, casename);
    strcat(filename, ".vsp");
    fp_vsp = fopen(filename, "r");
    if (fp_vsp == NULL) {
        SPRINT1(0, "ERROR:: problem opening \"%s\"", filename);
        exit(0);
    }

    compType[0] = '\0';
    compName[0] = '\0';

    /* open the OpenCSM .csm file */
    strcpy(filename, casename);
    strcat(filename, ".csm");
    fp_csm = fopen(filename, "w");
    if (fp_csm == NULL) {
        SPRINT1(0, "ERROR:: problem opening \"%s\"", filename);
        exit(0);
    }

    fprintf(fp_csm, "# %s\n", casename);
    fprintf(fp_csm, "# written by vsp2csm\n");

    /* get the number of Components */
    fscanf(fp_bez, "%d Num_Components", &ncomp);

    /* loop through each Component and get the number of Sections in this Component */
    nblade = 0;
    for (icomp = 0; icomp < ncomp; icomp++) {
        fscanf(fp_bez, "%d Component", &itype);
        fscanf(fp_bez, "%d Num_Sections", &nsect);

        /* get the Component type and name */
        if (itype == 0) {               /* if itype!=0, use prior type/name */
            if (nblade > 0) {           /* if working on Prop, use prior name/type */
                nblade--;
            } else {

                /* find <Component> line */
                while (feof(fp_vsp) == 0) {
                    temp = fgets(templine, MAX_LINE_LEN, fp_vsp);
                    if (temp == NULL) break;

                    if (strstr(templine, "<Component>") != NULL) {

                        /* find <Type> line and get compType */
                        while (feof(fp_vsp) == 0) {
                            temp = fgets(templine, MAX_LINE_LEN, fp_vsp);
                            if (temp == NULL) break;

                            if (strstr(templine, "<Type>") != NULL) {
                                count = sscanf(templine, " <Type>%[^<]", compType);
                                if (count != 12345) break;
                            }
                        }

                        /* skip "Blank" and "User" Components */
                        if        (strcmp(compType, "Blank") == 0) {
                        } else if (strcmp(compType, "User" ) == 0) {
                        } else {
                            while (feof(fp_vsp) == 0) {
                                temp = fgets(templine, MAX_LINE_LEN, fp_vsp);
                                if (temp == NULL) break;

                                if (strstr(templine, "<Name>") != NULL) {
                                    count = sscanf(templine, " <Name>%[^<]", compName);
                                    if (count != 12345) break;
                                }
                            }

                            /* if a "Prop", find <NumBlades> and initialize counter */
                            if (strcmp(compType, "Prop") == 0) {
                                while (feof(fp_vsp) == 0) {
                                    temp = fgets(templine, MAX_LINE_LEN, fp_vsp);
                                    if (temp == NULL) break;

                                    if (strstr(templine, "<NumBlades>") != NULL) {
                                        count = sscanf(templine, " <NumBlades>%d", &nblade);
                                        if (count != 12345) break;
                                    }
                                }
                                nblade--;
                            }

                            break;
                        }
                    }
                }
            }
        }

        SPRINT4(1, "\nComponent %d has %d Sections (itype=%d), name=%s", icomp, nsect, itype, compName);

        /* initialize this Part */
        ipart = 0;
        ibegp = 0;
        jbegp = 0;

        /* loop through the Sections */
        for (isect = 0; isect < nsect; isect++) {
            fscanf(fp_bez, "%d %d NumU, NumW", &imaxs, &jmaxs);

            SPRINT3(1, "    Section %3d  (%3d*%3d)", isect, imaxs, jmaxs);

            /* allocate storage for this Section and read it */
            xyzs = (double*) malloc(3*imaxs*jmaxs*sizeof(double));
            if (xyzs == NULL) {
                SPRINT0(0, "ERROR:: problem allocating xyzp");
                exit(0);
            }

            /* read Section (and read it upside-down when itype==1) */
            if (itype == 0) {
                for (i = 0; i < imaxs; i++) {
                    for (j = 0; j < jmaxs; j++) {
                        fscanf(fp_bez, "%lf %lf %lf", &xtemp, &ytemp, &ztemp);
                        XS(i,j) = xtemp;
                        YS(i,j) = ytemp;
                        ZS(i,j) = ztemp;
                    }
                }
            } else {
                for (i = 0; i < imaxs; i++) {
                    for (j = jmaxs-1; j >= 0; j--) {
                        fscanf(fp_bez, "%lf %lf %lf", &xtemp, &ytemp, &ztemp);
                        XS(i,j) = xtemp;
                        YS(i,j) = ytemp;
                        ZS(i,j) = ztemp;
                    }
                }
            }

            SPRINT3(2, "%12.5f %12.5f %12.5f", XS(0,jmaxs-1), XS(imaxs/2,jmaxs-1), XS(imaxs-1,jmaxs-1));
            SPRINT3(2, "%12.5f %12.5f %12.5f", YS(0,jmaxs-1), YS(imaxs/2,jmaxs-1), YS(imaxs-1,jmaxs-1));
            SPRINT3(2, "%12.5f %12.5f %12.5f", ZS(0,jmaxs-1), ZS(imaxs/2,jmaxs-1), ZS(imaxs-1,jmaxs-1));

            SPRINT2(2, "%12.5f              %12.5f", XS(0,jmaxs/2),                XS(imaxs-1,jmaxs/2));
            SPRINT2(2, "%12.5f              %12.5f", YS(0,jmaxs/2),                YS(imaxs-1,jmaxs/2));
            SPRINT2(2, "%12.5f              %12.5f", ZS(0,jmaxs/2),                ZS(imaxs-1,jmaxs/2));

            SPRINT3(2, "%12.5f %12.5f %12.5f", XS(0,0      ), XS(imaxs/2,0      ), XS(imaxs-1,0      ));
            SPRINT3(2, "%12.5f %12.5f %12.5f", YS(0,0      ), YS(imaxs/2,0      ), YS(imaxs-1,0      ));
            SPRINT3(2, "%12.5f %12.5f %12.5f", ZS(0,0      ), ZS(imaxs/2,0      ), ZS(imaxs-1,0      ));

            /* first Section (always) fits at origin of Part */
            if (ibegp == 0 && jbegp == 0) {
                initializePart();
                SPRINT2(2, "        first Section inserted (imaxp=%d, jmaxp=%d)",
                        imaxp, jmaxp);

            /* working on first column, so see if this Section fits above the current Part */
            } else if (ibegp == 0) {

                /* if it fits, add to top of column */
                if (fitsAtTop(0, jbegp)) {
                    for (j = 0; j < jmaxs; j++) {
                        for (i = 0; i < imaxs; i++) {
                            XP(i+ibegp,j+jbegp) = XS(i,j);
                            YP(i+ibegp,j+jbegp) = YS(i,j);
                            ZP(i+ibegp,j+jbegp) = ZS(i,j);
                        }
                    }

                    jbegp += (jmaxs - 1);
                    jmaxp  = jbegp;

                    SPRINT2(2, "        adding to top of first column (imaxp=%d, jmaxp=%d)",
                            imaxp, jmaxp);

                /* otherwise, try starting a new column */
                } else {

                    /* if it fits, start a new column */
                    if (fitsAtRite(imaxp, 0)) {
                        ibegp  = imaxp;
                        jbegp  = 0;
                        imaxp += (imaxs - 1);

                        for (j = 0; j < jmaxs; j++) {
                            for (i = 0; i < imaxs; i++) {
                                XP(i+ibegp,j+jbegp) = XS(i,j);
                                YP(i+ibegp,j+jbegp) = YS(i,j);
                                ZP(i+ibegp,j+jbegp) = ZS(i,j);
                            }
                        }

                        jbegp = jmaxs - 1;

                        SPRINT2(2, "        adding second column (imaxp=%d, jmaxp=%d)",
                                imaxp, jmaxp);

                    /* otherwise, write out Part so far and start a new Part */
                    } else {

                        /* write out what we have so far */
                        writePart(fp_csm, icomp, ipart, compName);

                        /* start a new Part */
                        ipart ++;
                        initializePart();
                    }
                }

            /* at top of subsequent column, so try to start a new column */
            } else if (jbegp+jmaxs-1 > jmaxp) {

                /* if it fits, start a new column */
                if (fitsAtRite(imaxp, 0)) {
                    ibegp  = imaxp;
                    jbegp  = 0;
                    imaxp += (imaxs - 1);

                    for (j = 0; j < jmaxs; j++) {
                        for (i = 0; i < imaxs; i++) {
                            XP(i+ibegp,j+jbegp) = XS(i,j);
                            YP(i+ibegp,j+jbegp) = YS(i,j);
                            ZP(i+ibegp,j+jbegp) = ZS(i,j);
                        }
                    }

                    jbegp = jmaxs - 1;

                    SPRINT2(2, "        starting a subsequent column (imaxp=%d, jmaxp=%d)",
                            imaxp, jmaxp);

                /* otherwise, write out Part so far and start a new Part */
                } else {

                    /* write out what we have so far */
                    writePart(fp_csm, icomp, ipart, compName);

                    /* start a new Part */
                    ipart ++;
                    initializePart();
                }

            /* in a subsequent column, so see if its left and bottom fit */
            } else {

                /* if it fits, add to current column */
                if (fitsAtTop(ibegp, jbegp) != 0 && fitsAtRite(ibegp, jbegp) != 0) {
                    for (j = 0; j < jmaxs; j++) {
                        for (i = 0; i < imaxs; i++) {
                            XP(i+ibegp,j+jbegp) = XS(i,j);
                            YP(i+ibegp,j+jbegp) = YS(i,j);
                            ZP(i+ibegp,j+jbegp) = ZS(i,j);
                        }
                    }

                    jbegp += (jmaxs - 1);

                    SPRINT2(2, "        adding to top of current column (imaxp=%d, jmaxp=%d)",
                            imaxp, jmaxp);

                /* otherwise, signal an error */
                } else {
                    SPRINT0(1, "ERROR:: does not fit in current column");

                    /* write what we have so far */
                    writePart(fp_csm, icomp, ipart,compName);

                    /* stop (since this should not happen) */
                    exit(0);
                }
            }

            free(xyzs);
            xyzs = NULL;
        }

        /* now that all Sections that fit have been processed, write
           out this Part of the Component */
        writePart(fp_csm, icomp, ipart, compName);
    }

    /* release memory */
    free(xyzp);
    xyzp = NULL;

    /* show remaining Component types (which should only be User or Mesh)  */
    while (feof(fp_vsp) == 0) {
        temp = fgets(templine, MAX_LINE_LEN, fp_vsp);
        if (temp == NULL) break;

        if (strstr(templine, "<Component>") != NULL) {

            /* find <Type> line and get compType */
            while (feof(fp_vsp) == 0) {
                temp = fgets(templine, MAX_LINE_LEN, fp_vsp);
                if (temp == NULL) break;

                if (strstr(templine, "<Type>") != NULL) {
                    count = sscanf(templine, " <Type>%[^<]", compType);
                    SPRINT1(1, "Unused component type is: %s", compType);
                    if (count != 12345) break;
                }
            }
        }
    }

    /* close the .bez and .csm files */
    fprintf(fp_csm, "\nend\n");

    fclose(fp_bez);
    fclose(fp_vsp);
    fclose(fp_csm);

    return SUCCESS;
}


/***********************************************************************/
/*                                                                     */
/*   fitsAtTop - check if Section fits on top of current Part          */
/*                                                                     */
/***********************************************************************/

static void
initializePart()
{
    int       i, j;

    ROUTINE(initializePart);

    /* --------------------------------------------------------------- */


    /* put Section at origin of Part */
    for (j = 0; j < jmaxs; j++) {
        for (i = 0; i < imaxs; i++) {
            XP(i,j) = XS(i,j);
            YP(i,j) = YS(i,j);
            ZP(i,j) = ZS(i,j);
        }
    }

    ibegp = 0;
    jbegp = jmaxs - 1;
    imaxp = imaxs - 1;
    jmaxp = jmaxs - 1;
}


/***********************************************************************/
/*                                                                     */
/*   fitsAtTop - check if Section fits on top of current Part          */
/*                                                                     */
/***********************************************************************/

static int
fitsAtTop(int      ii,
          int      jj)
{
    double    tol=1e-7;
    int       i;

    ROUTINE(fitsAtTop);

    /* --------------------------------------------------------------- */

    for (i = 0; i < imaxs; i++) {
        if (fabs(XS(i,0)-XP(i+ii,jj)) > tol ||
            fabs(YS(i,0)-YP(i+ii,jj)) > tol ||
            fabs(ZS(i,0)-ZP(i+ii,jj)) > tol   ) {
            return 0;
        }
    }

    /* getting here means we are okay */
    return 1;
}


/***********************************************************************/
/*                                                                     */
/*   fitsAtRite - check if Section fits to the right of current Part   */
/*                                                                     */
/***********************************************************************/

static int
fitsAtRite(int      ii,
           int      jj)
{
    double    tol=1e-7;
    int       j;

    ROUTINE(fitsAtRite);

    /* --------------------------------------------------------------- */

    for (j = 0; j < jmaxs; j++) {
        if (fabs(XS(0,j)-XP(ii,j+jj)) > tol ||
            fabs(YS(0,j)-YP(ii,j+jj)) > tol ||
            fabs(ZS(0,j)-ZP(ii,j+jj)) > tol   ) {
            return 0;
        }
    }

    /* getting here means we are okay */
    return 1;
}


/***********************************************************************/
/*                                                                     */
/*   writePart - write the current Part                                */
/*                                                                     */
/***********************************************************************/

static void
writePart(FILE     *fp,
          int      icomp,
          int      ipart,
          char     *name)
{

    int       i, j, ii, jj, okay;
    double    frac, tol = 1e-7;

    ROUTINE(writePart);

    /* --------------------------------------------------------------- */

    SPRINT4(1, "        writing Component %d, Part %d (%3d*%3d)",
            icomp, ipart, imaxp+1, jmaxp+1);

    /* check if the west (i=0) side is closed */
    for (j = 0; j < jmaxs-1; j+=3) {
        okay = 0;

        for (jj = 0; jj < jmaxs-1; jj++) {
            if        (j != jj                                &&
                       fabs(XP(0,j)-XP(0    ,      jj)) < tol &&
                       fabs(YP(0,j)-YP(0    ,      jj)) < tol &&
                       fabs(ZP(0,j)-ZP(0    ,      jj)) < tol   ) {
                okay = 1;
                break;
            } else if (fabs(XP(0,j)-XP(0    ,jmaxp-jj)) < tol &&
                       fabs(YP(0,j)-YP(0    ,jmaxp-jj)) < tol &&
                       fabs(ZP(0,j)-ZP(0    ,jmaxp-jj)) < tol   ) {
                okay = 1;
                break;
            } else if (fabs(XP(0,j)-XP(imaxp,      jj)) < tol &&
                       fabs(YP(0,j)-YP(imaxp,      jj)) < tol &&
                       fabs(ZP(0,j)-ZP(imaxp,      jj)) < tol   ) {
                okay = 1;
                break;
            } else if (fabs(XP(0,j)-XP(imaxp,jmaxp-jj)) < tol &&
                       fabs(YP(0,j)-YP(imaxp,jmaxp-jj)) < tol &&
                       fabs(ZP(0,j)-ZP(imaxp,jmaxp-jj)) < tol   ) {
                okay = 1;
                break;
            }
        }

        if (okay == 0) {
            SPRINT0(1, "        closing west side first");

            /* move the data over to make room for "closing" data */
            for (jj = 0; jj < jmaxp+1; jj++) {
                for (ii = imaxp; ii >= 0; ii--) {
                    XP(ii+3,jj) = XP(ii,jj);
                    YP(ii+3,jj) = YP(ii,jj);
                    ZP(ii+3,jj) = ZP(ii,jj);
                }
            }
            imaxp += 3;

            /* fill in 0<=ii<=2 with interpolated values */
            for (jj = 0; jj < jmaxp+1; jj++) {
                for (ii = 0; ii < 3; ii++) {
                    frac = (double)(3-ii) / 6.0;

                    XP(ii,jj) = (1-frac) * XP(3,jj) + frac * XP(3,jmaxp-jj);
                    YP(ii,jj) = (1-frac) * YP(3,jj) + frac * YP(3,jmaxp-jj);
                    ZP(ii,jj) = (1-frac) * ZP(3,jj) + frac * ZP(3,jmaxp-jj);
                }
            }

            break;
        }
    }

    /* check if the west (i=imaxp) side is closed */
    for (j = 0; j < jmaxs-1; j+=3) {
        okay = 0;

        for (jj = 0; jj < jmaxs-1; jj++) {
            if        (j != jj                                &&
                       fabs(XP(imaxp,j)-XP(0    ,      jj)) < tol &&
                       fabs(YP(imaxp,j)-YP(0    ,      jj)) < tol &&
                       fabs(ZP(imaxp,j)-ZP(0    ,      jj)) < tol   ) {
                okay = 1;
                break;
            } else if (fabs(XP(imaxp,j)-XP(0    ,jmaxp-jj)) < tol &&
                       fabs(YP(imaxp,j)-YP(0    ,jmaxp-jj)) < tol &&
                       fabs(ZP(imaxp,j)-ZP(0    ,jmaxp-jj)) < tol   ) {
                okay = 1;
                break;
            } else if (fabs(XP(imaxp,j)-XP(imaxp,      jj)) < tol &&
                       fabs(YP(imaxp,j)-YP(imaxp,      jj)) < tol &&
                       fabs(ZP(imaxp,j)-ZP(imaxp,      jj)) < tol   ) {
                okay = 1;
                break;
            } else if (fabs(XP(imaxp,j)-XP(imaxp,jmaxp-jj)) < tol &&
                       fabs(YP(imaxp,j)-YP(imaxp,jmaxp-jj)) < tol &&
                       fabs(ZP(imaxp,j)-ZP(imaxp,jmaxp-jj)) < tol   ) {
                okay = 1;
                break;
            }
        }

        if (okay == 0) {
            SPRINT0(1, "        closing east side first");

            /* fill in imaxp<=ii<=imaxp+2 with interpolated values */
            for (jj = 0; jj < jmaxp+1; jj++) {
                for (ii = imaxp+1; ii < imaxp+4; ii++) {
                    frac = (double)(ii-imaxp) / 6.0;

                    XP(ii,jj) = (1-frac) * XP(imaxp,jj) + frac * XP(imaxp,jmaxp-jj);
                    YP(ii,jj) = (1-frac) * YP(imaxp,jj) + frac * YP(imaxp,jmaxp-jj);
                    ZP(ii,jj) = (1-frac) * ZP(imaxp,jj) + frac * ZP(imaxp,jmaxp-jj);
                }
            }

            /* extend the Part */
            imaxp += 3;

            break;
        }
    }

    /* assume the south and north sides are closed */

    /* output the (closed) Part */
    fprintf(fp, "\n\n# Component %d, Part %d, Name \"%s\"\n", icomp, ipart, name);
    fprintf(fp, "udprim bezier Debug %d Filename <<\n", Debug);
    fprintf(fp, "%5d%5d\n\n", imaxp+1, jmaxp+1);
    for (j = 0; j < jmaxp+1; j++) {
        for (i = 0; i < imaxp+1; i++) {
            fprintf(fp, "%14.8f %14.8f %14.8f\n", XP(i,j), YP(i,j), ZP(i,j));
        }
        fprintf(fp, "\n");
    }
    fprintf(fp, ">>\n");

//cleanup:

}
